--[[ SRB2:CE v3 Effects Lua
    This is where all the effects will go; logic will be ran in a seperate script.
]]

local mapobjectscale = mapobjectscale or FRACUNIT -- SRB2 might get this in the future, so we're doing this for in case

SRB2CE.ueTimer = ueTimer
SRB2CE.ueTimerMax = ueTimerMax

SRB2CE.ValidityCheck = function(player)
    if not player.valid return false end
    if not player.mo return false end
    if player.spectator return false end
    if not player.khaos return false end
    if player.khaos.mapchange return true end -- force the game to remove all effects first
    if mapheaderinfo[gamemap].srb2cecity return false end -- don't play on tutorial/chaos city
    --if player.bot and gametype != GT_RACE return false end -- no longer needed, keeping commented out just in case
    if player.quittime return false end --TODO: does SRB2 even have this?
    if not SRB2CE_Enabled return false end

    return true
end

local function indexOf(array, value)
    for i, v in ipairs(array) do
        if v == value then
            return i
        end
    end
    return nil
end

local function isPlayerRespawning(player) -- TODO: fix me for SRB2
    if player.playerstate != PST_LIVE then
        return true
    end

    return false
end

local function SpawnFlingRing(mobj, timer, speed, z)
    local ring = P_SpawnMobj(mobj.x, mobj.y, z or mobj.z, MT_FLINGRING)
    ring.fuse = (timer and timer*TICRATE) or 5*TICRATE
    ring.target = mobj.mo
    ring.destscale = mobj.destscale
    ring.scale = mobj.scale
    ring.angle = mobj.angle
    if not speed then
        speed = 9
    end

    if (mobj.eflags & MFE_VERTICALFLIP) then
        ring.momz = mobj.momz - mobj.scale*speed
    else
        ring.momz = mobj.momz + mobj.scale*speed
    end

    ring.momx = mobj.momx - (P_RandomRange(-speed, speed)*mobj.scale)
    ring.momy = mobj.momy - (P_RandomRange(-speed, speed)*mobj.scale)

    return ring
end


local function keyOf(tbl, value)
    for k, v in pairs(tbl) do
        if v == value then
            return k
        end
    end
    return nil
end

local function indexOfName(array, value) --thanks to amperbee for telling me the obvious that i can just modify this function lol
    for i, v in ipairs(array) do
        if v.name == value then
            return i
        end
    end
    return nil
end

local function indexOfName2(array, value)
    for i, v in ipairs(array) do
        if string.lower(v.name) == string.lower(value) then -- makes this case insensitive
            return i
        end
    end
    return nil
end

local function ShuffleInPlace(t) -- thanks to https://stackoverflow.com/a/68486276 for this, i've only modified it slightly
    for i = #t, 2, -1 do
        local j = P_RandomRange(1, i)
        t[i], t[j] = t[j], t[i]
    end
end

local function fullcolors(string, offset) -- if you use this, leave this credit for me: Ibly/EeveeEuphoria
    local indexer = 0
    if offset then
        indexer = offset
    end
    local colortable = {"\x80", "\x81", "\x82", "\x83", "\x84", "\x85", "\x86", "\x87", "\x88", "\x89", "\x8A", "\x8B", "\x8C", "\x8D", "\x8E", "\x8F"}
    local rainbowed = ""

    for c in string:gmatch"." do
        if c != " " then
            indexer = $ + 1
            if indexer > #colortable then
                indexer = 1
            end
            c = colortable[indexer] + $
        end
        rainbowed = $ + c
    end
    return rainbowed
end

local function BootlegAttract(mo, target, speed, xoffs, yoffs, zoffs) -- originated from tatsuru's conversion of SRB2 source code
    xoffs = $ or 0
    yoffs = $ or 0
    zoffs = $ or 0

    local newx = target.x + xoffs
    local newy = target.y + yoffs
    local newz = target.z + zoffs
    local dist = R_PointToDist2(0, 0, R_PointToDist2(mo.x, mo.y, newx, newy), newz - mo.z)

    if speed < dist
        mo.momx = FixedMul(FixedDiv(newx - mo.x, dist), speed)
        mo.momy = FixedMul(FixedDiv(newy - mo.y, dist), speed)
        mo.momz = FixedMul(FixedDiv(newz - mo.z, dist), speed)
    elseif speed > dist
        mo.momx = 0
        mo.momy = 0
        mo.momz = 0
        local CoolTeleFunction = P_TeleportMove
        if P_SetOrigin then -- new function for interpolation mod; don't use it if it doesn't exist
            CoolTeleFunction = P_SetOrigin
        end
        CoolTeleFunction(mo, newx, newy, newz)
    end
end


local function GivePizza(player)
    local pizzatable = {{"small", FRACUNIT/4, (30*3)*FRACUNIT}, {"medium", FRACUNIT, 38*FRACUNIT}, {"large", FRACUNIT + FRACUNIT*(2/4), 44*FRACUNIT}, {"extra large", FRACUNIT*2,49*FRACUNIT}} --name, scale, speed
    local thetoppings = {{"pepperoni", S_CHAOTICPIZZA}, {"pizzaface", S_CHAOTICPIZZAFACE}, {"deton", S_CHAOTICXPLOSPIZZA}}
    if player.khaossp.drowned then
        table.insert(thetoppings, {"bones", S_CHAOSSKULL}) -- you'll have to ask for a boneless pizza
    end

    player.khaos.PizzaRemoval = nil
    local location = player.mo.ceilingz-(6*mapobjectscale)
    if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
        location = player.mo.floorz+(6*mapobjectscale)
    end
    local thepizzais = P_SpawnMobj(player.mo.x, player.mo.y, location, MT_JIMBO)
    thepizzais.target = player.mo
    local choosesize = P_RandomRange(1, 3)
    if SRB2CE.AmplifyChecker(player) then
        choosesize = P_RandomRange(2, 4)
    end
    local size = pizzatable[choosesize]
    local toppings = thetoppings[P_RandomRange(1, #thetoppings)]
    thepizzais.state = toppings[2]
    thepizzais.scale = size[2]
    thepizzais.speed = size[3]

    if not player.khaos.PizzaGivenAlready then
        local StringMan = khaos_rainbowtext(player.name) .. "\x80 has ordered a " .. size[1] .. " sized pizza, with " .. toppings[1] .. " toppings to themself."
        print(StringMan)
        if netgame then chatprint(StringMan, true) end
    end

    player.khaos.PizzaSpawned = true
end

SRB2CE.KhaosTotal = #KhaosEffectsTable

local function addEffect(tbl)
    -- basic sanity checks
    if not tbl.name then
        assert(nil, "ERROR: Invalid or no name given to effect!")
    end
    if tbl.chance == nil then
        assert(nil, "ERROR: Invalid or no 'chance' value set!")
    end
    if tbl.duration == nil or type(tbl.duration) != "number" then
        assert(nil, "ERROR: Invalid or no 'duration' value set!")
    end
    -- do NOT error out if no icon/sprite is there, that's something i use and is accounted for
    for key, value in pairs(KhaosEffectsTable) do
        if tbl.name == value.name then
            assert(nil, "ERROR: Effect name is already taken by something else!")
        end
    end
    if not tbl.noprepend and tbl.sprite then
        if VERSION >= 2 then
            tbl.sprite = "KH_" + $
        else
            tbl.sprite = "CH_" + $
        end
    end
    table.insert(KhaosEffectsTable, tbl)
    if not tbl.dontcountme then
        SRB2CE.KhaosTotal = $ + 1
    end
end

SRB2CE.addEffect = addEffect

local function GlobalNonConflict(effectid) -- this is all the stuff that isn't player specific. used mainly for "you get the bees"
    local effecter = KhaosEffectsTable[effectid]

    if effecter.dontrollme then
        SRB2CE.DebugPrint(nil, "don't roll this!")
        return false
    end

    if effecter.nonet and netgame then
        SRB2CE.DebugPrint(player, "did not choose due to being in a netgame")
        return false
    end

    if effecter.nomatch and SRB2CE.IsMatchMode() then
        SRB2CE.DebugPrint(player, "did not choose due to being in match mode")
        return false
    end

    --[[if effecter.nobattle and gametype == GT_BATTLE then
        SRB2CE.DebugPrint(nil, "we're in battle mode, skip")
        return false
    end

    if effecter.battleonly and gametype != GT_BATTLE then
        SRB2CE.DebugPrint(nil, "this isn't battle mode!")
        return false
    end]]

    if effecter.nonights and (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
        SRB2CE.DebugPrint(nil, "NiGHTs level, skip")
        return false
    end

    if effecter.nospecialstage and G_IsSpecialStage(gamemap) then
        SRB2CE.DebugPrint(nil, "we're in a special stage, skip")
        return false
    end

    if effecter.noboss and (mapheaderinfo[gamemap].bonustype == 1 or mapheaderinfo[gamemap].bonustype == 2) then
        SRB2CE.DebugPrint(nil, "no bosses allowed")
        return false
    end

    if effecter.amplifyonly and not SRB2CE.AmplifyChecker(player) then
        SRB2CE.DebugPrint(nil, "amplify mode isn't on, not rolling")
        return false
    end

    return true
end

local function CheckNonConflict(player, effectid, ignoreotherincompats, ignoreglobal)
    if not ignoreotherincompats then
        ignoreotherincompats = 0
    end
    local effecter = KhaosEffectsTable[effectid]
    -- check if effect exists first
    SRB2CE.DebugPrint(player, "checking nonconflict: " + effecter.name)

    for key, value in pairs(player.khaoseffects) do
        if value.effect == effectid then
            SRB2CE.DebugPrint(player, "effect is already applied to player, skip")
            return false
        end
        if KhaosEffectsTable[value.effect].incompatible and ignoreotherincompats <= 1 then -- check pre-existing effects on the player, to see if they're incompatible
            local incompatibleTable = KhaosEffectsTable[value.effect].incompatible
            for key2, value2 in pairs(incompatibleTable) do
                --print(value2)
                if value2 == effecter.name then
                    SRB2CE.DebugPrint(player, "incompatible effect found")
                    return false
                end
            end
        end
        if effecter.incompatible and ignoreotherincompats <= 1 then -- now, do the inverse; check this effect's incompatibilities, and see if they exist in the player
            for key2, value2 in pairs(effecter.incompatible) do
                if value2 == KhaosEffectsTable[value.effect].name then
                    SRB2CE.DebugPrint(player, "incompatible effect found second-way around")
                    return false
                end
            end
        end
    end

    for key, value in pairs(player.khaossp.khaosblacklist) do
        if key == effecter.name then
            SRB2CE.DebugPrint(player, "effect is blacklisted!")
            return false
        end
    end

    if effecter.savereq and (not player.khaossp or not player.khaossp[effecter.savereq]) then
        SRB2CE.DebugPrint(player, "player hasn't unlocked effect yet!")
        return false
    end


    if ignoreotherincompats <= 2 then
        if not ignoreglobal then
            if not GlobalNonConflict(effectid) then
                SRB2CE.DebugPrint(player, "effect doesn't pass globalnonconflict")
                return false
            end
        end

        if effecter.nobot and player.bot then
            SRB2CE.DebugPrint(player, "player is a bot, ergo invalid")
            return false
        end

        if effecter.fxrequired and #player.khaoseffects < effecter.fxrequired then
            SRB2CE.DebugPrint(player, "not enough effects. need: " .. effecter.fxrequired)
            return false
        end
    end


    if not ignoreotherincompats then -- put misc. incompatibilities here
        if effecter.requirerings and player.rings <= 0 then
            SRB2CE.DebugPrint(player, "player doesn't have rings!")
            return false
        end
        if effecter.norock and player.powers[pw_carry] == CR_ROLLOUT then
            SRB2CE.DebugPrint(player, "player is on The Rock. skip")
            return false
        end
        if effecter.nocart and player.powers[pw_carry] == CR_MINECART then
            SRB2CE.DebugPrint(player, "player is playing minecraft, skip")
            return false
        end

        if effecter.grounded and not P_IsObjectOnGround(player.mo) then
            SRB2CE.DebugPrint(player, "player isn't on the ground!")
            return false
        end
        if player.khaosoptions.SRB2CE_Memory > 0 then
            for key, value in pairs(player.khaosmemory) do
                if value == effectid then -- recently rolled, don't apply this again for a little bit
                    SRB2CE.DebugPrint(player, "found in memory: " .. value)
                    return false
                end
            end
        end
    end


    -- all else checks out, let's go
    return true
end

SRB2CE.GlobalNonConflict = GlobalNonConflict
SRB2CE.CheckNonConflict = CheckNonConflict

local function ApplyEffect(player, effect, extravalues, hudnumber)
    local dex = tonumber(effect) or indexOfName(KhaosEffectsTable, effect) -- dex 2: enter the decko. no but, get the index so we can point to the right table.
    local effecter = KhaosEffectsTable[dex]
    if effecter == nil then
        assert(nil, "ERROR: Effect not found!")
    end
    if player.khaosoptions.SRB2CE_HUD or effecter.forcestart then
        if effecter.startsfx then
            if effecter.localsfx then
                S_StartSound(player.mo, effecter.startsfx, player)
            else
                S_StartSound(player.mo, effecter.startsfx)
            end
        end
    end

    local funnyduration = 0

    if effecter.duration > 0 and effecter.range then
        local mintime = max(effecter.duration - (effecter.range/3), effecter.duration/2 + 1) -- done to prevent effect from going under half it's time
        local maxtime = effecter.duration + (effecter.range + 2)


        funnyduration = P_RandomRange(mintime, maxtime)

        if SRB2CE.AmplifyChecker(player) and effecter.addition then -- just for a lil more spice on top of the extended timers
            funnyduration = $ + effecter.addition
        end
    else
        funnyduration = effecter.duration
    end

    SRB2CE.DebugPrint(player, effect + ": " + dex)

    local snable = {timer = 0, effect = dex, duration = funnyduration, icon = effecter.sprite, noblink = effecter.noblink, notimer = effecter.notimer, removeme = 0, hudnumber = hudnumber or player.khaosHUDNumber} --hide = effecter.hide

    player.khaosHUDNumber = $ + 1 -- used to assign unique numbers to HUD elements, to keep track of what is what.
    --TODO: this number goes up even if the hudnumber is fed to it from the roulette. is that okay? i think it is?

    if extravalues then
        for key, value in pairs(extravalues) do
            snable[key] = value
        end
    end

    if not effecter.dontrememberme then
        local memsetting = player.khaosoptions.SRB2CE_Memory
        if memsetting > 0 then
            table.insert(player.khaosmemory, dex)
            if #player.khaosmemory > memsetting
                table.remove(player.khaosmemory, 1)
            end
        end
    end

    table.insert(player.khaoseffects, snable)

    if not effecter.nohudtext then
        local prefix = effecter.hudtextprefix or ""
        local color = SKINCOLOR_GREY
        if effecter.etype == "good" then
            color = SKINCOLOR_GREEN
        elseif effecter.etype == "bad" then
            color = SKINCOLOR_RED
        end
        local TheName = effecter.name
        if SRB2CE.AmplifyChecker(player) and effecter.amplifyname then
            TheName = effecter.amplifyname
            SRB2CE.DebugPrint(player, "Amplify Name Detected")
        end
        SRB2CE.addHUDText(player, prefix + TheName, nil, nil, color)
        if player.khaosoptions.SRB2CE_HUD and player.khaosoptions.SRB2CE_Print_Effect then
            CONS_Printf(player, "*\x83 Applied\x80: " .. effecter.name)
        end
    end

    --[[local snable2 = {timer = 0, effect = dex, position = #player.khaoshud, removeme = 0} --hide = effecter.hide
    table.insert(player.khaoshud, snable2)]]
    --play sound if it exists
end

SRB2CE.ApplyEffect = ApplyEffect

SRB2CE.TextBoxTable = { -- there's one more text box stapled to the very end of this lua script, just to count all the current effects lol
{{"Finally caught a",
"Simpleflips stream. He",
"spent 2 hours bouncing",
"on the floor as a"},
{"tree, then he fell",
"asleep live on stream.",
"Great entertainer he is.", ""}},
{{"The path is narrow here.",
"Easy does it! No one is",
"allowed on the top of the",
"mountain!"},
{"And if you know what's",
"good for you, you won't",
"wake anyone who's",
"sleeping!"},
{"Move slowly,",
"tread lightly.","",""}},
{{khaos_rainbowtext("it's look who it is"),"","",""}},
{{"You just lost The Game.","","","hah"}},
{{"<<" + "\x87" + "SRB2-CR4CK (LEGIT).exe" + "\x80" + ">>",
"The application isn't respond-",
"ing. The program may respond",
"again if you wait."},
{"Do you want to end",
"this process?",
"              [End] [Cancel]", ""}},
{{"welcom 2 my EPIC",
"minecraf base. DO NOT",
"TAKE MY DIAMONDS. THAT",
"MEANS U, JARED!!!!1!!"}},
{{"I had to change the",
"neon sign outside my",
"house from saying",
"..."},
{"\"I've always",
"had radishes\", to",
"\"I've never had",
"radishes\"."}},
{{"Busted! Y'all ready?",
"\"Rap Rat\" is the name,",
"The Video Boab Oyedivuth",
"The Video Boab Oyedivuth"},
{"The Video Boab Oyedivuth",
"The Video Boab Oyedivuth",
khaos_rainbowtext("The Video Boab Oyedivuth"), -- just to be silly
"The Video Boab Oyedivuth"},
{"The Video Boab Oyedivuth",
"The Video Boab Oyedivuth",
"The Video Boab Oyedivuth",
"The Video Boab Oyedivuth"}},
{{"Did you know? You can",
"use \x87\"SRB2CE_What_Is\"\x80 in the",
"console to find out more",
"about certain effects!"}},
{{"May I recommend trying",
"the Deaths Counter addon?",
"It'll give you infinite",
"lives to try Chaos with."}},
}

addEffect{name = "i taketh", chance = FRACUNIT/20, duration = 0, sprite = "ITAKETH", etype = "bad", startsfx = sfx_shldls, forcestart = true,
description = "This takes away various things from the player.",
func = function(player, selfid, value)
    player.powers[pw_super] = 0
    P_RemoveShield(player)
    if player.rings > 0 then
        player.powers[pw_invulnerability] = 0
    end
    player.powers[pw_sneakers] = 0
    player.powers[pw_flashing] = 0
    player.powers[pw_tailsfly] = 0
    player.powers[pw_shield] = SH_NONE
    player.powers[pw_nights_superloop] = 0
    player.powers[pw_nights_helper] = 0
    player.powers[pw_nights_linkfreeze] = 0
    if player.bot == 0 and player.rings > 0 then
        if SRB2CE.AmplifyChecker(player) then
            P_GivePlayerRings(player, -7)
        else
            P_GivePlayerRings(player, -3)
        end
    end
    player.glidetime = 0

    player.khaosmemory = {}

    if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
        if player.drillmeter > 0 then
            player.drillmeter = $ / 2
        elseif player.chaotic.OGdrillmeter > 0
            player.chaotic.OGdrillmeter = $ / 2
        end
        if player.spheres > 2 then
            player.spheres = $ - 2
        end
    end

    player.climbing = 0
    player.homing = 0
    --player.mo.color = player.skincolor
    player.pflags = $1 | PF_THOKKED
    player.pflags = $ & ~PF_GLIDING
    player.pflags = $ & ~PF_SPINNING
    P_RestoreMusic(player)
    player.khaos.Check4Hit = max($, 20)
    value.fxdone = true
end}

addEffect{name = "PRESS X NOW", ignoremultiplier = true, nonet = true, nohudtext = true, keeprolling = true, startexception = true, sprite = "NOTHING",
chance = FRACUNIT/39, duration = 2, etype = "bad", description = "Press this button and/or suffer the consequences! ...or don't!",
func = function(player, selfid, value)
    if not value.started then
        local funnybutton = false
        local buttonlist = {
            GC_FORWARD,
            GC_BACKWARD,
            GC_STRAFELEFT,
            GC_STRAFERIGHT,
            GC_TURNLEFT,
            GC_TURNRIGHT,
            GC_WEAPONNEXT,
            GC_WEAPONPREV,
            GC_WEPSLOT1,
            GC_WEPSLOT2,
            GC_WEPSLOT3,
            GC_WEPSLOT4,
            GC_WEPSLOT5,
            GC_WEPSLOT6,
            GC_WEPSLOT7,
            GC_FIRE,
            GC_FIRENORMAL,
            GC_TOSSFLAG,
            GC_SPIN,
            GC_CAMRESET,
            GC_LOOKUP,
            GC_LOOKDOWN,
            GC_TALKKEY, -- make singleplayer only? but only if this can't detect the chat key in multiplayer
            GC_TEAMKEY, -- ditto
            GC_SCORES,
            GC_JUMP,
            GC_CUSTOM1,
            GC_CUSTOM2,
            GC_CUSTOM3,
            GC_FIRE,
            GC_FIRENORMAL,
            GC_VIEWPOINT
        }
        repeat
            local randombutton = P_RandomRange(1, #buttonlist)
            randombutton = buttonlist[randombutton]

            SRB2CE.DebugPrint(player, input.gameControlDown(randombutton))
            SRB2CE.DebugPrint(player, randombutton)
            SRB2CE.DebugPrint(player, input.keyNumToName(input.gameControlToKeyNum(randombutton))) -- GC_*
            if input.keyNumToName(input.gameControlToKeyNum(randombutton)) != "KEY0" then -- if it's "KEY0" that means it's not bound, try again!
                funnybutton = true
            end
            value.funnybutton = randombutton
        until (funnybutton) --if you have all your buttons unbounded, this could result in an infinite loop. that's really funny honestly

        --print(input.gameControlToKeyNum(player.chaotic.funnybutton))

        if (not SRB2CE.AmplifyChecker() and P_RandomChance(FRACUNIT/5)) or (SRB2CE.AmplifyChecker() and P_RandomChance(FRACUNIT/4 + FRACUNIT/9)) then
        -- random chance for this to be flipped to make it so you *shouldn't* press the button
            local phrase = P_RandomRange(1, 3)
            local funnybuttontext = input.keyNumToName(input.gameControlToKeyNum(value.funnybutton))
            if phrase == 1 then
                SRB2CE.addHUDText(player, "PRESS " .. funnybuttontext .. " NOW, OR PROSPER!", true)
            elseif phrase == 2 then
                SRB2CE.addHUDText(player, "Press " .. funnybuttontext .. " so you get hit!", true)
            else
                SRB2CE.addHUDText(player, "Push " .. funnybuttontext .. " to get hit", true)
            end
            value.theopposite = true
        else -- regular mode
            local phrase = P_RandomRange(1, 3)
            local funnybuttontext = input.keyNumToName(input.gameControlToKeyNum(value.funnybutton))
            if phrase == 1 then
                SRB2CE.addHUDText(player, "PRESS " .. funnybuttontext .. " NOW, OR SUFFER!", true)
            elseif phrase == 2 then
                SRB2CE.addHUDText(player, "Press " .. funnybuttontext .. " or you get hit!", true)
            else
                SRB2CE.addHUDText(player, "Push " .. funnybuttontext .. " or get hit", true)
            end
        end
        value.started = true
    end
    if input.gameControlDown(value.funnybutton) then
        value.hitbutton = true
        value.timer = 690*TICRATE -- nice
    end
end,
endfunc = function(player, selfid, value)
    if (value.hitbutton and value.theopposite) or (not value.hitbutton and not value.theopposite) then
        player.powers[pw_invulnerability] = 0
        player.powers[pw_flashing] = 0
        P_DamageMobj(player.mo, nil, nil, DMG_SPIKE)
        S_StartSound(player.mo, sfx_s3kb2, player)
        player.khaos.Check4Hit = max($, 1)
        SRB2CE.addHUDText(player, "Should've listened!", true)
    else
        S_StartSound(player.mo, sfx_ideya, player)
    end
end
}



addEffect{name = "Ghost", chance = FRACUNIT/29, duration = 8, range = 4, sprite = "GHOST", etype = "bad", -- CEICT0
startsfx = sfx_boo, endsfx = sfx_kc65, forcestart = true,
description = "You died?!?! You now phase through all objects, and enemies that don't ",
description2 = "already know you exist can't see you.",
--incompatible = {"Invisible"},
func = function(player, selfid, value)
    if SUBVERSION < 14 and VERSION == 202 then
        player.khaos.GhostNow = true
    else
        player.mo.alpha = max(FRACUNIT/3, $ - (FRACUNIT/35))
    end

    player.pflags = $1 | PF_INVIS
    player.mo.flags = $ | MF_NOCLIPTHING
end,
endfunc = function(player, selfid, value)
    player.mo.flags = $ & ~MF_NOCLIPTHING
    player.pflags = $ & ~PF_INVIS
    player.powers[pw_flashing] = 1*TICRATE
    if SUBVERSION < 14 and VERSION == 202 then
        player.khaos.GhostNow = false
        player.khaos.GhostBegone = true
    else
        player.mo.alpha = FRACUNIT
    end
end
}

local P_GivePlayerSpheres = P_GivePlayerSpheres or function(player, amount) -- 2.2.14 introduces this, might handle things better. or might literally just do this, who knows
    player.spheres = $ + amount
end

addEffect{name = "Money Bag", chance = FRACUNIT/31, duration = 4, range = 2, etype = "good", sprite = "MONEYBAG", nonights = true, -- CEICN0
savereq = "fxpax1", hidename = true,
startsfx = sfx_s3kb3, endsfx = sfx_chchng,
incompatible = {"Bleeding", "Super Explosive 64"},
description = "This will give you around 7 rings every second!",
func = function(player, selfid, value)
    if not value.started then
        value.timer = $ - P_RandomRange(0, TICRATE+(TICRATE)/2) -- approximately 17 tics, or half a second
        if G_IsSpecialStage(gamemap) or (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
            value.duration = $ - 2
        end
        value.started = true
    end

    local amount = 1
    if SRB2CE.AmplifyChecker(player) then
        amount = P_RandomRange(1, 2)
    end

    if (value.timer % 5 == 0) then
        if G_IsSpecialStage(gamemap) or (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
            P_GivePlayerSpheres(player, amount)
            if player.khaosoptions.SRB2CE_HUD then
                S_StartSound(player.mo, sfx_s3k65)
            end
        else
            P_GivePlayerRings(player, amount)
            if player.khaosoptions.SRB2CE_HUD then
                S_StartSound(player.mo, sfx_itemup)
            end
            if SRB2CE.IsMatchMode() then
                local WEPON = P_RandomRange(pw_infinityring, pw_railring)
                player.powers[WEPON] = $ + amount
            end
        end
    end
end}

addEffect{name = "Super Explosive 64", chance = FRACUNIT/39, duration = 8, range = 5, etype = "bad", sprite = "EXPLODE", nospecialstage = true, nonights = true, noghost = true, drainrespawntimer = true,
startsfx = sfx_kc59, endsfx = sfx_kc5a, localsfx = true,
incompatible = {"Lucky!", "Nuke", "Superring", "Ghost", "Bleeding", "Money Bag"},
description = "\x84\"The all-time classic!\"\x80 Bumping into walls will make you explode! Some of the ",
description2 = "rings you collect and enemies you kill are also explosive!",
func = function(player, selfid, value)
    if not value.started then
        value.prevrings = player.rings
        value.prevscoreadd = player.scoreadd
        value.started = true
    end

    if (player.rings > value.prevrings and P_RandomChance(FRACUNIT/15))
    or (player.scoreadd > value.prevscoreadd and P_RandomChance(FRACUNIT/6))
    or player.khaos.wallblock == TICRATE-1
    then
        player.powers[pw_flashing] = 0 -- evil
        player.khaos.Check4Hit = max($, 1)
        local Explode = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_BOSSEXPLODE)
        S_StartSound(Explode, sfx_bkpoof)
        Explode.destscale = 4*player.mo.scale
        Explode.scalespeed = player.mo.scale/6
        P_DamageMobj(player.mo)
        P_RadiusAttack(player.mo, player.mo, 128*FRACUNIT, nil, false)
    end
    value.prevrings = player.rings
    value.prevscoreadd = player.scoreadd
end}

addEffect{name = "Effect Storage", chance = 1/FRACUNIT, duration = 0, etype = "bees", dontrollme = true, nobees = true, nohudtext = true, dontcountme = true, notimer = true, noblacklist = true, dontrememberme = true, startexception = true,
description = "When certain events like \"You Get The Bees\" wants to apply an effect, but you have",
description2 = "an incompatible effect applied, it will be stored here, until it can be given to you.",
func = function(player, selfid, value)
    if not value.icon then
        value.icon = KhaosEffectsTable[value.storeeffect].sprite
    end
    if value.timer > TICRATE and (leveltime % 10) == 0 then
        if CheckNonConflict(player, value.storeeffect, 1, true) then
            value.fxdone = true
        end
    end
end,
endfunc = function(player, selfid, value)
    ApplyEffect(player, KhaosEffectsTable[value.storeeffect].name, nil) -- value.hudnumber used to be placed here, but that seems to be a bit buggy?
end}

addEffect{name = "Rewind", chance = FRACUNIT/30, duration = 0, etype = "bad", sprite = "REWIND", startsfx = sfx_chrewi, forcestart = true, nonights = true, startexception = true,
description = "Every 3 seconds a new position is recorded down, and when you get rewind, it ",
description2 = "will send you back there!",
func = function(player, selfid, value)
    P_SetOrigin(player.mo, player.khaos.RewindX, player.khaos.RewindY, player.khaos.RewindZ)
    if not SRB2CE.AmplifyChecker(player) then
        player.mo.angle = player.khaos.RewindAngle
        COM_BufInsertText(consoleplayer, "resetcamera") -- what no access to the camera variable does to a mfer
    end
    player.khaos.Check4Hit = max($, 20)
    value.fxdone = true
end}

addEffect{name = "I WAS FROZEN TODAY!", chance = FRACUNIT/26, duration = 2, range = 1, sprite = "ICE", addition = 1, etype = "bad", drainrespawntimer = true, nonights = true,
startsfx = sfx_s3k80, endsfx = sfx_shattr, forcestart = true, forceend = true,
incompatible = {"Change Color", "Cement Shoes", "Change Sprite", "Black Hole", "Laggy Model", "Hey! Listen!"},
description = "Your controls are frozen! Shake wildly to speed up the timer!",
func = function(player, selfid, value)
    if not value.started then
        player.khaos.FrozenNow = true
        player.khaos.FrozenWait = 3
        value.FrozenPrevColor = player.mo.color
        value.started = true
    end

    player.mo.colorized = true
    player.mo.color = SKINCOLOR_ICY
    player.skincolor = player.mo.color

    player.pflags = $|PF_FULLSTASIS

    if not player.khaos.FrozenState then
        player.khaos.FrozenState = player.mo.state
        player.khaos.FrozenFrame = player.mo.frame
        if player.followmobj then
            player.khaos.FollowFrozenState = player.followmobj.state
            player.khaos.FollowFrozenFrame = player.followmobj.frame
        end
    end

    if player.bot and P_RandomChance(FRACUNIT/7) then
        player.khaos.FrozenShake = true
    end

    if player.khaos.FrozenShake and player.khaos.FrozenWait == 0 then
        if VERSION >= 2 then
            S_StartSound(player.mo, sfx_iceb)
        else
            S_StartSound(player.mo, sfx_cdfm41)
        end
        local z = (player.mo.z + player.mo.height) + 6*player.mo.scale
        if (player.mo.eflags & MFE_VERTICALFLIP) then
            z = (player.mo.z - player.mo.height) - 6*player.mo.scale
        end
        local piecespawn = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
        piecespawn.number = P_RandomRange(1, 4)
        piecespawn.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
        value.timer = $ + P_RandomRange(3, 6)
        player.khaos.FrozenWait = 4
    end

    if player.khaos.FrozenWait > 0 then
        player.khaos.FrozenWait = $ - 1
        player.mo.spritexoffset = 0
        player.mo.spriteyoffset = 0
    end

    if player.khaos.FrozenWait == 0 then
        player.khaos.FrozenShake = false
    end
    player.khaos.Check4Hit = max($, 1)
end,
endfunc = function(player, selfid, value)
    if value.FrozenPrevColor then -- seems to complain sometimes, so, we'll make sure this exists first
        player.mo.color = value.FrozenPrevColor
        player.skincolor = player.mo.color
    end
    player.khaos.Check4Hit = max($, 8)
    player.mo.colorized = false
    player.khaos.FrozenNow = false
    player.khaos.FrozenShake = false
    player.khaos.FrozenState = nil
    player.khaos.FollowFrozenState = nil
    player.khaos.FrozenFrame = nil
    player.khaos.FollowFrozenFrame = nil
    player.mo.spritexoffset = 0
    player.mo.spriteyoffset = 0

    local z = (player.mo.z + player.mo.height) + 6*player.mo.scale
    if (player.mo.eflags & MFE_VERTICALFLIP) then
        z = (player.mo.z - player.mo.height) - 6*player.mo.scale
    end

    -- stolen from ice springs, forgive me for this sloppy code, i refuse to fix it just yet
    local piece1 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece2 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece3 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece4 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece5 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece6 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece7 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    local piece8 = P_SpawnMobj(player.mo.x, player.mo.y, z, MT_ICESHRAPNELCHAOTIC)
    piece1.number = 1
    piece2.number = 2
    piece3.number = 3
    piece4.number = 4
    piece1.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece2.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece3.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece4.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))

    piece5.number = 1
    piece6.number = 2
    piece7.number = 3
    piece8.number = 4
    piece5.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece6.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece7.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
    piece8.angle = player.mo.angle + (ANG1 * (P_RandomRange(-360, 360)))
end
}

addEffect{name = "Giant", chance = FRACUNIT/10, duration = 7, sprite = "GIANT", range = 3, etype = "neutral", restartrespawn = true,
startsfx = sfx_mario3, endsfx = sfx_kc59, forcestart = true, forceend = true,
incompatible = {"Tiny", "HE"},
description = "This makes you giant!",
func = function(player, selfid, value)
    if not value.started then
        value.setscale = (mapobjectscale*3)
        player.mo.destscale = value.setscale
        value.started = true
    end
end,
endfunc = function(player, selfid, value)
    player.mo.destscale = mapobjectscale --TODO: check if there's other circumstances where player size changes
end}

addEffect{name = "Tiny", chance = FRACUNIT/13, duration = 6, sprite = "TINY", range = 2, addition = 2, etype = "bad", restartrespawn = true,
startsfx = sfx_mario8, endsfx = sfx_kc5a, forcestart = true, forceend = true,
incompatible = {"Giant"},
description = "This makes you very small!",
func = function(player, selfid, value)
    if not value.started then
        value.setscale = (mapobjectscale/4) + (mapobjectscale/3) --TODO: determine if this new size is better than before
        if SRB2CE.AmplifyChecker(player) then
            value.destscale = (mapobjectscale/4)
        end
        player.mo.destscale = value.setscale
        value.started = true
    end
end,
endfunc = function(player, selfid, value)
    player.mo.destscale = mapobjectscale
end}

addEffect{name = "Super Jump", chance = FRACUNIT/30, duration = 8, range = 8, sprite = "HOP", addition = 1, nocart = true, nonights = true, etype = "neutral",
startsfx = sfx_kc5a, endsfx = sfx_kc59, localsfx = true,
incompatible = {"Gravity Coil", "Bouncy!", "Moonjump", "Pancake"},
description = "Your jump got super-charged, making you go 4-5 times higher!",
func = function(player, selfid, value)
    if SRB2CE.AmplifyChecker() then
        player.jumpfactor = skins[player.mo.skin].jumpfactor * 5
    else
        player.jumpfactor = skins[player.mo.skin].jumpfactor * 4
    end
end,
endfunc = function(player, selfid, value)
    player.jumpfactor = skins[player.mo.skin].jumpfactor
end}

addEffect{name = "Omochao", chance = FRACUNIT/12, duration = 15, range = 4, sprite = "OMO", etype = "bad", addition = 7, -- fucking suffer
description = "Am I annoying you?",
savereq = "hunter", hidename = true,
func = function(player, selfid, value)
    if not value.started then
        value.omochao = {}
        value.omochao[1] = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_CHAOSOMO)
        value.omochao[1].target = player.mo
        value.randomtimer = P_RandomRange(10, 3*TICRATE)
        value.countdown = 3
        value.started = true
    end

    if SRB2CE.AmplifyChecker(player) and (value.timer % (value.randomtimer) == 0) then
        value.randomtimer = P_RandomRange(2*TICRATE, 5*TICRATE)
        value.omochao[#value.omochao+1] = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_CHAOSOMO)
        value.omochao[#value.omochao].target = player.mo
    end

end,
endfunc = function(player, selfid, value)
    for key, v2 in pairs(value.omochao) do
        if v2 and v2.valid then
            v2.killtimer = value.countdown
            value.countdown = $ + 3
            --S_StartSound(v2, sfx_pop)
            --P_KillMobj(v2)
        end
    end
end}

addEffect{name = "Gravity Coil", chance = FRACUNIT/16, duration = 15, range = 4, sprite = "GRAVITYCOIL", addition = 3, nonights = true, etype = "good", norock = true,
savereq = "fxpax1",
startsfx = sfx_chgc, endsfx = sfx_kc65, localsfx = true,
incomaptible = {"Super Jump", "Bouncy!", "Cement Shoes"},
description = "This makes your falls descend a whole lot slower.",
func = function(player, selfid, value)

    if skins[player.mo.skin].jumpfactor >= FRACUNIT then
        player.jumpfactor = (skins[player.mo.skin].jumpfactor + (FRACUNIT/2) + (FRACUNIT/4))
    else
        player.jumpfactor = (FRACUNIT + (FRACUNIT/2) + (FRACUNIT/4))
    end

    if not P_IsObjectOnGround(player.mo) then
        if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
            if player.mo.momz > 0 then
                player.mo.momz = $ - $ / 25
            elseif player.mo.momz < (9*FRACUNIT * FRACUNIT) then
                player.mo.momz = $ + $ / 2
            end
        else
            if player.mo.momz < 0 then
                player.mo.momz = $ - $ / 25
            elseif player.mo.momz < (9*FRACUNIT * FRACUNIT) then
                player.mo.momz = $ + $ / 2
            end
        end
    end
end,
endfunc = function(player, selfid, value)
    player.jumpfactor = skins[player.mo.skin].jumpfactor
end}

addEffect{name = "Sticky Slopes", chance = FRACUNIT/26, duration = 10, range = 5, sprite = "STICKY", addition = 5, etype = "bad", nonights = true, -- CEIDJ0
startsfx = sfx_kc59, endsfx = sfx_kc5a, localsfx = true,
incompatible = {"Pancake", "Butter Physics", "Cement Shoes"},
description = "You'll have a hard time going up hill! And it's also hard going down hill!",
func = function(player, selfid, value)
    if P_IsObjectOnGround(player.mo) and player.mo.standingslope then
        --player.mo.friction = FRACUNIT - FRACUNIT/10
        player.mo.friction = 23*FRACUNIT/32
    else
        if player.khaos.OGFriction then -- srb2kart complained about this when a player was respawning
            player.mo.friction = player.khaos.OGFriction
        end
    end
end,
endfunc = function(player, selfid, value)
    if player.khaos.OGFriction then -- srb2kart complained about this when a player was respawning
        player.mo.friction = player.khaos.OGFriction
    end
end}

addEffect{name = "Drifting", chance = FRACUNIT/22, duration = 10, range = 3, sprite = "DRIFTINGLEFT", addition = 1, etype = "bad", nocart = true, nohudtext = true, nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Invert Controls"},
description = "Simulates Joy-Con(tm) Drift! You will be constantly veering either to ",
description2 = "the left or right.",
func = function(player, selfid, value)
    --player.cmd.turning -- ranges from -800 to 800, positive is left

    if not value.started then
        value.started = true
        value.adjustvalue = 450
        if P_RandomChance(FRACUNIT/2) then
            value.icon = "KH_DRIFTINGRIGHT"
            value.adjustvalue = -450
            value.adjustangle = -ANG2
            SRB2CE.addHUDText(player, "Drifting Right")
        else
            value.adjustangle = ANG2
            SRB2CE.addHUDText(player, "Drifting Left")
        end
    end

    if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
        player.flyangle = $ + value.adjustvalue
        player.anotherflyangle = $ + value.adjustvalue
    else
        if SRB2CE.AmplifyChecker(player) then
            player.mo.angle = (player.cmd.angleturn << FRACBITS) + value.adjustangle + value.adjustangle
        else
            player.mo.angle = (player.cmd.angleturn << FRACBITS) + value.adjustangle
        end
    end
end
}

addEffect{name = "Earthquake", chance = FRACUNIT/16, duration = 12, range = 8, etype = "bad", sprite = "EARTHQUAKE", hudtextprefix = "\202",
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
description = "Causes your screen to shake violently! If you're near anyone else, their screen ",
description2 = "will also be shaking!",
func = function(player, selfid, value)
    --if (leveltime % 10 == 0) then
        if SRB2CE.AmplifyChecker(player) then
            P_StartQuake(30*FRACUNIT, 3, {player.mo.x, player.mo.y, player.mo.z}, 700*FRACUNIT) -- the latter 2 don't do anything pre-v2.2.14
        else
            P_StartQuake(20*FRACUNIT, 3, {player.mo.x, player.mo.y, player.mo.z}, 700*FRACUNIT)
        end
    --end
end
}

addEffect{name = "Nuke", chance = FRACUNIT/40, duration = 0, sprite = "NUKE", nonights = true, etype = "good", startsfx = sfx_bkpoof, nospecialstage = true, forcestart = true,
savereq = "fxpax2",
description = "Sets off an explosion, the same explosion used by the Armageddon Shield.",
func = function(player, selfid, value)

    P_NukeEnemies(player.mo, player.mo, (2160*2)*FRACUNIT)

    --[[if value.NukePrevShield then
        player.powers[pw_shield] = value.NukePrevShield
    end

    --P_BlackOw(player)

    if player.powers[pw_shield] then
        value.NukePrevShield = player.powers[pw_shield]
    end]]

    --[[local JankSplosion = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_SSMINE)
    JankSplosion.scale = mapobjectscale * 5
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    local JankSplosion = P_SpawnMobj(player.mo.x+50*FRACUNIT, player.mo.y, player.mo.z, MT_SSMINE)
    JankSplosion.scale = mapobjectscale * 3
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    local JankSplosion = P_SpawnMobj(player.mo.x, player.mo.y+50*FRACUNIT, player.mo.z, MT_SSMINE)
    JankSplosion.scale = mapobjectscale * 3
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    local JankSplosion = P_SpawnMobj(player.mo.x-50*FRACUNIT, player.mo.y, player.mo.z, MT_SSMINE)
    JankSplosion.scale = mapobjectscale * 3
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    local JankSplosion = P_SpawnMobj(player.mo.x, player.mo.y-50*FRACUNIT, player.mo.z, MT_SSMINE)
    JankSplosion.scale = mapobjectscale * 3
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    -- this cause a shit ton of lag. i am not fixing this.
    player.khaos.DodgeHit = false]]

    value.fxdone = true
end}


addEffect{name = "Reversed!", chance = FRACUNIT/33, duration = 0, etype = "bad", sprite = "REVERSED", amplifyonly = true, startsfx = sfx_s3kaa, nonights = true,
description = "This instantly turns you around 180 degrees!",
savereq = "amplifier",
func = function(player, selfid, value)
    player.mo.angle = player.mo.angle + ANGLE_180
    --if SRB2CE.AmplifyChecker(player) then
        player.mo.momx = -player.mo.momx
        player.mo.momy = -player.mo.momy
    --end
    player.khaos.Check4Hit = max($, 20)
    value.fxdone = true
end}

addEffect{name = "the pizza is aggressive", chance = FRACUNIT/37, duration = 13, range = 12, sprite = "PIZZA", addition = 3, etype = "bad", noghost = true, nonights = true,
description = "\x84\"Bond with me, Jimmy.\"\x80 You ordered a pizza, but it's aggressive! This will chase ",
description2 = "you until it hits you, or the timer runs out.",
startsfx = sfx_chapis, endsfx = sfx_kc65, forcestart = true,
incompatible = {"Ghost"},
func = function(player, selfid, value)
    if not player.khaos.PizzaSpawned then
        GivePizza(player)
        player.khaos.PizzaGivenAlready = true
    end
    if player.khaos.PizzaHitTarget then
        if player.khaos.PizzaHitTarget then
            SRB2CE.addHUDText(player, "Delivery complete, enjoy your pizza!", true)
            player.khaos.Check4Hit = max($, 1)
            value.timer = 69*TICRATE -- nice
        end
    end
end,
endfunc = function(player, selfid, value)
    player.khaos.PizzaSpawned = false
    player.khaos.PizzaRemoval = true
    player.khaos.PizzaHitTarget = false
    player.khaos.PizzaGivenAlready = false
end}

--TODO: maybe remix this to work w/ SRB2?
--[[  addEffect{name = "Insta-Boost", chance = FRACUNIT/17, duration = 0, sprite = "BOOST", etype = "neutral",
description = "\x84\"You got boost power!\"\x80 This instantly applies a sneaker boost to the player.",
func = function(player, selfid, value)
    if VERSION >= 2 then
        player.ringboost = $ + 600
    end
    if P_IsObjectOnGround(player.mo) then
        K_DoSneaker(player, 0)
    else
        K_DoSneaker(player, 2)
    end
    value.fxdone = true
end}]]

addEffect{name = "No Turning", chance = FRACUNIT/38, etype = "bad", duration = 3, sprite = "NOTURNING", addition = 2, drainrespawntimer = true, nonights = true,
description = "Disables your turning very briefly!",
startsfx = sfx_s3kb2, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Drifting"},
func = function(player, selfid, value)
    if not value.started then
        value.pangle = player.mo.angle
        value.started = true
    end
    player.khaos.Check4Hit = max($, 1)
    player.mo.angle = value.pangle
end}

addEffect{name = "Hop!", chance = FRACUNIT/12, duration = 0, sprite = "HOP", startsfx = sfx_sprong, kartstartsfx = sfx_kpogos, etype = "neutral", forcestart = true, norock = true, nocart = true, nonights = true,
description = "Springs you upward with great velocity!",
func = function(player, selfid, value)
    if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
        player.mo.momz = min($, 0) + P_RandomRange(-15, -30)*FRACUNIT
    else
        player.mo.momz = max($, 0) + P_RandomRange(15, 30)*FRACUNIT
    end
    value.fxdone = true
end}

addEffect{name = "Laggy Model", chance = FRACUNIT/31, duration = 12, range = 5, sprite = "LAG", etype = "bad", --freezeremove = true,
startsfx = sfx_kc65, endsfx = sfx_kc65, forcestart = true, forceend = true,
incompatible = {"I WAS FROZEN TODAY!", "Invisible", "Change Sprite"},
description = "Causes your player's sprite/model to lag behind.",
func = function(player, selfid, value)
    if not value.started then
        --value.thej = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_LAGPLAYER)
        --value.thej.gayer = player
        player.khaos.InvisibleNow = true
        player.khaos.InvisibleBegone = false
        value.started = true
        value.divisor = P_RandomRange(3, 7)
    end


    if (leveltime % (TICRATE/value.divisor) == 0) then
        value.divisor = P_RandomRange(3, 7)
        if value.thej and value.thej.valid then
            if value.thej.tracer and value.thej.tracer.valid then
                P_RemoveMobj(value.thej.tracer)
            end
            P_RemoveMobj(value.thej)
        end
        value.thej = P_SpawnGhostMobj(player.mo)
        value.thej.frame = $ & ~FF_TRANSMASK
        --value.thej.frame = $1 | FF_TRANS10
        value.thej.fuse = 2*TICRATE
        value.thej.shadowscale = FRACUNIT
        if value.thej.tracer and value.thej.tracer.valid then
            value.thej.tracer.frame = $ & ~FF_TRANSMASK
            value.thej.tracer.fuse = 2*TICRATE
        end

        --FF_TRANSMASK
        --P_RemoveMobj(value.thej)
        --value.started = false
    end

    --[[if value.thej and value.thej.valid then
        if (player.mo.flags2 & MF2_DONTDRAW) then -- used to get invincibility frames to render
            value.thej.Blinker = true
        else
            value.thej.Blinker = false
        end
    end]]
end,
endfunc = function(player, selfid, value)
    if value.thej and value.thej.valid then
        P_RemoveMobj(value.thej)
    end
    player.khaos.InvisibleNow = false
    player.khaos.InvisibleBegone = true

    local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
    dust.scale = player.mo.scale*2
    dust.state = S_XPLD1
end}

addEffect{name = "Change Sprite", chance = FRACUNIT/10, duration = 21, range = 15, sprite = "CHSPRITE", etype = "neutral",
startsfx = sfx_kc65, endsfx = sfx_kc65, forcestart = true, forceend = true,
incompatible = {"I WAS FROZEN TODAY!", "Invisible", "Laggy Model"},
description = "Changes your player's sprite/model to be completely different!",
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        player.khaos.ChangeSpriteNow = true
        player.khaos.InvisibleNow = true

        local funnystates = {S_CHAOSMAC, S_CHAOSWIN9X, S_CHAOSTUX, S_UNKNOWN, S_YELLOWSPRING, S_YHORIZ1, S_BUSH, S_SOSS_RUN1, S_GFZTREE, S_GARGOYLE, S_COLORRING, S_CHAOTICPIZZA}
        if player.khaossp.drowned then
            table.insert(funnystates, S_CHAOSSKULL)
            table.insert(funnystates, S_CHAOSSKULL) -- do it twice because funny
        end
        local SPRITE = funnystates[P_RandomRange(1, #funnystates)]
        -- S_SSHELL_RUN1, doesn't work :(

        local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
        player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
        dust.scale = player.mo.scale*2
        dust.state = S_XPLD1

        player.khaos.ChangeSpriteMOBJ = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_CHANGESPRITE)
        player.khaos.ChangeSpriteMOBJ.gayer = player
        player.khaos.ChangeSpriteMOBJ.state = SPRITE
        --player.khaos.ChangeSpriteMOBJ.flags = $ & ~MF_ENEMY
        player.khaos.ChangeSpriteMOBJ.flags = $ & ~MF_NOTHINK
        player.khaos.ChangeSpriteMOBJ.summonedasstate = SPRITE
        player.khaos.ChangeSpriteMOBJ.dontdrawforviewmobj = player.mo
    end

    if player.khaos.ChangeSpriteMOBJ and player.khaos.ChangeSpriteMOBJ.valid then
        if (player.mo.flags2 & MF2_DONTDRAW) then -- used to get invincibility frames to render
            player.khaos.ChangeSpriteMOBJ.Blinker = true
        else
            player.khaos.ChangeSpriteMOBJ.Blinker = false
        end
    end
end,
endfunc = function(player, selfid, value)
    player.khaos.ChangeSpriteNow = false
    if player.khaos.ChangeSpriteMOBJ and player.khaos.ChangeSpriteMOBJ.valid then
        player.khaos.ChangeSpriteMOBJ.despawn = true
        player.khaos.ChangeSpriteMOBJ = nil
    end
    player.khaos.InvisibleNow = false
    player.khaos.InvisibleBegone = true

    local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, MT_DUST)
    dust.scale = player.mo.scale*2
    dust.state = S_XPLD1
end}

addEffect{name = "Hey! Listen!", chance = FRACUNIT/34, duration = 12, sprite = "LISTEN", etype = "bad", nobees = true, nobot = true, immune2remove = true, notimer = true,
startsfx = sfx_chtxst, endsfx = sfx_chtxed, forcestart = true, forceend = true,
incompatible = {"Hop!", "Fling Player", "Rewind", "Bouncy!", "Windy Day", "I WAS FROZEN TODAY!"},
description = "Pops up an annoying text box. Press jump to get through the text!",
func = function(player, selfid, value)
    if not value.started then
        player.khaos.TextBoxStation = P_RandomRange(1, #SRB2CE.TextBoxTable)
        player.khaos.TextBoxPage = 1 --#SRB2CE.TextBoxTable[TextBoxStation]
        player.khaos.TextBoxFader = 8
        player.khaos.TextBoxAdvancePage = 30
        player.khaos.FadeTextBoxIn = 1
        player.khaos.prevmomx = player.mo.momx
        player.khaos.prevmomy = player.mo.momy
        player.khaos.prevmomz = player.mo.momz
        player.khaos.keepangle = player.mo.angle
        value.started = true
    end

    if player.khaos.TextBoxFlipPage then
        player.khaos.TextBoxFlipPage = false
        if player.khaos.TextBoxPage + 1 > #SRB2CE.TextBoxTable[player.khaos.TextBoxStation] then
            player.khaos.FadeTextBoxOut = 1
            value.timer = (value.duration * TICRATE) - (3*TICRATE)
        else
            player.khaos.TextBoxPage = $ + 1
        end
    end

    player.mo.angle = player.khaos.keepangle
    if not isPlayerRespawning(player) then
        player.mo.momx = 0
        player.mo.momy = 0
        player.mo.momz = 0
    end
    --player.spinouttype = KSPIN_IFRAMES
    player.powers[pw_flashing] = 2
    player.pflags = $|PF_FULLSTASIS

    if player.khaos.FadeTextBoxIn then
        player.khaos.FadeTextBoxIn = $ + 1
        if ((player.khaos.FadeTextBoxIn-2*TICRATE)*2) % 2 == 0 then
            player.khaos.TextBoxFader = $ - 1
        end
        if player.khaos.TextBoxFader == 0 then
            player.khaos.FadeTextBoxIn = 0
        end
    end

    if not player.khaos.FadeTextBoxOut and value.timer == (value.duration * TICRATE) - (1*TICRATE) then -- fade out text box near timer end, to make it look smoother
        player.khaos.FadeTextBoxOut = 1
    end

    if player.khaos.FadeTextBoxOut then
        player.khaos.FadeTextBoxOut = $ + 1
        if ((player.khaos.FadeTextBoxOut-2*TICRATE)*2) % 2 == 0 then
            player.khaos.TextBoxFader = $ + 1
        end
        if player.khaos.TextBoxFader > 9 then
            player.khaos.TextBoxStation = nil
            player.khaos.FadeTextBoxOut = 0
            player.khaos.TextBoxFader = 0
            value.fxdone = true
        end
    end

    if player.khaos.TextBoxAdvancePage > 0 then
        player.khaos.TextBoxAdvancePage = $ - 1
    end

    if (player.cmd.buttons & BT_JUMP) --TODO: replace with PBT3
        if player.khaos.TextBoxStation and not player.khaos.HoldAccelerate and not player.khaos.TextBoxAdvancePage then
            player.khaos.TextBoxAdvancePage = 20
            player.khaos.TextBoxFlipPage = true
            if player.khaos.TextBoxPage + 1 <= #SRB2CE.TextBoxTable[player.khaos.TextBoxStation] then
                S_StartSound(player.mo, sfx_chtxad, player)
            end
        end
        player.khaos.HoldAccelerate = true
    else
        player.khaos.HoldAccelerate = false
    end

end,
endfunc = function(player, selfid, value)
    player.mo.momx = player.khaos.prevmomx
    player.mo.momy = player.khaos.prevmomy
    player.mo.momz = player.khaos.prevmomz

    player.khaos.TextBoxStation = nil
    player.khaos.FadeTextBoxOut = 0
    player.khaos.TextBoxFader = 0
    player.khaos.Check4Hit = max($, 20)
end}


--TODO: remix this, make it just take control away from the player for a bit, then the sprite re-emerges b4 the effect ends
--[[  addEffect{name = "Quicksand!", chance = FRACUNIT/50, duration = 2, sprite = "QUICKSAND", etype = "bad", nobees = true, grounded = true, drainrespawntimer = true, nosrb2kart = true, nospecialstage = true,
startsfx = sfx_chslrp, forcestart = true,
incompatible = {"Bouncy!", "Windy Day", "Insta-Boost", "Hey! Listen!"},
description = "You got slurped by the ground! You respawn after this effect goes away.",
func = function(player, selfid, value)
    if not value.started then
        value.offsetter = 0
        value.adder = FRACUNIT*2
        player.spinouttype = KSPIN_WIPEOUT
        player.spinouttimer = value.duration*TICRATE
        player.mo.state = S_KART_SPINOUT
        value.started = true
    end

    value.offsetter = $ - value.adder
    value.adder = $ + (FRACUNIT/10)

    player.mo.momx = 0
    player.mo.momy = 0
    player.mo.momz = 0

    -- player.mo.angleroll = $ + ANG1
    player.mo.spriteyoffset = value.offsetter
    player.mo.spritexoffset = 1
end,
endfunc = function(player, selfid, value)
    player.mo.spritexoffset = 0
    player.mo.spriteyoffset = 0
    P_DamageMobj(player.mo, nil, nil, 1, DMG_DEATHPIT)
end}]]

local InvalidBHObjects = {}

if VERSION >= 2 then
    InvalidBHObjects = {MT_PLAYER, MT_STEAM, MT_CHAINPOINT}
else
    InvalidBHObjects = {MT_PLAYER}
end

local function ValidBHObject(mobj)
    if (mobj.flags & (MF_SOLID)) == (MF_SOLID) and not (mobj.type != MT_SPIKE and SRB2CE.AmplifyChecker()) then
        return false -- exception for all things solid
    elseif (mobj.flags & MF_SPRING) --and not tonumber(CV_FindVar("SRB2CE_Amplify").value) then
        return false
    elseif (mobj.flags & MF_PUSHABLE) then
        return false  -- just because those gargoyles are annoying too
    end
    for key, value in pairs(InvalidBHObjects) do
        if value == mobj.type then
            return false
        end
    end

    return true
end

addEffect{name = "Black Hole", chance = FRACUNIT/20, duration = 20, range = 5, sprite = "BLACKHOLE", addition = 2, etype = "bad",  -- CEIC50
nobees = true, dontcountme = true, dontrollme = true, drainrespawntimer = true, immune2remove = true, noghost = true, startsfx = sfx_s3k74, endsfx = sfx_kc65,
description = "You were granted the title of 'Most Attractive Player'! Ergo, nearby objects ",
description2 = "will start flying towards you.",
incompatible = {"Super Explosive 64", "Ghost", "Bleeding"},
func = function(player, selfid, value)
    -- thanks to amperbee for providing sample code!
    local m = player.mo --used to be realmo, but ring racers is silly and thinks we shouldn't have realmo!
    local range = 2160*mapobjectscale
    if SRB2CE.AmplifyChecker(player) then
        range = $ + 512 * mapobjectscale
    end
    local range2 = range - (64*mapobjectscale)
    searchBlockmap("objects", function(ref, found)
        if ValidBHObject(found) then
            if R_PointToDist2(m.x, m.y, found.x, found.y) <= range2 then
                if VERSION >= 2 then
                    if found.type == MT_RING or found.type == MT_BLUESPHERE then
                        found.tracer = m
                    end
                end
                -- min = 3 FRACUNIT
                -- max = 22 FRACUNIT
                local dist = R_PointToDist2(m.x, m.y, found.x, found.y)
                if dist < FRACUNIT then
                    dist = FRACUNIT
                end
                local PullForce = range/dist * 3
                if PullForce > 40 then
                    PullForce = 40
                end
                PullForce = $ * FRACUNIT

                BootlegAttract(found, m, PullForce)
            end
        end
    end, m, m.x-range, m.x+range, m.y-range, m.y+range)
    player.khaos.Check4Hit = max($, 1)
end}

addEffect{name = "HE", chance = FRACUNIT/39, duration = 16, range = 6, sprite = "HE", addition = 3, etype = "neutral", nonights = true, nocart = true, -- our lord and savior, Lanky "Christ" Kong
startsfx = sfx_chlank, endsfx = sfx_kc65, forcestart = true,
incompatible = {"Pancake", "Giant", "Cement Shoes"},
description = "You have no style, you have no grace, t-t-t-this player ",
description2 = "has unusual speed properties!",
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        value.GOALY = FRACUNIT*4 - FRACUNIT/4
        value.GOALX = FRACUNIT/3 + FRACUNIT/4
        value.Defaultkartspeed = player.kartspeed
        value.Defaultkartweight = player.kartweight
        if SRB2CE.AmplifyChecker(player) then
            value.GOALY = FRACUNIT*4 - FRACUNIT/6
            value.GOALX = FRACUNIT/2 + FRACUNIT/4
        end
        player.khaos.RestoreSpriteXScaleUp = nil
        player.khaos.RestoreSpriteYScaleUp = nil
        player.khaos.RestoreSpriteXScaleDown = nil
        player.khaos.RestoreSpriteYScaleDown = nil
    end

    if player.mo.spritexscale != value.GOALX then
        if player.mo.spritexscale < value.GOALX then
            player.mo.spritexscale = $ + FRACUNIT/2
        elseif player.mo.spritexscale > value.GOALX then
            player.mo.spritexscale = value.GOALX
        end
    end
    if player.mo.spriteyscale != value.GOALY then
        if player.mo.spriteyscale < value.GOALY then
            player.mo.spriteyscale = $ + FRACUNIT/2
        elseif player.mo.spriteyscale > value.GOALY then
            player.mo.spriteyscale = value.GOALY
        end
    end

    --[[if tonumber(CV_FindVar("SRB2CE_Amplify").value)  then
        player.normalspeed = skins[player.mo.skin].normalspeed * 3
        player.runspeed = skins[player.mo.skin].runspeed * 3
        player.actionspd = skins[player.mo.skin].actionspd * 3
        player.thrustfactor = skins[player.mo.skin].thrustfactor * 3
        player.mindash = FRACUNIT * 80
        player.maxdash = skins[player.mo.skin].maxdash * 2
    else
        player.normalspeed = (skins[player.mo.skin].normalspeed * 2) + (skins[player.mo.skin].normalspeed / 4)
        player.runspeed = (skins[player.mo.skin].runspeed * 2) + (skins[player.mo.skin].runspeed / 3)
        player.actionspd = skins[player.mo.skin].actionspd * 2
        player.thrustfactor = (skins[player.mo.skin].thrustfactor * 2) + (skins[player.mo.skin].thrustfactor / 3)
    end]]

    if SRB2CE.AmplifyChecker(player) then
        player.normalspeed = skins[player.mo.skin].normalspeed * 3
        player.runspeed = skins[player.mo.skin].runspeed * 3
        player.actionspd = skins[player.mo.skin].actionspd * 3
        player.thrustfactor = skins[player.mo.skin].thrustfactor * 3
        player.mindash = FRACUNIT * 80
        player.maxdash = skins[player.mo.skin].maxdash * 2
    else
        player.normalspeed = (skins[player.mo.skin].normalspeed * 2) + (skins[player.mo.skin].normalspeed / 4)
        player.runspeed = (skins[player.mo.skin].runspeed * 2) + (skins[player.mo.skin].runspeed / 3)
        player.actionspd = skins[player.mo.skin].actionspd * 2
        player.thrustfactor = (skins[player.mo.skin].thrustfactor * 2) + (skins[player.mo.skin].thrustfactor / 3)
    end
end,
endfunc = function(player, selfid, value)
    player.normalspeed = skins[player.mo.skin].normalspeed
    player.runspeed = skins[player.mo.skin].runspeed
    player.actionspd = skins[player.mo.skin].actionspd
    player.thrustfactor = skins[player.mo.skin].thrustfactor
    player.mindash = skins[player.mo.skin].mindash
    player.maxdash = skins[player.mo.skin].maxdash
    player.height = skins[player.mo.skin].height
    player.khaos.RestoreSpriteXScaleDown = true
    player.khaos.RestoreSpriteYScaleDown = true
    --player.mo.spriteyscale = FRACUNIT
end}

addEffect{name = "Rainbow Sneakers", chance = FRACUNIT/30, duration = 16, range = 6, sprite = "RSPEED", addition = 3, etype = "neutral", nonights = true, nocart = true, restartrespawn = true,
startsfx = sfx_kc65, endsfx = sfx_kc65, forcestart = true,
incompatible = {"Pancake", "Giant", "Cement Shoes", "HE"},
description = "Your speed properties have been randomized!",
func = function(player, selfid, value)
    if not value.started then
        if not value.PosOrNeg then
            if P_RandomChance(FRACUNIT/2) then
                value.PosOrNeg = -1
            else
                value.PosOrNeg = 1
            end
        end
        local PosOrNeg = value.PosOrNeg -- too lazy to change all of the below
        player.normalspeed = max($ / 5, skins[player.mo.skin].normalspeed + (PosOrNeg * (skins[player.mo.skin].normalspeed / P_RandomRange(1, 3))))
        player.runspeed = max($ / 4, skins[player.mo.skin].runspeed + (PosOrNeg * (skins[player.mo.skin].runspeed / P_RandomRange(1, 3))))
        player.actionspd = max($ / 3, skins[player.mo.skin].actionspd + (PosOrNeg * (skins[player.mo.skin].actionspd / P_RandomRange(1, 3))))
        player.thrustfactor = max($ / 3, skins[player.mo.skin].thrustfactor + (PosOrNeg * (skins[player.mo.skin].thrustfactor / P_RandomRange(1, 3))))
        player.mindash = max($ / 3, skins[player.mo.skin].mindash + (PosOrNeg * (skins[player.mo.skin].mindash / P_RandomRange(1, 3))))
        player.maxdash = max($ / 3, skins[player.mo.skin].maxdash + (PosOrNeg * (skins[player.mo.skin].maxdash / P_RandomRange(1, 3))))
        value.started = true
    end
end,

endfunc = function(player, selfid, value)
    player.normalspeed = skins[player.mo.skin].normalspeed
    player.runspeed = skins[player.mo.skin].runspeed
    player.actionspd = skins[player.mo.skin].actionspd
    player.thrustfactor = skins[player.mo.skin].thrustfactor
    player.mindash = skins[player.mo.skin].mindash
    player.maxdash = skins[player.mo.skin].maxdash
end}

addEffect{name = "Remove Random Effect", chance = FRACUNIT/21, duration = 0, sprite = "TRASH", etype = "neutral", fxrequired = 2, nobees = true, startexception = true, notimer = true,
description = "Does what it says on the tin. If it can't remove any given effect,",
description2 = "it'll just stick around until it can remove an effect.",
func = function(player, selfid, value)
    local RANDNUMB = P_RandomRange(1, #player.khaoseffects)

    for index2, value2 in ipairs(player.khaoseffects) do
        local TheNamer = KhaosEffectsTable[value2.effect]
        if index2 != selfid and index2 == RANDNUMB and not TheNamer.immune2remove and TheNamer.duration > 0 and TheNamer.name != "Roulette" and TheNamer.name != "Remove Random Effect" then
            value2.timer = value2.duration*TICRATE + 5
            value.fxdone = true
            break
        end
    end
end}

addEffect{name = "Invert Controls", chance = FRACUNIT/35, duration = 10, range = 7, sprite = "INVERT", etype = "bad", amplifyonly = true,
savereq = "amplifier",
description = "This inverts your controls, shocker. Good luck navigating before and after this.",
description2 = "There's also a small chance this will invert your camera controls too.",
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"I WAS FROZEN TODAY!", "Drifting"},
func = function(player, sselfid, value)
    if not value.started then
        if P_RandomChance(FRACUNIT/5) then
            value.InvertCam = true
        end
        value.started = true
    end
    player.khaos.InvertControlsNow = true

    if value.InvertCam then -- code taken from "MirrorControls_v2.1.lua" by QuickWolfPT
        if (player.powers[pw_carry] == CR_MINECART or player.powers[pw_carry] == CR_PTERABYTE or player.mo.state == S_PLAY_FIRE) then
            player.mo.angle = player.drawangle
            value.OldAngle = player.mo.angle
            return
        end

        if (player.mo.state == S_PLAY_CLING or player.mo.state == S_PLAY_CLIMB or (player.onconveyor == 2 and player.mo.state == S_PLAY_PAIN) or player.powers[pw_carry] == CR_ZOOMTUBE or player.powers[pw_carry] == CR_PTERABYTE) then
            value.OldAngle = player.mo.angle
            return
        end

        if ((player.pflags & PF_ANALOGMODE) and 1 or 0) then
            value.OldAngle = $ or player.mo.angle

            local angdiff = (player.cmd.angleturn << FRACBITS) - value.OldAngle
            player.mo.angle = $ - 2 * angdiff

            value.OldAngle = player.mo.angle
        end
    end

end,
endfunc= function(player, sselfid, value)
    player.khaos.InvertControlsNow = false
end}

addEffect{name = "Moonjump", chance = FRACUNIT/15, etype = "good", duration = 9, range = 4, sprite = "MOON", addition = 2, nobot = true, nonights = true,
startsfx = sfx_kc5a, endsfx = sfx_kc65, localsfx = true,
description = "You can now rapidly double jump!",
incompatible = {"Cement Shoes", "Zero Gravity", "Pancake"},
func = function(player, sselfid, value)
    if not value.started then
        value.started = true
        SRB2CE.addHUDText(player, "You can now double jump infinitely!", true)
    end

    player.charability = CA_DOUBLEJUMP
    player.pflags = $ & ~PF_THOKKED
end,
endfunc = function(player, sselfid, value)
    player.charability = skins[player.mo.skin].ability
end}

-- include super duper rare chance of brak eggman spawning

local ObjectsTable = {} -- used for "Spawn Object" effect

local function addObject(tbl)
    table.insert(ObjectsTable, tbl)
end

SRB2CE.addObject = addObject

--addObject{id = MT_EGGTRAP, name = "Eggtrap"} -- doesn't work :(

addObject{id = MT_TURRET, name = "Turret"}
addObject{id = MT_REDSPRING, name = "Red Spring", flagremover = MF_NOGRAVITY}
addObject{id = MT_BLUESPRING, name = "Blue Spring", flagremover = MF_NOGRAVITY}
addObject{id = MT_RING_BOX, name = "Super Ring Monitor"}
addObject{id = MT_BLUECRAWLA, name = "Blue Crawla"}
addObject{id = MT_GOLDBUZZ, name = "Gold Buzz"}
addObject{id = MT_JETTBOMBER, name = "Jetty-Syn Bomber"}
addObject{id = MT_DETON, name = "Deton"}
addObject{id = MT_SPINCUSHION, name = "Spincushion"}
addObject{id = MT_FACESTABBER, name = "Lance-a-Bot"}
addObject{id = MT_BLUECRAWLA, name = "Blue Crawla"}
addObject{id = MT_EGGMOBILE, name = "Eggman"}
addObject{id = MT_GOOMBA, name = "Goomba"}
addObject{id = MT_TNTBARREL, name = "TNT Barrel"}
addObject{id = MT_PYREFLY, name = "Pyre Fly"}
addObject{id = MT_PENGUINATOR, name = "Penguinator"}
addObject{id = MT_ROBOHOOD, name = "Robo-Hood"}
addObject{id = MT_GFZTREE, name = "Tree"}
addObject{id = MT_PITY_BOX, name = "Pity Shield Box"}
addObject{id = MT_TURRET, name = "Turret"}
addObject{id = MT_TURRET, name = "Turret"} -- makes the turret more common. why? funny
addObject{id = MT_MINUS, name = "Minus"}
addObject{id = MT_YELLOWSHELL, name = "Yellow Spring Shell"}
addObject{id = MT_FLAMEAURA_BOX, name = "Flame Shield Monitor"}
addObject{id = MT_BUBBLEWRAP_BOX, name = "Bubble Shield Monitor"}
addObject{id = MT_THUNDERCOIN_BOX, name = "Lightning Shield Monitor"}
addObject{id = MT_EGGGUARD, name = "Egg Guard"}

addEffect{name = "Spawn Object", chance = FRACUNIT/18, duration = 0, sprite = "PORTAL", nohudtext = true, etype = "neutral", nonights = true,
savereq = "fxpax1",
incompatible = {"I WAS FROZEN TODAY!"},
description = "Spawns an object above your head! It'll randomly be thrown forward, backwards, ",
description2 = "or just drop right down where you are/were.",
func = function(player, selfid, value)
S_StartSound(player.mo, sfx_s3kcas)
    local RANDY1 = P_RandomRange(1, #ObjectsTable)

    local theobj = ObjectsTable[RANDY1].id
    local name = ObjectsTable[RANDY1].name
    local tic


    local thej = player.mo.z + (player.mo.height + (32 * player.mo.scale))
    if ObjectsTable[RANDY1].onground then
        thej = player.mo.floorz
    end

    local Obj = P_SpawnMobj(player.mo.x, player.mo.y, thej, theobj)
    local RANDY2 = P_RandomRange(1, 3)
    if ObjectsTable[RANDY1].flagremover then
        Obj.flags = $ & ~ObjectsTable[RANDY1].flagremover
    end
    Obj.target = player.mo
    --Obj.tracer = player.mo
    Obj.angle = player.mo.angle
    if tic then
        Obj.fuse = tic
    end


    if Obj.ceilingrover and Obj.ceilingrover.valid then
        if Obj.z > Obj.ceilingrover.bottomheight then
            Obj.z = Obj.ceilingrover.bottomheight - Obj.height - 4*player.mo.scale
        end
    else
        if Obj.z > Obj.subsector.sector.ceilingheight then
            Obj.z = Obj.subsector.sector.ceilingheight - Obj.height - 4*player.mo.scale
        end
    end
    if Obj and Obj.valid then -- thanks game for being a bitch about this
        local Velocity = ObjectsTable[RANDY1].velocity and ObjectsTable[RANDY1].velocity*mapobjectscale or 22*mapobjectscale
        if not ObjectsTable[RANDY1].nothrust then
            if RANDY2 == 1 then
                P_InstaThrust(Obj, Obj.angle, Velocity)
                Obj.angle = $ + ANGLE_180
            elseif RANDY2 == 3 then
                P_InstaThrust(Obj, Obj.angle, -Velocity)
            end
        end
    end

    SRB2CE.addHUDText(player, "\x87" + "Spawned" + "\x80: " .. name)

    value.fxdone = true
end
}

--[[  addEffect{name = "Amy Invasion", chance = FRACUNIT/38, duration = 0, sprite = "PLACEHOLDER", nohudtext = true, etype = "neutral", nospecialstage = true,
--savereq = "fxpax1",
description = "Amy is here. You are not safe.",
func = function(player, selfid, value)
    local Obj = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z + (player.mo.height + (32 * player.mo.scale)), MT_ROSY)

    value.fxdone = true
end
}]]

addEffect{name = "Firework", chance = FRACUNIT/24, etype = "bad", duration = 1, sprite = "FIREWORK", addition = 1, nocart = true, nospawn = true, requirerings = true, nonights = true,
startsfx = sfx_chadet, forcestart = true,
incompatible = {"Cement Shoes", "Zero Gravity"},
description = "Shoots you really far into the sky! There's a chance the fireworks",
description2 = "will be explosive.",
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        value.explosive = false
        local amnt = FRACUNIT/2-FRACUNIT/5
        if SRB2CE.AmplifyChecker(player) then
            amnt = FRACUNIT/2+FRACUNIT/3
        end
        --[[if SRB2CE.IsMatchMode() then
            amnt = $ / 2
        end]]
        if P_RandomChance(amnt) then
            value.explosive = true
        end
    end

    local additioner = FRACUNIT + (FRACUNIT/2) + (FRACUNIT/4)

    --[[if (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
        additioner = $ * 20
    end]]

    if not isPlayerRespawning(player) then
        if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
            player.mo.momz = $ - additioner
        else
            player.mo.momz = $ + additioner
        end
    end

    local function SpawnSpark(x, y, z, state)
        local Sperk = P_SpawnMobj(player.mo.x + x*2, player.mo.y + y*3, player.mo.z + z/4, MT_BOXSPARKLE)
        Sperk.scale = player.mo.scale
        Sperk.momx = x * 2
        Sperk.momy = y * 2
        Sperk.momz = z * 10
        if state then
            Sperk.state = state
            Sperk.scale = $ / 2
        end
    end

    for i=6,1,-1 do
        SpawnSpark(P_RandomFixed(), P_RandomFixed(), P_RandomFixed())
        if value.explosive then
            SpawnSpark(P_RandomFixed(), P_RandomFixed(), P_RandomFixed(), S_SPINDUST_FIRE1)
        end
    end
end,
endfunc = function(player, selfid, value)
    --S_StartSound(player.mo, sfx_kc65, player)

    local function SpawnSpark(x, y, z, state) --gotta do this twice now :pensive:
        local Sperk = P_SpawnMobj(player.mo.x + x*2, player.mo.y + y*3, player.mo.z + z/4, MT_BOXSPARKLE)
        Sperk.scale = player.mo.scale
        Sperk.momx = x * 2
        Sperk.momy = y * 2
        Sperk.momz = z * 10
        if state then
            Sperk.state = state
            Sperk.scale = $ / 2
        end
    end

    for i=69,1,-1 do -- nice
        SpawnSpark(P_RandomFixed()*4, P_RandomFixed()*5, P_RandomFixed()*6)
    end

    if value.explosive then
        local explode = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_TNTBARREL)
        explode.state = S_TNTBARREL_EXPL1
        player.khaos.Check4Hit = max($, 1)
    end
end
}

addEffect{name = "Bouncy!", chance = FRACUNIT/37, duration = 4, range = 2, etype = "bad", sprite = "BOUNCE", addition = 2, nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Gravity Coil", "Cement Shoes", "Butter Physics", "Zero Gravity"},
description = "The floors don't like you, and keep bouncing you back up!",
description2 = "The walls have similar feelings!",
func = function(player, selfid, value)
    if not value.started then
        player.mo.momz = 3*FRACUNIT
        player.mo.z = $ + 7*FRACUNIT
        value.LastMomZ = player.mo.momz
        value.started = true
    end

    if player.mo.momz/FRACUNIT == 0 then
        if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
            if value.LastMomZ > 0 then
                value.BounceTime = true
            end
        else
            if value.LastMomZ < 0 then
                value.BounceTime = true
            end
        end
    else
        value.LastMomZ = player.mo.momz
    end

    if value.LastMomZ < FRACUNIT and P_IsObjectOnGround(player.mo) then
        value.LastMomZ = $ + 7*FRACUNIT
    end

    value.BounceTime = true

    if value.BounceTime then
        if P_IsObjectOnGround(player.mo) then
            player.mo.momz = abs(value.LastMomZ)
            if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
                player.mo.momz = -$
            end
            S_StartSoundAtVolume(player.mo, sfx_kc52, 150)
            value.BounceTime = false
        end
    end

    if player.khaos.wallblock == TICRATE-1 then
        P_InstaThrust(player.mo, player.mo.angle-ANGLE_180, 15*FRACUNIT)
        S_StartSoundAtVolume(player.mo, sfx_kc52, 150)
    end
end}

addEffect{name = "Bleeding", chance = FRACUNIT/18, duration = 12, range = 5, sprite = "BLEED", addition = 7, etype = "bad", requirerings = true, nonights = true, -- CEIDK0
startsfx = sfx_kc59, endsfx = sfx_kc5a, localsfx = true,
noboss = true,
incompatible = {"Money Bag"},
description = "You're bleeding rings! You will frequently drop the rings you have collected.",
func = function(player, selfid, value)
    local function SetTime()
        if SRB2CE.AmplifyChecker(player)  then
            value.timeseed = P_RandomRange(TICRATE/4, TICRATE-TICRATE/5)
        else
            value.timeseed = P_RandomRange(TICRATE/2, TICRATE+(TICRATE/5))
        end
    end
    if not value.started then
        value.started = true
        SetTime()
    end

    -- every second, player is drained one ring, goes invicible for a split second, and a ring pops out of them

    local function Bleed(silent)
        if player.rings > 0 then
            --[[if player.powers[pw_flashing] == 0 then
                player.powers[pw_flashing] = TICRATE/3 + 3
            end]]
            local z = (player.mo.z + player.mo.height) + 16*FRACUNIT
            if (player.mo.eflags & MFE_VERTICALFLIP) then
                z = (player.mo.z - player.mo.height) - 16*FRACUNIT
            end
            local ring = SpawnFlingRing(player.mo, nil, nil, z)

            P_GivePlayerRings(player, -1)
            if not silent then
                S_StartSound(player.mo, sfx_antiri)
                if P_RandomChance(FRACUNIT/4) then
                    SetTime()
                end
            end

        end
    end

    if leveltime % value.timeseed == 0 then
        Bleed()
        if SRB2CE.AmplifyChecker(player) then
            Bleed(true)
        end
    end
end}


addEffect{name = "Windy Day", chance = FRACUNIT/26, duration = 6, range = 3, sprite = "WIND", addition = 3, etype = "bad", drainrespawntimer = true, nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
description = "A mighty gust is blowing you every which way! You're blown in a certain direction, ",
description2 = "which changes every now and then to blow you a different way",
func = function(player, selfid, value)
    if not value.started then
        S_StartSound(player.mo, sfx_s3kcel, player)
        value.angle = player.mo.angle-ANGLE_180
        value.changeang = 0
        value.thrust = FRACUNIT/2 - FRACUNIT/8
        value.posneg = false -- honestly i have no fucking clue what this does. what drugs were you on, 2020-or-so jasmine?
        value.started = true
    end


    if value.changeang > 0 then
        if value.posneg then
            value.angle = $ - (ANG2 + ANG1)
        else
            value.angle = $ + ANG2 + ANG1
        end
        value.changeang = $ - 1
        value.thrust = P_RandomRange(FRACUNIT/2 - FRACUNIT/8, FRACUNIT/2 - FRACUNIT/4) + P_RandomFixed()/4
        if SRB2CE.AmplifyChecker(player) then
            value.thrust = $ + P_RandomFixed()/2
        end
        if VERSION < 2 then
            value.thrust = $ * 2
        end
    else
        value.changeang = 0
        value.posneg = not value.posneg
        if P_RandomChance(FRACUNIT/5) then
            value.changeang = P_RandomRange(5, 25)
            if SRB2CE.AmplifyChecker(player) then
                value.changeang = $ + 2
            end
            --SRB2CE.DebugPrint(player, "Changing Angle!")
        end
    end

    P_Thrust(player.mo, value.angle, value.thrust)
end}

addEffect{name = "Invisible Treadmill", chance = FRACUNIT/29, duration = 10, range = 3, sprite = "TREADMILL", addition = 3, etype = "bad", drainrespawntimer = true, nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Windy Day", "Always Watching."},
description = "You're getting constantly pushed the direction your character is facing!",
func = function(player, selfid, value)
    if not value.started then
        value.thrust = FRACUNIT + FRACUNIT/8
        if SRB2CE.AmplifyChecker(player) then
            value.thrust = $ + (P_RandomFixed())*2
        end
        value.started = true
    end

    P_Thrust(player.mo, player.drawangle, value.thrust)
end}

addEffect{name = "Zero Gravity", chance = FRACUNIT/26, etype = "bad", duration = 5, range = 2, sprite = "ZEROGRAV", addition = 3, resetvalues = true, nonights = true,
startsfx = sfx_s3k9c, endsfx = sfx_kc65,
incompatible = {"Cement Shoes", "Moonjump", "Bouncy!", "Gravity Coil"},
description = "Your gravity has been turned off! Often results in you being catapulted ",
description2 = "to the sky.",
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        value.laststoredmom = player.mo.momz
    end

    if not isPlayerRespawning(player) then
        if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
            if value.laststoredmom < player.mo.momz then
                player.mo.momz = value.laststoredmom
            end
        else
            if value.laststoredmom > player.mo.momz then
                player.mo.momz = value.laststoredmom
            end
        end
    end

    value.laststoredmom = player.mo.momz
end}

addEffect{name = "Eye for an Eye", chance = FRACUNIT/33, etype = "bad", duration = 13, range = 3, sprite = "KNIFE", addition = 2, nonights = true,
startsfx = sfx_s3k9c, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Lucky!", "Timebomb"},
description = "When you deal damage, you take damage. But when you take damage, ",
description2 = "your inflictor takes damage.",
func = function(player, selfid, value)
    player.khaos.Eye4Eye = true
end,
endfunc = function(player, selfid, value)
    player.khaos.Eye4Eye = false
end}


addEffect{name = "Timebomb", chance = FRACUNIT/31, duration = 3, addition = -1, sprite = "BOMB", etype = "bad", drainrespawntimer = true, noghost = true, requirerings = true,
startsfx = sfx_s3k5c,
incompatible = {"Ghost", "Eye for an Eye"},
description = "\x84\"Somebody set us up the bomb!\"\x80 You have a few seconds to react to the fact ",
description2 = "you're gonna explode!",
func = function(player, selfid, value)
    if not value.started then
        value.repeatsound = 1
        value.subtracttime = 8
        value.sublimetime = 3
        value.defaulttimer = value.duration*TICRATE - value.timer
        value.started = true
        if P_RandomChance(FRACUNIT/8) then
            S_StartSound(player.mo, sfx_cabomb)
            S_StartSound(player.mo, sfx_cabomb)
        end
    end

    local function SpawnSplode(x, y, z)
        local theobj = MT_SONIC3KBOSSEXPLODE
        if P_RandomChance(FRACUNIT/5) then
            theobj = MT_BOSSEXPLODE
        end
        local Sperk = P_SpawnMobj(player.mo.x + x*2, player.mo.y + y*3, player.mo.z + z/4, theobj)
        Sperk.scale = player.mo.scale
        Sperk.momx = x * 2
        Sperk.momy = y * 2
        Sperk.momz = z * 5
    end

    local function Blinker()
        if not value.back then
            player.mo.colorized = true
            value.prevcolor = player.mo.color
            player.mo.color = SKINCOLOR_RED
            value.back = true
        else
            player.mo.colorized = false
            player.mo.color = value.prevcolor
            value.back = false
        end
    end

    SpawnSplode(P_RandomFixed()*2, P_RandomFixed()*3, P_RandomFixed()*2)

    if value.defaulttimer > (value.duration*TICRATE - value.timer) - value.repeatsound then -- beeping
        value.repeatsound = $ - value.subtracttime
        value.sublimetime = $ - 1
        if value.sublimetime <= 0 then
            value.sublimetime = 3
            value.subtracttime = $ - 1
        end
        S_StartSound(player.mo, sfx_deton) --sfx_kc2a
        Blinker()
    end

    if value.duration*TICRATE - value.timer == TICRATE/2 + TICRATE/3 then
        S_StartSound(player.mo, sfx_s3k86)
    end
end,
endfunc = function(player,selfid, value)
    local JankSplosion = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_TNTBARREL)
    JankSplosion.target = player.mo
    P_KillMobj(JankSplosion)
    S_StartSound(player.mo, sfx_s3k4e)
    player.mo.color = value.prevcolor or $ -- for some reason a horrible bug occured because this wasn't set while a level was ending, weird
end}

addEffect{name = "the ride never ends", chance = FRACUNIT/20, duration = 15, range = 8, sprite = "THERIDE", addition = 4, etype = "bad", nonights = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Downside Up"},
description = "\x84\"I want to get off TAILS WILD RIDE.\"\x80 Your screen's perspective gets angled!",
description2 = "It gets worse in Amplify Mode.",
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        local RANDYSPIN = P_RandomChance(FRACUNIT/2)
        if not RANDYSPIN then
            player.khaos.SpinOtherWay = true
        else
            player.khaos.SpinOtherWay = false
        end
        value.angler = (ANG30 + ANG15)
        if SRB2CE.AmplifyChecker(player) then
            value.angler = (ANG30 + ANG30)
        end
    end
    if not player.khaos.SpinOtherWay then
        if player.viewrollangle < value.angler then
            player.viewrollangle = $ + (ANG1/5)
            player.mo.rollangle = $ - (ANG1/5)
        end
    else
        if player.viewrollangle > -value.angler then
            player.viewrollangle = $ - (ANG1/5)
            player.mo.rollangle = $ + (ANG1/5)
        end
    end
end,
endfunc = function(player, selfid, value)
    player.khaos.RestoreViewAngle = true
end}

addEffect{name = "Pancake", chance = FRACUNIT/10, duration = 8, range = 5, sprite = "PANCAKE", addition = 7, etype = "bad", nonights = true,
startsfx = sfx_chcake, endsfx = sfx_kc65, forcestart = true,
description = "You get squashed like a pancake, and your max speed gets halved. This is ",
description2 = "more common than Cement Shoes, and less incompatible with other effects.",
incompatible = {"Cement Shoes", "Sticky Slopes", "Moonjump", "HE"},
func=function(player, selfid, value)
    if not value.started then
        value.started = true
        value.GOALY = FRACUNIT/4 + FRACUNIT/20
        if SRB2CE.AmplifyChecker(player) then
            value.GOALY = FRACUNIT/6 + FRACUNIT/9 -- nice
        end
        player.khaos.RestoreSpriteXScaleUp = nil
        player.khaos.RestoreSpriteYScaleUp = nil
        player.khaos.RestoreSpriteXScaleDown = nil
        player.khaos.RestoreSpriteYScaleDown = nil
    end

    player.mo.spritexscale = FRACUNIT + FRACUNIT

    if player.mo.spriteyscale != value.GOALY then
        if player.mo.spriteyscale > value.GOALY then
            player.mo.spriteyscale = $ - FRACUNIT/2
        elseif player.mo.spriteyscale < value.GOALY then
            player.mo.spriteyscale = value.GOALY
        end
    end

    if SRB2CE.AmplifyChecker(player) then
        player.normalspeed = skins[player.mo.skin].normalspeed / 3
        player.runspeed = skins[player.mo.skin].runspeed / 3
        player.actionspd = skins[player.mo.skin].actionspd / 3
        player.thrustfactor = skins[player.mo.skin].thrustfactor / 3
        player.mindash = FRACUNIT
        player.maxdash = skins[player.mo.skin].maxdash / 2
    else
        player.normalspeed = (skins[player.mo.skin].normalspeed / 2) + (skins[player.mo.skin].normalspeed / 4)
        player.runspeed = (skins[player.mo.skin].runspeed / 2) + (skins[player.mo.skin].runspeed / 3)
        player.actionspd = skins[player.mo.skin].actionspd / 2
        player.thrustfactor = (skins[player.mo.skin].thrustfactor / 2) + (skins[player.mo.skin].thrustfactor / 3)
    end
end,
endfunc=function(player, selfid, value)
    player.normalspeed = skins[player.mo.skin].normalspeed
    player.runspeed = skins[player.mo.skin].runspeed
    player.actionspd = skins[player.mo.skin].actionspd
    player.thrustfactor = skins[player.mo.skin].thrustfactor
    player.mindash = skins[player.mo.skin].mindash
    player.maxdash = skins[player.mo.skin].maxdash
    player.height = skins[player.mo.skin].height
    player.khaos.RestoreSpriteYScaleUp = true
    player.mo.spritexscale = FRACUNIT
end}

addEffect{name = "Give Good Stuff", chance = FRACUNIT/24, duration = -1, notimer = true, etype = "good", sprite = "GOODSTUFF", startsfx = sfx_chchng, forcestart = true, localsfx = true, nohudtext = true,
description = "Gives the player a few good things, like shields, power-ups, etc.",
func = function(player, selfid, value)
    local GivenGoodStuff = false

    if not player.powers[pw_shield] and (P_RandomChance(FRACUNIT/2) or (SRB2CE.AmplifyChecker(player))) then
        player.powers[pw_shield] = SH_PITY
        P_SpawnShieldOrb(player)
        SRB2CE.addHUDText(player, "Given a Shield")
        GivenGoodStuff = true
    end

    if player.powers[pw_underwater] > 0 then
        player.powers[pw_underwater] = 0
        SRB2CE.addHUDText(player, "Reset Drowning Timer")
    end

    if P_RandomChance(FRACUNIT/2) then
        player.powers[pw_sneakers] = $ + 20*TICRATE
        P_RestoreMusic(player)
        SRB2CE.addHUDText(player, "Given Super Sneakers")
        GivenGoodStuff = true
    end

    if P_RandomChance(FRACUNIT/3) or (SRB2CE.AmplifyChecker(player) and P_RandomChance(FRACUNIT/2)) then
        if SRB2CE.IsMatchMode() then
            player.powers[pw_invulnerability] = $ + 10*TICRATE
        else
            player.powers[pw_invulnerability] = $ + 20*TICRATE
        end
        P_RestoreMusic(player)
        SRB2CE.addHUDText(player, "Given Invulernability")
        GivenGoodStuff = true
    end

    if P_RandomChance(FRACUNIT/15) or (SRB2CE.AmplifyChecker(player) and P_RandomChance(FRACUNIT/8)) then
        P_GivePlayerRings(player, 111)
        SRB2CE.addHUDText(player, "Given 111 rings")
        GivenGoodStuff = true
    end

    if P_RandomChance(FRACUNIT/99) or (SRB2CE.AmplifyChecker(player) and P_RandomChance(FRACUNIT/75)) then
        P_GivePlayerRings(player, 999)
        if SRB2CE.IsMatchMode() then
            player.powers[pw_invulnerability] = $ + 30*TICRATE
        else
            player.powers[pw_invulnerability] = $ + 500*TICRATE
        end
        --P_RestoreMusic(player)
        SRB2CE.addHUDText(player, "Given 999 rings, and a lot of Invulnerability!")
        GivenGoodStuff = true
    end

    if not GivenGoodStuff then
        --debugprint("somehow the player got nothing, giving them a bone")
        if SRB2CE.IsMatchMode() then
            player.powers[pw_invulnerability] = $ + 5*TICRATE
        else
            player.powers[pw_invulnerability] = $ + 15*TICRATE
        end
        P_GivePlayerRings(player, 15)
        SRB2CE.addHUDText(player, "Given some invulnerability, and 15 rings")
        P_RestoreMusic(player)
    end

    value.fxdone = GivenGoodStuff
end
}


addEffect{name = "Cement Shoes", chance = FRACUNIT/21, duration = 7, range = 4, sprite = "CEMENTSHOES", addition = 1, nocart = true, norock = true, nonights = true, etype = "bad",
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
description = "Your shoes have been filled with cement! Your max speed gets reduced by a lot.",
incompatible = {"Super Jump", "HE", "Pancake", "Sticky Slopes", "I WAS FROZEN TODAY!", "Change Color", "Butter Physics", "Bouncy!", "Gravity Coil", "Moonjump", "Zero Gravity"}, -- so many items
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        value.CementPrevColor = player.mo.color
        value.Defaultkartspeed = player.kartspeed
        value.Defaultkartweight = player.kartweight
        if not SRB2CE.IsMatchMode() then
            player.mo.colorized = true
            player.mo.color = SKINCOLOR_GREY
        end
    end
    if SRB2CE.AmplifyChecker(player) then
        player.normalspeed = skins[player.mo.skin].normalspeed / 7
        player.runspeed = skins[player.mo.skin].runspeed / 5
        player.actionspd = skins[player.mo.skin].actionspd / 9
        player.thrustfactor = skins[player.mo.skin].thrustfactor / 5
        player.mindash = FRACUNIT
        player.maxdash = skins[player.mo.skin].maxdash / 3
    else
        player.normalspeed = skins[player.mo.skin].normalspeed / 5
        player.runspeed = skins[player.mo.skin].runspeed / 4
        player.actionspd = skins[player.mo.skin].actionspd / 7
        player.thrustfactor = skins[player.mo.skin].thrustfactor / 3
    end

    -- P_SetObjectMomZ(player.mo, -gravity, true)

    if (player.mo.eflags & MFE_VERTICALFLIP) or (player.mo.flags2 & MF2_OBJECTFLIP) then
        if player.mo.momz > 0 then
            player.mo.momz = $ + $ / 10
        else
            player.mo.momz = $ - $ / 8
        end
    else
        if player.mo.momz > 0 then
            player.mo.momz = $ - $ / 10
        else
            player.mo.momz = $ + $ / 8
        end
    end

    if not P_IsObjectOnGround(player.mo) then
        player.khaos.CementShoesSfx = true
        if player.mo.momz < -(3*FRACUNIT) and not player.khaos.CementShoesFallSfx then
            S_StartSound(player.mo, sfx_s3k51, player)
            player.khaos.CementShoesFallSfx = true
        end
    else
        if player.khaos.CementShoesSfx then
            S_StartSound(player.mo, sfx_s3k5f) --previously sfx_pstop, and before that, sfx_doord2
            --P_StartQuake(15, 10*FRACUNIT, 10*FRACUNIT, player.mo)
            P_StartQuake(10*FRACUNIT, 5, {player.mo.x, player.mo.y, player.mo.z}, 700*FRACUNIT)
            player.khaos.CementShoesSfx = false
        end
        player.khaos.CementShoesFallSfx = false

        value.angle = player.mo.angle - ANGLE_180
        value.thrust = player.speed / 50

        P_Thrust(player.mo, value.angle, value.thrust)
    end
end,
endfunc = function(player,selfid, value)
    player.normalspeed = skins[player.mo.skin].normalspeed
    player.runspeed = skins[player.mo.skin].runspeed
    player.actionspd = skins[player.mo.skin].actionspd
    player.thrustfactor = skins[player.mo.skin].thrustfactor
    player.mindash = skins[player.mo.skin].mindash
    player.maxdash = skins[player.mo.skin].maxdash
    player.mo.color = value.CementPrevColor
    player.mo.colorized = false
end}

addEffect{name = "Butter Physics", chance = FRACUNIT/14, duration = 12, range = 7, sprite = "BUTTER", addition = 2, etype = "bad", nonights = true, nocart = true, norock = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Cement Shoes", "Sticky Slopes"},
description = "Forget ice physics, your shoes have been slathered with butter. You are basically ",
description2 = "unstoppable. You might go zooming around the level, or into pits.",
func = function(player, selfid, value)
    if not player.khaos.OGFriction then player.khaos.OGFriction = player.mo.friction end
    if SRB2CE.AmplifyChecker(player) then
        player.mo.friction = FRACUNIT + (FRACUNIT/69) -- nice
    else
        player.mo.friction = FRACUNIT + (FRACUNIT/56)
    end
end,
endfunc = function(player,selfid, value)
    player.mo.friction = player.khaos.OGFriction
end}

addEffect{name = "Fling Player", chance = FRACUNIT/10, duration = 0, sprite = "CANNON", nohudtext = true, etype = "bad", norock = true, nocart = true, nonights = true,
incompatible = {"I WAS FROZEN TODAY!"},
description = "You get tossed around, either forward or backwards.",
func=function(player, selfid, value)
    local RANDYSLAP
    local RANDYYEET = P_RandomChance(FRACUNIT/20) -- 5%
    local Angler = player.mo.angle + (P_RandomRange(0, 4) * ANGLE_90)
    if SRB2CE.AmplifyChecker(player) then
        P_InstaThrust(player.mo, Angler, 310*FRACUNIT)
        if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/3) end -- 33.3%, reduced chances since spinout causes you to lose control
    else
        P_InstaThrust(player.mo, Angler, 210*FRACUNIT)
        if P_IsObjectOnGround(player.mo) then RANDYSLAP = P_RandomChance(FRACUNIT/2 + FRACUNIT/6) end -- 66.6%
    end
    if RANDYSLAP or RANDYYEET then
        if RANDYYEET then
            SRB2CE.addHUDText(player, "YEET", true)
            local yeetsfx = sfx_chyee2
            if P_RandomChance(FRACUNIT/2) then
                yeetsfx = sfx_chyeet
            end
            S_StartSound(player.mo, yeetsfx)
            S_StartSound(player.mo, sfx_kc5b, player)
        else
            SRB2CE.addHUDText(player, "Slapped!")
            S_StartSound(player.mo, sfx_slappy)
        end
        player.mo.state = S_PLAY_PAIN
        player.climbing = 0
        player.powers[pw_flashing] = TICRATE
        if SRB2CE.AmplifyChecker(player) then
            player.mo.momz = $ + 35*FRACUNIT
        else
            player.mo.momz = $ + 20*FRACUNIT
        end
    else
        SRB2CE.addHUDText(player, "Fling Player")
        S_StartSound(player.mo, sfx_s3k81)
        player.mo.momz = $ + 2*FRACUNIT
    end
    value.fxdone = true
end
}

local eatmydust

if VERSION >= 2 then
    eatmydust = MT_DUST
else
    eatmydust = MT_BOOSTSMOKE
end

addEffect{name = "Change Character", chance = FRACUNIT/29, duration = 3, range = 2, sprite = "CHARS", etype = "neutral", endsfx = sfx_kc65, forceend = true,
incompatible = {"I WAS FROZEN TODAY!", "Change Color"},
description = "Changes your current character repeatedly until the effect expires. ",
description2 = "This won't pick any characters you haven't unlocked yet!",
func = function(player, selfid, value)
    if not value.started then
        local RANDYC = P_RandomRange(-5, TICRATE-(TICRATE)/2) -- approximately 17 tics, or half a second
        value.timer = $ - RANDYC
        value.skinzor = skins[1].name --iniitialize it to something so we don't get an error
        value.started = true
    end
    if (leveltime % 5 == 0) then
        S_StartSoundAtVolume(player.mo, sfx_kc56, 136)
        repeat
            value.skinzor = skins[P_RandomRange(0, #skins-1)].name
        until value.skinzor and (VERSION < 2 or (VERSION >= 2 and R_SkinUsable(player, value.skinzor))) -- have to check skinzor is unlocked by the player
        R_SetPlayerSkin(player, value.skinzor)
    end
end,
endfunc = function(player, selfid, value)
    if not value.freeze then
        local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
        player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, eatmydust)
        dust.scale = FRACUNIT*2
        dust.state = S_XPLD1
        SRB2CE.addHUDText(player, "Character changed to: " .. skins[player.mo.skin].realname)
        --[[if player.bot then
            player.name = skins[player.mo.skin].realname -- doesn't work :(
        end]]
    end
end
}

addEffect{name = "Change Color", chance = FRACUNIT/13, duration = 3, range = 2, sprite = "COLORS", etype = "neutral", nomatch = true, endsfx = sfx_kc65, forceend = true, nonights = true,
incompatible = {"I WAS FROZEN TODAY!", "Cement Shoes", "Change Character"},
description = "This changes your character's color until the timer expires.",
func = function(player, selfid, value)
    if (leveltime % 4 == 0) then
        local funnytablelength = 0
        if VERSION >= 2 then
            funnytablelength = #skincolors
        else
            funnytablelength = MAXSKINCOLORS
        end
        player.mo.color = P_RandomRange(1, funnytablelength-1)
        S_StartSoundAtVolume(player.mo, sfx_kc56, 136)
        player.skincolor = player.mo.color
    end
end,
endfunc = function(player,selfid, value)
    if not value.freeze then
        local dust = P_SpawnMobj(player.mo.x + P_ReturnThrustX(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)),
        player.mo.y + P_ReturnThrustY(player.mo, player.mo.angle, FixedMul(-1*FRACUNIT, player.mo.scale)), player.mo.z, eatmydust)
        dust.scale = FRACUNIT*2
        dust.state = S_XPLD1
        if VERSION >= 2 then
            SRB2CE.addHUDText(player, "Color changed to: " .. R_GetNameByColor(player.mo.color))
        end
        player.skincolor = player.mo.color -- one last time, just to be sure
    end
end
}

addEffect{name = "Lucky!", chance = FRACUNIT/16, etype = "good", duration = 15, range = 8, sprite = "LUCK", addition = 1,
startsfx = sfx_s1c9, endsfx = sfx_kc65, localsfx = true,
incompatible = {"PANIC", "Eye for an Eye"},
description = "This makes it so most of the time you take a hit, you instead dodge it, you're also ",
description2 = "more likely to roll good effects than bad, and time between effects is increased.",
func=function(player, selfid, value)
    player.khaos.IsLucky = true
    if player.khaos.timer > 0 and ((SRB2CE.AmplifyChecker(player) and P_RandomChance(FRACUNIT/2)) or P_RandomChance(FRACUNIT/4)) then -- reduce da timer just a bit
        player.khaos.timer = $ - 1
    end
end,
endfunc = function(player)
    player.khaos.IsLucky = false
end
}

addEffect{name = "PANIC", chance = FRACUNIT/30, etype = "bad", duration = 14, range = 6, sprite = "PANIC", addition = 3,
startsfx = sfx_alarm, endsfx = sfx_kc65, localsfx = true,
incompatible = {"Lucky!"},
description = "Effects get applied to you alarmingly quick! You're also ",
description2 = "more likely to roll bad effects!",
func=function(player, selfid, value)
    player.khaos.timer = $ + 1
    player.khaos.IsPanicking = true
    if SRB2CE.AmplifyChecker(player) or P_RandomChance(FRACUNIT/14) then -- do it even m o r e
        player.khaos.timer = $ + 2
    end
end,
endfunc = function(player)
    player.khaos.IsPanicking = false
end
}

addEffect{name = "Kids These Days", chance = FRACUNIT/32, duration = 11, range = 4, sprite = "IPADKID", etype = "bad", addition = 5, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65,
incompatible={"Darkened Vision", "DVD Screensaver", ":3"},
description = "\x84\"Back in my day...\"\x80 The sides of your screen are obscured!",
func=function(player,selfid,value)
    player.khaos.ipadkid = true
end,

endfunc=function(player,selfid,value)
player.khaos.ipadkid = false
end}

addEffect{name = ":3", chance = FRACUNIT/25, duration = 11, range = 4, sprite = "COLONTHREE", etype = "bad", addition = 5, nobot = true,
startsfx = sfx_ksquee, endsfx = sfx_kc65, forcestart = true,
incompatible = {"Kids These Days"},
description = "Two creatures are being held up for your viewing delight.",
func=function(player,selfid,value)
    player.khaos.animaltime = true
end,

endfunc=function(player,selfid,value)
    player.khaos.animaltime = false
end}

addEffect{name = "Busted Controller", chance = FRACUNIT/30, duration = 8, range = 3, sprite = "BUSTEDC", etype = "bad", amplifyonly = true, nocart = true, norock = true, ignoremultiplier = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible = {"No Turning", "I WAS FROZEN TODAY!"},
description = "Your little sibling busted up your controller! How rude of them.",
description2 = "Random control inputs will be given at random intervals.",
func=function(player,selfid,value)
    player.khaos.BustedController = true
end,

endfunc=function(player,selfid,value)
    player.khaos.BustedController = false
end}


addEffect{name = "Broken Screen", chance = FRACUNIT/43, duration = 12, range = 5, sprite = "SCREEN", etype = "bad", addition = 8, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
savereq = "brokenscreen",
incompatible={"Kids These Days", "Darkened Vision"},
description="\x84\"You've done incurred my wrath, son!\"\x80 Your screen is broken, ",
description2 = "just like what you did to Tails' screen.",
func=function(player,selfid,value)
    player.khaos.BrokenScreenNow = true
end,

endfunc=function(player,selfid,value)
    player.khaos.BrokenScreenNow = false
end}

addEffect{name = "Darkened Vision", chance = FRACUNIT/25, duration = 18, range = 5, sprite = "DARKV", etype = "bad", addition = 5, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible={"Kids These Days", "Broken Screen"},
description="\x84\"Ya bring a light? No!\"\x80 Your vision is significantly darkened!",
func=function(player,selfid,value)
    player.khaos.DarkenedVisionNow = true
end,

endfunc=function(player,selfid,value)
    player.khaos.DarkenedVisionNow = false
end}

addEffect{name = "YOU'RE WINNER", chance = FRACUNIT/20, duration = 18, range = 7, sprite = "WINNER", etype = "bad", addition = 5, nobot = true,
startsfx = sfx_ideya, endsfx = sfx_kc65, localsfx = true, --sfx_xideya?
incompatible={"Kids These Days"},
description="You win! Congratulations! Wait, you didn't? Well, you'll get ",
description2 = "showered in confetti anyways.",
func=function(player,selfid,value)
    player.khaos.yourewinner = true
end,

endfunc=function(player,selfid,value)
    player.khaos.yourewinner = false
end}

--[[  addEffect{name = "Marked For Death", chance = FRACUNIT/36, duration = 25, range = 5, sprite = "MARKED", etype = "bad", addition = 5, nospecialstage = true, nosrb2kart = true,
startsfx = sfx_chmark, endsfx = sfx_kc65, localsfx = true,
incompatible={"Super Explosive 64", "Lucky!", "Nuke"},
description  = "Every hit you take will make you tumble instead! There's also a chance you just get stunlocked for a little bit, watch out!",
func = function(player, selfid, value)
    if not value.started then
        player.khaos.Marked4Death = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z+105*(player.mo.scale/2), MT_MARKED4DEATH)
        player.khaos.Marked4Death.gayer = player
        value.started = true
    end

    player.khaos.ImDyingSquirtle = true
    if (leveltime % 6) == 0 then
        player.khaos.Tumbler = $ - 1
    end
end,
endfunc=function(player,selfid,value)
    player.khaos.ImDyingSquirtle = false
    player.khaos.Tumbler = 0
    if player.khaos.Marked4Death and player.khaos.Marked4Death.valid then
        P_KillMobj(player.khaos.Marked4Death)
    end
    player.khaos.Marked4Death = nil
end}]]

addEffect{name = "Pong", chance = FRACUNIT/45, etype = "bad", duration = 12, range = 5, sprite = "PONG", addition = 5, startexception = true,
--hidename = true, savereq = "arcade",
startsfx = sfx_kc65, endsfx = sfx_kc65, forcestart = true,
description="You have to play pong! If you lose, you get hit, no matter what.", description2="And in Amplify Mode, you just die. Your life will depend on pong.",
func=function(player,selfid,value)
    if not value.started then
        player.khaos.ponglost = false
        player.khaos.pongwon = false
        player.khaos.pongtime = true
        value.started = true
    end
    if player.khaos.ponglost or player.khaos.pongwon then
        value.fxdone = true
    end
end,
endfunc=function(player,selfid,value)
    player.khaos.pongtime = false
    player.khaospong = nil
    if player.khaos.ponglost then
        P_RemoveShield(player)
        player.powers[pw_invulnerability] = 0
        player.powers[pw_flashing] = 0
        player.powers[pw_super] = 0
        if SRB2CE.AmplifyChecker(player) then
            P_KillMobj(player.mo)
        else
            P_DamageMobj(player.mo)
        end
    end
    if not player.khaos.pongwon and not player.khaos.ponglost and (value.timer >= value.duration*TICRATE) then
        player.khaossp.playerpoints = ($ and $ + 1) or 1
    end
    player.khaos.pongwon = false
    player.khaos.ponglost = false
end}

addEffect{name = "Invisible", chance = FRACUNIT/26, etype = "bad", duration = 15, range = 5, sprite = "NOTHING", addition = 3,
startsfx = sfx_kc65, endsfx = sfx_kc65, forcestart = true,
incompatible = {"Ghost", "Change Sprite", "Laggy Model"},
description="\x84\"I'm the invisible man.\"\x80 You can't see yourself, but enemies still can.",
func=function(player,selfid,value)
    if SUBVERSION < 14 and VERSION == 202 then
        player.khaos.InvisibleNow = true
    else
        player.mo.alpha = max(0, $ - (FRACUNIT/50))
    end
end,
endfunc=function(player, selfid, value)
    if SUBVERSION < 14 and VERSION == 202 then
        player.khaos.InvisibleNow = false
    else
        player.mo.alpha = FRACUNIT
    end
end}

addEffect{name = "DVD Screensaver", chance = FRACUNIT/27, duration = 12, range = 8, sprite = "DVD", etype = "bad", addition = 5, nobot = true,
startsfx = sfx_kc59, endsfx = sfx_kc65, localsfx = true,
incompatible={"Kids These Days"}, -- how fitting
description="Your screen has the DVD logo bouncing around! This hinders vision a lot.",
func=function(player,selfid,value)
    if not value.started then
        value.started = true
        local RANDY = P_RandomChance(FRACUNIT/8) -- pretty rare chance, just like how rare HD DVDs are to come across in the real world
        if RANDY then
            player.khaos.HDDVD = true
        end
    end

    player.khaos.DVDHud = true
end,

endfunc=function(player,selfid,value)
    player.khaos.DVDHud = false
    player.khaos.HDDVD = false
end}


addEffect{name = "Always Watching.", chance = FRACUNIT/19, duration = 18, range = 15, sprite = "EYE", etype = "neutral", nonights = true,
startsfx = sfx_s3k8a, endsfx = sfx_kc65,
incompatible = {"I WAS FROZEN TODAY!", "Invisible"},
description="Your character is turned around 180 degrees, to face you. They know what you did.",
func=function(player,selfid, value)
    player.drawangle = player.mo.angle + ANGLE_180
end}


addEffect{name = "but here's the spinner", chance = FRACUNIT/21, duration = 14, range = 12, sprite = "SPIN", etype = "neutral", nonights = true,
startsfx = sfx_s3k8a, endsfx = sfx_kc65,
savereq = "fxpax2",
incompatible = {"I WAS FROZEN TODAY!", "Invisible", "Always Watching."},
description="\x84\"Spin Mode Activated.\"\x80  You're spinning!",
func=function(player,selfid, value)
    if not value.started then
        value.spinangle = 0
        value.started = true
    end
    value.spinangle = $ + ANG2 + ANG2


    player.drawangle = player.mo.angle + value.spinangle
end}

addEffect{name = "Superring", chance = FRACUNIT/15, duration = 0, etype = "good", sprite = "RING", startsfx = sfx_itemup, nonights = true,
description="Gives you 10 rings! 20 rings in Amplify Mode.",
savereq = "fxpax2",
func=function(player,selfid, value)
    if SRB2CE.AmplifyChecker(player) then
    P_GivePlayerRings(player, 20)
    else
        P_GivePlayerRings(player, 10)
    end
    value.fxdone = true
end}

addEffect{name = "Money!", chance = FRACUNIT/45, duration = 0, etype = "good", sprite = "KRAB", startsfx = sfx_chmony, noboss = true, nonights = true, nomatch = true,
description="All enemies in a considerably large radius are turned into rings!",
savereq = "fxpax2",
func=function(player,selfid, value)
    local m = player.mo --used to be realmo, but ring racers is silly and thinks we shouldn't have realmo!
    local range = 4180*mapobjectscale
    if SRB2CE.AmplifyChecker(player) then
        range = $ + 512 * mapobjectscale
    end
    local range2 = range - (64*FRACUNIT)
    searchBlockmap("objects", function(ref, found)
        if (found.flags & MF_ENEMY) then
            if R_PointToDist2(m.x, m.y, found.x, found.y) <= range2 then
                P_KillMobj(found, m, m)
                A_KhaosMobjRingBurst(found, 10, 30)
            end
        end
    end, m, m.x-range, m.x+range, m.y-range, m.y+range)

    value.fxdone = true
end}


addEffect{name = "Downside Up", chance = FRACUNIT/14, duration = 15, range = 5, sprite = "DOWNSIDEUP", etype = "neutral", nonights = true,
startsfx = sfx_s3k73, endsfx = sfx_kc65,
savereq = "fxpax2",
incompatible = {"Invisible", "the ride never ends"},
description = "You're upside down!",
func=function(player,selfid, value)
    player.mo.rollangle = ANGLE_180
end,
endfunc=function(player,selfid, value)
    player.khaos.RestoreViewAngle = true
    player.khaos.SpinOtherWay = true
    --player.mo.rollangle = 0
end}

addEffect{name = "Anti-Superring", chance = FRACUNIT/18, duration = 0, etype = "bad", sprite = "ANTIRING", startsfx = sfx_antiri, requirerings = true, nonights = true, startexception = true, noboss = true,
description="Takes away 10 rings! It takes away 20 rings in Amplify Mode!",
func=function(player,selfid, value)
    if SRB2CE.AmplifyChecker(player) then
        P_GivePlayerRings(player, -20) -- devastating
    else
        P_GivePlayerRings(player, -10)
    end
    value.fxdone = true
end}

addEffect{name = "Funky Camera", chance = FRACUNIT/23, duration = 10, range = 13, sprite = "CAMERA", etype = "bad", addition = 2,
description = "\x84\"This Lakitu is Slakituing!\"\x80 Your camera's either zoomed way too out ", description2 = "or far too close!",
nocart = true, nonights = true, nobot = true,  nomatch = true,
localsfx = true,
func = function(player, selfid, value)
    if not value.started then
        value.started = true
        local RANDY = P_RandomChance(FRACUNIT/2)

        if RANDY then
            value.CameraDist = true
            S_StartSound(player.mo, sfx_camout, player)
        else
            value.CameraDist = false
            S_StartSound(player.mo, sfx_camin, player)
        end
    end

    if SRB2CE.AmplifyChecker(player) then
        if value.CameraDist then
            player.camerascale = skins[player.mo.skin].camerascale * 7
        else
            player.camerascale = skins[player.mo.skin].camerascale / 9
        end
    else
        if value.CameraDist then
            player.camerascale = skins[player.mo.skin].camerascale * 4
        else
            player.camerascale = skins[player.mo.skin].camerascale / 5
        end
    end
end,
endfunc = function(player, selfid, value)
    player.camerascale = skins[player.mo.skin].camerascale
    if value.CameraDist then
        S_StartSound(player.mo, sfx_camin, player)
    else
        S_StartSound(player.mo, sfx_camout, player)
    end
end
}

local function IterateMaps()
    local MapTable = {}
    local MapNumber = 0
    repeat
        MapNumber = $ + 1
        if mapheaderinfo[MapNumber] then
            table.insert(MapTable, MapNumber)
        end
    until MapNumber == #mapheaderinfo

    return MapTable
end

addEffect{name = "Random Music", chance = FRACUNIT/14, duration = 0, sprite = "MUSIC", etype = "neutral", nobot = true,
description = "This changes the music to random music found in any of the levels.",
func = function(player, selfid, value)
    local musictable = {}
    local MapTable = IterateMaps()
    for key, value in pairs(MapTable) do
        if mapheaderinfo[value].musname and S_MusicExists(mapheaderinfo[value].musname) then
            table.insert(musictable, mapheaderinfo[value].musname)
        end
    end
    local RANDY = P_RandomRange(1, #musictable) -- our friend randy has returned from the war
    --S_ChangeMusic(musictable[RANDY], true, player)
    COM_BufInsertText(player, "tunes " + musictable[RANDY]) -- i don't know why but changemusic just gets overwritten by jingles, this forces it to stay.
    value.fxdone = true
end}

addEffect{name = "Flamethrower", chance = FRACUNIT/16, duration = 15, sprite = "FLAME", addition = 5, range = 3, etype = "good", nospecialstage = true,
description = "You're spewing constant fire!!! The fire won't show up in first person.",
savereq = "fxpax1",
startsfx = sfx_kc5a, endsfx = sfx_kc65, localsfx = true,
--incompatible = {"Negative Rings"},
func = function(player,selfid,value)
    if not value.started then
        value.started = true
    end

    local flamer = P_SpawnMobj(player.mo.x, player.mo.y, (player.mo.z + player.mo.height) - (player.mo.height/2), MT_FLAMEJETFLAME)
    flamer.scale = player.mo.scale
    flamer.momx = player.mo.momx
    flamer.momy = player.mo.momy
    flamer.momz = player.mo.momz
    flamer.target = player.mo
    flamer.dontdrawforviewmobj = player.mo
    flamer.flags = $ & ~MF_NOGRAVITY
    if SRB2CE.AmplifyChecker(player) then
        flamer.flags = $ & ~MF_FIRE
    end
    P_Thrust(flamer, player.mo.angle + (P_RandomRange(-10, 10)*ANG1), 40*player.mo.scale)
end}

addEffect{name = "2x Rings", amplifyname = "4x Rings", chance = FRACUNIT/13, duration = 10, sprite = "2XRING", addition = 5, range = 3, etype = "good", nonights = true, nobot = true,
noboss = true,
description="Every ring you are given is worth twice the amount! Becomes 4x in Amplify Mode!",
savereq = "fxpax1",
startsfx = sfx_kc5a, endsfx = sfx_kc65, localsfx = true,
--incompatible = {"Negative Rings"},
func = function(player,selfid,value)
    if not value.started then
        value.started = true
        value.previousrings = player.rings

        --[[if tonumber(CV_FindVar("SRB2CE_Amplify").value) then
            SRB2CE.addHUDText(player, "4x Rings")
        else
            SRB2CE.addHUDText(player, "2x Rings")
        end]]
    end

    if player.rings > value.previousrings then
        local soup = player.rings - value.previousrings
        if SRB2CE.AmplifyChecker(player) then
            soup = ($ * 3)
        end
        P_GivePlayerRings(player, soup)
    end

    value.previousrings = player.rings
end}

addEffect{name = "COMBO TIME!", chance = FRACUNIT/18, duration = 0, etype = "neutral", sprite = "PANIC", notimer = true,
description="This applies two roulettes at the same time!",
startsfx = sfx_kc48,
func=function(player, selfid, value)
    ApplyEffect(player, "Roulette")
    ApplyEffect(player, "Roulette")
    value.fxdone = true
end}

addEffect{name = "STOP!", chance = FRACUNIT/28, duration = 0, etype = "bad", sprite = "STOP", notimer = true, startexception = true,
description = "This will halt all your momentum in all directions.",
startsfx = sfx_chstop,
incompatible = {"Zero Gravity"},
savereq = "annoyingfriend",
func=function(player, selfid, value)
    A_ForceStop(player.mo)
    value.fxdone = true
end}

local function ConstructQuickTable(player)
    local Tabler = {}
    local position = 0
    repeat
        position = $ + 1
        if CheckNonConflict(player, position) then
            table.insert(Tabler, position)
        end
    until position == #KhaosEffectsTable

    return Tabler
end

addEffect{name = "Roulette", chance = 1, duration = -1, sprite = "PLACEHOLDER", nohudtext = true, notimer = true, dontrollme = true, noblink = true, dontcountme = true, noblacklist = true, removeimmediately = true, startexception = true, dontrememberme = true,
description = "This is the Roulette, given to everyone who isn't safe from the chaos. Every time",
description2 = "it's given to you, it will roll a random effect, depending on various conditions.",
func=function(player, selfid, value)
    if not value.started then
        value.QTable = ConstructQuickTable(player)
        if player.khaossp.gamblecore and tonumber(CV_FindVar("SRB2CE_Gamblecore").value) then
            S_StartSound(player.mo, sfx_lggamb, player)
        end
        value.VRoller = P_RandomRange(1, #value.QTable-1)
        value.VRollerPos = 1
        value.started = true
    end

    value.VRoller = value.QTable[value.VRollerPos]

    value.VRollerPos = $ + 1

    if value.VRollerPos > #value.QTable then
        value.VRollerPos = 1
    end

    local evalue = KhaosEffectsTable[value.VRoller]

    if player.khaosoptions.SRB2CE_Roulette_SFX and player.khaosoptions.SRB2CE_HUD then
        S_StartSoundAtVolume(player.mo, sfx_cdfm46, 80, player) -- tink sound that plays when rolling an effect
    end
    value.icon = evalue.sprite
    if not evalue.sprite then
        value.icon = "KH_NOTHING"
    end
    local ChanceTime = evalue.chance

    if player.khaos.IsLucky then
        if evalue.etype == "bad" then
            ChanceTime = $ - ($/4)
        elseif evalue.etype == "good" then
            ChanceTime = $ + ($/3)
        end
    elseif player.khaos.IsPanicking then
        if evalue.etype == "bad" then
            ChanceTime = $ + ($/5)
        elseif evalue.etype == "good" then
        ChanceTime = $ - ($/5)
        end
    end

    if player.khaossp.rdice then
        if evalue.etype == "bad" then
            ChanceTime = $ - ($/6)
        elseif evalue.etype == "good" then
            ChanceTime = $ + ($/3)
        end
    end

    if value.timer > TICRATE and not value.freeze and not evalue.dontrollme and P_RandomChance(ChanceTime) and CheckNonConflict(player, value.VRoller) then
        ApplyEffect(player, evalue.name, nil, value.hudnumber)
        if not evalue.keeprolling then -- used specifically for "PRESS X NOW"
            if player.khaossp.gamblecore and tonumber(CV_FindVar("SRB2CE_Gamblecore").value) then
                if evalue.etype == "good" then
                    S_StartSound(player.mo, sfx_gbcjac, player)
                elseif evalue.etype == "bad" then
                    S_StartSound(player.mo, sfx_dangit, player)
                else
                    S_StartSound(player.mo, sfx_ohokay, player)
                end
            end
            value.fxdone = true
        end
    end
end
}

local UniversalEffectsTable = {}
SRB2CE.UniversalEffectsTable = UniversalEffectsTable
rawset(_G, "UniversalEffectsTable", UniversalEffectsTable)

local function addUniversalEffect(tbl)
    if VERSION < 2 and tbl.nosrb2kart then
        return
    end
    table.insert(UniversalEffectsTable, tbl)
end

SRB2CE.addUniversalEffect = addUniversalEffect -- this is probably a bad idea since this hasn't been tested at all, but whatever

-- addEffect{name = "TEST!", chance = FRACUNIT/99, effect = "good", duration = 10, sprite = "PLACEHOLDER", addition = 1, func = TestFunc}
--[[
addUniversalEffect{name = "Supersonic Racing", subtext = {"Try to keep your tires right on the ground!", "There's no time to look around!", "Come on, let the fire burn!"}, hudtextprefix = "\199\201",
sounds = {sfx_cssrac, sfx_cssra2}, microsounds = {sfx_mssrac, sfx_mssra2}, chance = FRACUNIT/32,
func = function()
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            if VERSION >= 2 then
                player.invincibilitytimer = $ + 8*TICRATE
                local TwilightSparkle = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_INVULNFLASH)
                TwilightSparkle.target = player.mo
                TwilightSparkle.destscale = player.mo.scale
                P_SetScale(TwilightSparkle, player.mo.scale)
            else
                player.kartstuff[k_invincibilitytimer] = $ + 8*TICRATE
            end
        end
    end
    SRB2CE.RunFunc = nil
end
}]]


addUniversalEffect{name = "BOMBZ", subtext = {"You want it?", "It's yours my friend!"}, hudtextprefix = "\202",
sounds = {sfx_chbomb}, chance = FRACUNIT/32,
func = function()
    local randomer = 8
    if SRB2CE.AmplifyChecker() then
        randomer = 5
    end

    for mo in mobjs.iterate() do
        if mo.type != MT_TNTBARREL and P_RandomChance(FRACUNIT/randomer) then --don't do it on tnt barrels, it might be one we already spawned! best hope you don't play on a map that's mostly, if not all, TNT barrels (cursed map idea for those reading)
            local ceiling = mo.ceilingz - ((mobjinfo[MT_TNTBARREL].height + P_RandomRange(1, 3)) * FRACUNIT)
            if (mo.flags2 & MF2_OBJECTFLIP) then
                ceiling = mo.floorz + ((mobjinfo[MT_TNTBARREL].height + P_RandomRange(1, 3)) * FRACUNIT)
            end
            local tnt = P_SpawnMobj(mo.x, mo.y, ceiling, MT_TNTBARREL)
            tnt.flags2 = $1 & (mo.flags2 & MF2_OBJECTFLIP)
            tnt.flags = $ & ~MF_SOLID -- this is to prevent objects from being crushed. instead they get blown to smithereens
            tnt.khaostime = true
            S_StartSound(tnt, sfx_s3k51)
        end
    end
    SRB2CE.RunFunc = nil
end
}

--TODO: don't roll for players if it's singleplayer
addUniversalEffect{name = "The Most Attractive Player", subtext = {"Finding player..."}, hudtextprefix = "\201",
sounds = {sfx_cattra, sfx_cattr2}, microsounds = {sfx_mattra, sfx_mattr2}, chance = FRACUNIT/48, nonights = true,
func = function()
    local TheName = ""
    local RanEffect = false
    local bhid = indexOfName(KhaosEffectsTable, "Black Hole")
    local pretext = " is"

    for player in players.iterate do
        if not RanEffect and SRB2CE.ValidityCheck(player) and P_RandomChance(FRACUNIT/10) then -- small random chance just to keep it rolling a little bit, to heighten the tension
            if CheckNonConflict(player, bhid, 1, true) then
                ApplyEffect(player, "Black Hole")
            else
                ApplyEffect(player, "Effect Storage", {storeeffect = bhid})
                pretext = " will be"
            end
            TheName = player.name
            RanEffect = true
        end
    end

    if RanEffect then
        for player in players.iterate do
            if SRB2CE.ValidityCheck(player) then
                player.khaos.UniEffectSubText = "\x82" + TheName + "\x80" + pretext + " the most attractive player!"
                player.khaos.UniEffecTextBufferNumber = 0
            end
        end
        SRB2CE.RunFunc = nil
    end
end
}

addHook("ThinkFrame", function()
    if SRB2CE.TimeStop != nil then
        SRB2CE.TimeStop = $ - 1
        for mobj in mobjs.iterate() -- rip global mobj thinkers
            if mobj.timestop == nil then continue end
            if mobj.type == MT_PLAYER then continue end
            mobj.momx = 0
            mobj.momy = 0
            mobj.momz = 0
            mobj.timestop = $ - 1

            if mobj.tslaststate and mobj.state != mobj.tslaststate then
                mobj.timestop = 0
            end
            mobj.tslaststate = mobj.state

            if mobj.timestop <= 0 then
                mobj.momx = mobj.khaosmomstorex
                mobj.momy = mobj.khaosmomstorey
                mobj.momz = mobj.khaosmomstorez
                mobj.khaosmomstorex = nil
                mobj.khaosmomstorey = nil
                mobj.khaosmomstorez = nil
                mobj.timestop = nil
                mobj.tslaststate = nil
            end
        end
        if SRB2CE.TimeStop <= 0 then
            SRB2CE.TimeStop = nil
        end
    end
end)

addUniversalEffect{name = "Timestop", subtext = {"Something went wrong... you're not supposed to see this!"}, hudtextprefix = "\201",
sounds = {sfx_ctimes, sfx_ctime2}, microsounds = {sfx_mtimes, sfx_mtime2},
chance = FRACUNIT/21,
func = function()
    local timer = P_RandomRange(8, 20)
    if SRB2CE.AmplifyChecker() then
        timer = $ + P_RandomRange(3, 10)
    end

    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            player.khaos.UniEffectSubText = "Time has stopped for everything except players for " .. tostring(timer) .. " seconds!"
            player.khaos.UniEffecTextBufferNumber = 0
        end
    end

    SRB2CE.TimeStop = timer*TICRATE

    --TODO: iterate over other players and give them "khaos.spookytrails". Have them spawn a bunch of ghosts that all disappear after 5 seconds.
    for mo in mobjs.iterate() do
        if mo.type != MT_PLAYER then
            if mo.type == MT_TURRET then -- specifically to stop it from firing temporarily
                mo.flags2 = $ & ~MF2_FIRING
                mo.flags2 = $ & ~MF2_SUPERFIRE
            end
            if mo.tics > 0 then
                mo.tics = $ + timer*TICRATE
            end
            if mo.fuse > 0 and mo.fuse != timer*TICRATE then
                mo.fuse = $ + timer*TICRATE
            end
            mo.anim_duration = timer*TICRATE
            mo.timestop = timer*TICRATE
            mo.khaosmomstorex = mo.momx
            mo.khaosmomstorey = mo.momy
            mo.khaosmomstorez = mo.momz
            mo.momx = 0
            mo.momy = 0
            mo.momz = 0
        else
            local thej = P_SpawnGhostMobj(mo)
            thej.frame = $ & ~FF_TRANSMASK
            --thej.colorized = true
            --value.thej.frame = $1 | FF_TRANS10
            thej.fuse = timer*TICRATE
            if thej.tracer and thej.tracer.valid then
                thej.tracer.frame = $ & ~FF_TRANSMASK
                thej.tracer.fuse = timer*TICRATE
                --thej.tracer.colorized = true
            end
        end
    end

    SRB2CE.RunFunc = nil
end
}


addUniversalEffect{name = "Position Swap", subtext = {"What was someone doing here?", "Why were you this far behind?!?"}, requiredplayers = 2,
sounds = {sfx_cposwa, sfx_cposw2}, microsounds = {sfx_mposwa, sfx_mposw2}, chance = FRACUNIT/40,
func = function()
    --[[local PlayerPositions = {}
    -- playercount
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            table.insert(PlayerPositions, {player.mo.x, player.mo.y, player.mo.z, player.mo.momx, player.mo.momy, player.mo.momz, player.mo.angle, player.position})
            -- would really be nice to be able to swap player's current waypoints... but they specifically don't expose that currentwaypoint and nextwaypoint. why :(
        end
    end

    ShuffleInPlace(PlayerPositions)
    ShuffleInPlace(PlayerPositions)
    ShuffleInPlace(PlayerPositions)
    ShuffleInPlace(PlayerPositions) -- do it a few times to hopefully guarante everything's shuffled

    local index = 1 -- make it 1 to follow the table
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            P_SetOrigin(player.mo, PlayerPositions[index][1], PlayerPositions[index][2], PlayerPositions[index][3])
            player.mo.momx,
            player.mo.momy,
            player.mo.momz,
            player.mo.angle,
            player.position = PlayerPositions[index][4], PlayerPositions[index][5], PlayerPositions[index][6], PlayerPositions[index][7], PlayerPositions[index][8]
            player.lastpickupdistance = 0 -- in v2.3 of ring racers, this doesn't do anything. this should hopefully be fixed by an MR i made though!
            index = $ + 1
        end
    end
    S_StartSoundAtVolume(nil, sfx_s3k73, 255)
    S_StartSoundAtVolume(nil, sfx_s3k73, 255)
    S_StartSoundAtVolume(nil, sfx_s3k73, 255)

    SRB2CE.RunFunc = nil]]

    local thing = P_SpawnMobj(0, 0, 0, MT_MIXUP_ICON)
    thing.flags2 = $|MF2_DONTDRAW
    SRB2CE.RunFunc = nil
end
}
--sfx_mixup

--[[local TrapTable = {}
if VERSION >= 2 then
    TrapTable = {MT_BANANA, MT_LANDMINE, MT_ORBINAUT, MT_DROPTARGET, MT_GACHABOM, MT_SSMINE}
else
    TrapTable = {MT_BANANA, MT_ORBINAUT}
end

addUniversalEffect{name = "Trap Rain", subtext = {"Ohhhhh Banana!", "Why is it THERE?"}, nobattle = true,
nosrb2kart = true, --TODO: fix for srb2kart. does it have a waypoint system? if so, what the fuck is the waypoint object called
sounds = {sfx_ctrain, sfx_ctrai2}, microsounds = {sfx_mtrain, sfx_mtrai2}, chance = FRACUNIT/52,
func = function()
    for mapthing in mapthings.iterate do
        if mapthing.type == mobjinfo[MT_WAYPOINT].doomednum and mapthing.mobj and mapthing.mobj.valid and (tonumber(CV_FindVar("SRB2CE_Amplify").value) or P_RandomChance(FRACUNIT/3)) then
            local TheSlobject = TrapTable[P_RandomRange(1, #TrapTable)]
            if P_RandomChance(FRACUNIT/99) then
                TheSlobject = MT_SINK
            end
            local TheGlobject = P_SpawnMobj(mapthing.mobj.x + P_RandomRange(-32, 32)*FRACUNIT, mapthing.mobj.y + P_RandomRange(-32, 32)*FRACUNIT, mapthing.mobj.ceilingz-mobjinfo[MT_WAYPOINT].height, TheSlobject)
            P_InstaThrust(TheGlobject, FixedAngle(P_RandomRange(1, 359)*FRACUNIT), 5*mapobjectscale)
            if P_RandomChance(FRACUNIT/10) then
                TheGlobject.scale = $ * 2
            end
        end
    end
    SRB2CE.RunFunc = nil
end
}]]

-- mercury: 0.378, (defaultgravity/4) + (defaultgravity/8)
-- venus: 0.907, (defaultgravity/2) + (defaultgravity/5) + (defaultgravity/5)
-- earth: 1 (duh)
-- moon: 0.166, (defaultgravity/10) + (defaultgravity/15)
-- mars: 0.377, (defaultgravity/5) + (defaultgravity/6)
-- jupiter: 2.36, (defaultgravity*2) + (defaultgravity/3)
-- saturn: 0.916, (defaultgravity/2) + (defaultgravity/6) + (defaultgravity/4)
-- uranus: 0.889, (defaultgravity/3) + (defaultgravity/3) + (defaultgravity/5)
-- neptune: 1.12, defaultgravity + defaultgravity/9
-- pluto: 0.071, (defaultgravity/14)

addUniversalEffect{name = "Out Of This World", subtext = {"Oops, you shouldn't be seeing this text!"},
nosrb2kart = true,
sounds = {sfx_coutwo, sfx_coutw2}, microsounds = {sfx_moutwo, sfx_moutw2}, chance = FRACUNIT/32,
func = function()
    local defaultgravity = mapheaderinfo[gamemap].gravity or FRACUNIT/2 --+ FRACUNIT/5 + FRACUNIT/10 -- should equal 0.8, which we will then use this as the basis for our gravity calculations
    -- mapheaderinfo[gamemap].gravity
    local GravityTable = {
    {"Mercury", (defaultgravity/4) + (defaultgravity/8)},
    {"Venus", (defaultgravity/2) + (defaultgravity/5) + (defaultgravity/5)},
    {"Earth's Moon", (defaultgravity/10) + (defaultgravity/15)},
    {"Mars", (defaultgravity/5) + (defaultgravity/6)},
    --{"Saturn", (defaultgravity/2) + (defaultgravity/6) + (defaultgravity/4)},
    {"Uranus",  (defaultgravity/3) + (defaultgravity/3) + (defaultgravity/5)},
    {"Neptune", defaultgravity + (defaultgravity/9)},
    --{"Ring Racers", FRACUNIT/2 + FRACUNIT/5 + FRACUNIT/1},
    }

    if SRB2CE.AmplifyChecker() then
        table.insert(GravityTable, {"Pluto", (defaultgravity/14)}) -- hell
        if P_RandomChance(FRACUNIT/6) then -- make this far less likely, since it can break some maps
            table.insert(GravityTable, {"Jupiter", (defaultgravity*2) + (defaultgravity/3)})
        end
    end

    if gravity != defaultgravity then
        table.insert(GravityTable, {"Regular Gravity", (defaultgravity)})
    end

    local TheOne = P_RandomRange(1, #GravityTable) --haha, ha, one
    local TheName = GravityTable[TheOne][1]
    gravity = GravityTable[TheOne][2]

    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            player.khaos.UniEffectSubText = "\x86\201" + "Planetary Gravity" + "\x80\200 has been set to "+ ": " + TheName
            player.khaos.UniEffecTextBufferNumber = 0
        end
    end
    SRB2CE.GravityTimer = P_RandomRange(30, 90)*TICRATE
    SRB2CE.RunFunc = nil
end
}

addUniversalEffect{name = "Extra Life", subtext = {"You've been given an extra life!"}, gamemodeneedslives = true,
nosrb2kart = true,
sounds = {sfx_cexlif, sfx_cexli2}, microsounds = {sfx_mexlif, sfx_mexli2}, chance = FRACUNIT/150,
func = function()
    for player in players.iterate do
        player.lives = $ + 1
    end
    SRB2CE.RunFunc = nil
end
}

--[[  addUniversalEffect{name = "Insta-item", subtext = {"FOR FREE?", "It's free real estate.", "But what if I already have an item :(", "Please don't give me BAR again."}, --TODO: give player a hidden khaos effect for this instead.
sounds = {sfx_cinsit, sfx_cinsi2}, microsounds = {sfx_minsit, sfx_minsi2}, chance = FRACUNIT/18,
func = function()
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            if not player.khaos.ItsFreakingFluttershy then
                player.lastpickupdistance = 0 -- in v2.3 of ring racers, this doesn't do anything. this should hopefully be fixed by an MR i made though!
                player.khaos.ItsFreakingFluttershy = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_RANDOMITEM) -- might as well continue the MLP:FiM reference. it's freaking fluttershy
                player.khaos.ItsFreakingFluttershy.flags2 = $ | MF2_DONTRESPAWN -- just double insurance to make sure it doesn't respawn, or if it deleting somehow fails, that after pickup it goes away
                player.khaos.ItsFreakingFluttershy.target = player.mo
                if VERSION >= 2 and gametype != GT_BATTLE and P_RandomChance(FRACUNIT - FRACUNIT/4) then -- 75% chance for item, 25% chance for rings
                    player.khaos.ItsFreakingFluttershy.state = S_RANDOMITEM1
                end
                player.khaos.FlutterTimer = 0
            end
        end
    end
    SRB2CE.RunFunc = nil
end
}]]

addUniversalEffect{name = "You Get The Bees", subtext = {"Rolling effect..."},
sounds = {sfx_cbees, sfx_cbees2}, microsounds = {sfx_mbees, sfx_mbees2}, chance = FRACUNIT/13,
func = function()
    local RanEffect = false
    local StatusEffect = nil
    local StatusRand = 1
    repeat
        StatusRand = P_RandomRange(1, #KhaosEffectsTable)
        StatusEffect = KhaosEffectsTable[StatusRand]
        if P_RandomChance(StatusEffect.chance) and not StatusEffect.nobees and not StatusEffect.savereq and GlobalNonConflict(StatusRand)
        and ((not StatusEffect.nobot) or (StatusEffect.nobot and SRB2CE.botcount-1 < SRB2CE.playercount/2)) -- don't roll nobot effects if half the server is bots.
        then
            RanEffect = true
        else
            return
        end
    until RanEffect

    if RanEffect then -- just for in case the loop somehow breaks early
        for player in players.iterate do
            if SRB2CE.ValidityCheck(player) then
                if CheckNonConflict(player, StatusRand, 1) and not player.khaos.Freezer then
                    ApplyEffect(player, StatusEffect.name)
                else
                    ApplyEffect(player, "Effect Storage", {storeeffect = StatusRand, icon = StatusEffect.sprite})
                end
                local TheName = StatusEffect.name
                if SRB2CE.AmplifyChecker() and StatusEffect.amplifyname then
                    TheName = StatusEffect.amplifyname
                end
                player.khaos.UniEffectSubText = "\x82" + "Bees give everyone"+ "\x80:" + TheName
                player.khaos.UniEffecTextBufferNumber = 0
            end
        end
        SRB2CE.RunFunc = nil
    end
end
}


addUniversalEffect{name = "Negative Rings", subtext = {"Some rings have turned negative! Avoid them!", "What is wrong with you? Why are you RED?"}, nonights = true,
sounds = {sfx_crings, sfx_cring2}, microsounds = {sfx_mrings, sfx_mring2},
chance = FRACUNIT/22, func = function()
    for mo in mobjs.iterate() do
        if (mo.type == MT_RING or mo.type == MT_FLINGRING) and (SRB2CE.AmplifyChecker() or P_RandomChance(FRACUNIT/3)) then
            mo.negativering = 45*TICRATE
        end
    end

    SRB2CE.RunFunc = nil
end
}

local function NegativeRingFunction(mobj)
    if mobj.negativering != nil then
        mobj.negativering = $ - 1
        if not mobj.negativeringsetup then
            mobj.state = S_COLORRING
            mobj.color = SKINCOLOR_RED
            mobj.negativeringsetup = true
        end
        if mobj.negativering <= 1 then
            mobj.negativering = nil
            mobj.state = S_RING
            mobj.negativeringsetup = nil
        end
    end
end

addHook("MobjThinker", NegativeRingFunction, MT_RING)
addHook("MobjThinker", NegativeRingFunction, MT_FLINGRING)

local function NegativeRingsTouch(ring, plyr) local player = plyr.player
    if ring.negativering then
        P_KillMobj(ring)
        S_StartSound(ring, sfx_antiri)
        ring.shadowscale = 0
        player.rings = max(0, $ - 1)
        return true
    else
        return false
    end
end

addHook("TouchSpecial", NegativeRingsTouch, MT_RING)
addHook("TouchSpecial", NegativeRingsTouch, MT_COIN)


local TheExceptions = {"F_SKY%d",
"CHRGN%d%d", "CHRGN%d%ds",
"BOST%u%d%d",
"BOST%u%d%dS",
"BOST%u%d%dX",
"BOST%uD",
"DP_%d_M%u%d",
"SPRING%u%d",
"TRIPWIRE", "TWIRE%d%d", "VRIPWIRE", "VWIRE%d%d", "4RIPWIRE", "4WIRE%d%d",
"SPP32F%d", "SPP64F%d", "SPPING%u", "SPPING%uS",
"PITOFF", "SNOWOFF", --idk what these do, but they sound like they respawn the player, so... better safe than sorry?
"RINDRA%d%d", --i don't think this is actually used in any maps. but, maybe it is in custom maps.
}

local function Check4Exception(pic) -- slightly expensive operation, but this only runs on 1 tic so it should be fine.
    if P_RandomChance(FRACUNIT/69) then -- just to be a bit silly
        return false
    end
    for key, value in pairs(TheExceptions) do
        if string.find(pic, value) then
            return false
        end
    end
    return true
end

addUniversalEffect{name = "Wacky Textures", subtext = {"why are the walls screaming?", "Using AI, you too can make maps this beautiful!", "You forgot to install Counter-Strike: Source.", "i will force you to touch grass"}, --"i made you a sneaker panel, but i eated it"
sounds = {sfx_cwacky, sfx_cwack2}, microsounds = {sfx_mwacky, sfx_mwack2}, chance = FRACUNIT/121, func = function()
    local TheTables = {{"KH_BUP", "DEV_GA1", "BLUFLAT", "FINISH_W", "~113", "RINDRA00", "BPZFLD01", "BPZFLD04", "FFZGRSS2", "DVBR1C", "JPPI01", "SPWALG1"}, -- this just produces a bunch of invalid textures, but that looks funny so i'm keeping it. clown vomit central. ...if you break this STJR, like you rightly should, i'll cry
    {"KH_BUP"},
    {"2FORTWL1", "2FORTWL9", "APBOXW3", "ACWOODST", "CAVEF2", "ACDEBRI2", "ACZFLR02", "DCFLR14", "DSFLOOR2", "DSZRCKF1", "BSFLR2D", "MHFL4", "CASINOF2", "GRASSY", "SEGGF01", "SPACEYM", "GFZFLR02", "GFZFLR01"},
    {"CASINOF2", "GRASSY", "SEGGF01", "SPACEYM", "GFZFLR02", "GFZFLR01", "EGRIDF1", "ERLAVA1", "ERZRCKF1", "MMFLRB2", "MMFLRE", "~005", "SFLR10", "SFLR05", "SFLR01", "SFLR01B"},
    {"2FORTWL1", "2FORTWL9", "APBOXW3", "ACWOODST", "CAVEF2", "ACDEBRI2", "ACZFLR02", "DCFLR14", "DSFLOOR2", "DSZRCKF1", "BSFLR2D", "MHFL4", "CASINOF2", "GRASSY", "SEGGF01", "SPACEYM", "GFZFLR02", "GFZFLR01", "EGRIDF1", "ERLAVA1", "ERZRCKF1", "MMFLRB2", "MMFLRE", "~005", "SFLR10", "SFLR05", "SFLR01", "SFLR01B", "KH_BUP"}, -- maximum carnage
    --{"FINISH_W"}
    }

    local ListOfWacky = TheTables[P_RandomRange(1, #TheTables)]

    for sector in sectors.iterate do
        if Check4Exception(sector.floorpic) then
            sector.floorpic = ListOfWacky[P_RandomRange(1, #ListOfWacky)]
        end
        if Check4Exception(sector.ceilingpic) then
            sector.ceilingpic = ListOfWacky[P_RandomRange(1, #ListOfWacky)]
        end
    end

    for line in lines.iterate do
        if line.frontside and line.frontside.toptexture
        and Check4Exception(line.frontside.toptexture) then
            line.frontside.toptexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end
        if line.frontside and line.frontside.midtexture
        and Check4Exception(line.frontside.midtexture) then
            line.frontside.midtexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end
        if line.frontside and line.frontside.bottomtexture
        and Check4Exception(line.frontside.bottomtexture) then
            line.frontside.bottomtexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end

        if line.backside and line.backside.toptexture
        and Check4Exception(line.backside.toptexture) then
            line.backside.toptexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end
        if line.backside and line.backside.midtexture
        and Check4Exception(line.backside.midtexture) then
            line.backside.midtexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end
        if line.backside and line.backside.bottomtexture
        and Check4Exception(line.backside.bottomtexture) then
            line.backside.bottomtexture = R_TextureNumForName(ListOfWacky[P_RandomRange(1, #ListOfWacky)])
        end
    end
    SRB2CE.RunFunc = nil
end
}

--TODO: make this apply to random enemies on the map too
--[[  addUniversalEffect{name = "Ka-boom!", subtext = {"\x85" + "None of you are free from sin.", "Why did you sneeze that loud?", "bone hurting juice :(", "*Blood Explosion Spell*"}, hudtextprefix = "\202",
sounds = {sfx_ckabom, sfx_ckabo2}, microsounds = {sfx_mkabom, sfx_mkabo2}, chance = FRACUNIT/74, func = function()
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            player.khaos.UniEffecTextBufferNumber = 9001 -- forcefully sets the text buffer to the end so the text actually shows up
            if VERSION < 2 then
                local JankSplosion = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_TNTBARREL)
                JankSplosion.target = player.mo
                P_KillMobj(JankSplosion)
                S_StartSound(player.mo, sfx_s3k4e)
            end
            --P_KillMobj(player.mo)
        end
    end
    SRB2CE.RunFunc = nil
end
}]]
