--[[ SRB2:CE v3
Pong Script
]]

local function CollisionCheck(paddle, ball, value)
    local checky = false
    if ball.ignorepcollision then
        return false
    end
    if value == "player" then
        if (ball.x <= paddle.x+paddle.width)
        and (ball.x >= paddle.x) then
            checky = true
        end
    end

    if value == "bot" then
        if (ball.x+ball.size <= paddle.x+paddle.width)
        and (ball.x+ball.size >= paddle.x) then
            checky = true
        end
    end

    if checky then
        if ball.y+ball.size >= paddle.y and (ball.y <= paddle.y + paddle.height) then
            return true
        end
    end

    return false
end


addHook("PlayerThink", function(player)
    if not player.khaos then return end
    if not player.khaos.pongtime then return end
    if not player.khaospong or player.khaospong.reinit then
        --initialize
        player.khaospong = {}
        local pheight = 30
        local pwidth = 4
        local bsize = 4
        local distancefromborder = 20

        if player.khaos.chaospong then
            pheight = P_RandomRange(20, 50)
            bsize = P_RandomRange(1, 6)
            distancefromborder = P_RandomRange(10, 45)
            pwidth = P_RandomRange(1, 8)
        end

        player.khaospong["player"] = {
            y = 100 - (pheight/2),
            x = distancefromborder, -- this shouldn't change, it's just here for the renderer & logic
            height = pheight,
            width = pwidth,
        }
        player.khaospong["ball"] = { --consider making this a ring?
            y = 100 - (bsize/2),
            x = 160 - (bsize/2),
            size = bsize,
            directionx = -1, -- -1 means left, 1 means right
            directiony = 0,
            momentumx = 5,
            ignorepcollision = 0, -- ignore player collision after impact very briefly
            momentumy = 0, -- the further away from the center of the paddle on impact, the stronger this should be
        }
        player.khaospong["bot"] = {
            y = 100 - (pheight/2),
            x = 320-distancefromborder,
            height = pheight,
            width = pwidth,
            speed = 4, -- randomize this, the bigger the number the more likely they are to catch up to the ball
            direction = 0,
            delay = (player.khaos.sppong and P_RandomRange(5, 7)) or 7, -- this determines how many tics should pass til it should check where the ball currently is
        }
        player.khaospong["info"] = {
            timer = 0,
            timeshit = 0,
            playerpoints = player.khaossp.playerpoints or 0,
            botpoints = player.khaossp.botpoints or 0
        }
        player.khaospong.reinit = nil
    end
    local khaospong = player.khaospong
    local ball = khaospong.ball
    local paddles = {"player", "bot"}
    khaospong.info.timer = $ + 1

    --if timer is greater than 1*TICRATE, then we start
    if khaospong.info.timer > 35 then
        local movex = ball.momentumx
        local movey = abs(ball.momentumy) -- this became a lot harder because i didn't implement momentumy like i did momentumx
        local moveydir = (ball.momentumy < 0 and -1) or 1
        repeat -- collision checker, done in steps to prevent the ball from just phasing through the paddles
            if movex > 0 then
                ball.x = $ + (ball.directionx * 1)
                movex = $ - 1
            end
            if movey > 0 then
                ball.y = $ + moveydir
                movey = $ - 1
            end
            for key, value in pairs(paddles) do
                local paddle = khaospong[value]
                if CollisionCheck(paddle, ball, value) then
                    ball.directionx = -$
                    if player.khaos.chaospong and P_RandomChance(FRACUNIT/2 + FRACUNIT/3) then
                        ball.momentumx = P_RandomRange(3, 9)
                    end
                    ball.ignorepcollision = 3
                    khaospong.info.timeshit = $ + 1
                    local paddlecenter = paddle.y + (paddle.height/2)
                    local delta = (paddlecenter - ball.y)
                    ball.momentumy = -(delta/4) -- this would probably be better if i used a percentage of some sort
                    if khaospong.info.timeshit == 2 and ball.momentumy == 0 then -- anti-cheese
                        ball.momentumy = P_RandomRange(1, 4) * ((P_RandomChance(FRACUNIT/2) and -1) or 1)
                    end
                    S_StartSound(player.mo, sfx_chpong, player)
                    movex = 0
                    movey = 0
                end
            end
        until (movex <= 0) and (movey <= 0)
    end

    if (ball.y < 0) or ball.y + ball.size > 200 then
        ball.momentumy = -$
        S_StartSound(player.mo, sfx_chping, player)
    end

    ball.ignorepcollision = max($ - 1, 0)

    for key, value in pairs(paddles) do
        local paddle = khaospong[value]
        local poscheck = -1
        if value == "player" then
            -- check for inputs, and move the paddle accordingly.
            if khaospong.info.timer > 20 or player.khaos.sppong then
                if player.khaosbt.up then
                    paddle.y = $ - 4
                elseif player.khaosbt.down then
                    paddle.y = $ + 4
                end
            end
        else
            local paddlecenter = paddle.y + (paddle.height/2)
            local delta = (paddlecenter - ball.y)
            if (leveltime % paddle.delay) == 0 then --move the paddle
                if delta > 3 then
                    paddle.direction = -1
                elseif delta < -6
                    paddle.direction = 1
                else
                    paddle.direction = 0
                end
            end
            paddle.y = $ + (paddle.direction * paddle.speed)
        end

        paddle.y = min($, 200 - paddle.height)
        paddle.y = max($, 0)
    end

    if ball.x+ball.size < 0 then
        player.khaos.ponglost = true
    elseif ball.x+ball.size > 320 then
        player.khaos.pongwon = true
    end

    if player.khaos.sppong then
        player.pflags = $|PF_FULLSTASIS
    end

    if player.khaos.sppong and player.khaosbt.spin then
        S_StartSound(player.mo, sfx_chpggo, player)
        player.khaospong = nil
        player.khaos.sppong = false
        player.khaos.pongtime = false
        player.khaos.pongwon = false
        player.khaos.ponglost = false
        player.khaos.chaospong = false
    end

    if player.khaos.ponglost or player.khaos.pongwon then
        S_StartSound(player.mo, sfx_chpggo, player)
        if player.khaos.pongwon then
            ball.directionx = 1
            khaospong.info.playerpoints = $ + 1
            player.khaossp.playerpoints = khaospong.info.playerpoints
        else
            ball.directionx = -1
            khaospong.info.botpoints = $ + 1
            player.khaossp.botpoints = khaospong.info.botpoints
        end
        if player.khaos.sppong then
            player.khaos.pongwon = false
            player.khaos.ponglost = false
            ball.x = 160 - (ball.size/2)
            ball.y = 100 - (ball.size/2)
            khaospong.bot.delay = P_RandomRange(5, 7)

            ball.directiony = 0
            ball.momentumy = 0
            khaospong.info.timer = 0
            if player.khaos.chaospong then
                player.khaospong.reinit = true
            end
        end
        --COM_BufInsertText(player, "CESAVE 1")
    end

end)

hud.add(function(v, player)
    if not player.khaospong then return end

    local paddlerenderers = {"player", "bot"}
    local khaospong = player.khaospong

    if player.khaos.sppong then
        v.draw(0, 0, v.cachePatch("KH_BDARK"), 2<<V_ALPHASHIFT)
    end

    -- center line
    v.drawFill(160-1, 0, 2, 200, 3)

    --vertical |  |
    v.drawFill(-3, -3, 3, 200+6, 31)
    v.drawFill(320, -3, 3, 200+6, 31)

    --horizontal __
    v.drawFill(-3, -3, 320+6, 3, 31)
    v.drawFill(-3, 200, 320+6, 3, 31)

    for key, value in pairs(paddlerenderers) do
        local paddle = khaospong[value]
        local color = (value == "player" and khaospong.info.playerpoints > 9 and SRB2CE.rainbowcolors[((leveltime/2) % #SRB2CE.rainbowcolors)+1]) or 3
        v.drawFill(paddle.x-1, paddle.y-1, paddle.width+2, paddle.height+2, 31)
        v.drawFill(paddle.x, paddle.y, paddle.width, paddle.height, color)
    end

    local ball = khaospong.ball
    v.drawFill(ball.x-1, ball.y-1, ball.size+2, ball.size+2, 31)
    v.drawFill(ball.x, ball.y, ball.size, ball.size, 3)

    Fontlib.drawString(v, 5, 5, "\x80" + tostring(khaospong.info.playerpoints), {font = "KHMNT"}) -- "left"
    Fontlib.drawString(v, 320-5, 5, "\x80" + tostring(khaospong.info.botpoints), {font = "KHMNT"}, "right") -- "left"
end, "game")
