--[[

SRB2: Chaotic Edition v3

This is a from-the-ground-up rewrite of the chaos script, done for Dr. Robotnik's Ring Racers, and now backported to SRB2. This mod will refer to itself internally as "khaos" for simplicity sake.
I've made this to allow for mod support (see SRB2CE-Mod-Example-lua.txt), and cleaned up a lot of the sloppy code I made from years ago. I've overhauled how effects are applied, the HUD system, and the settings system to be more reliable, readable, and generally more netgame-safe. This is an overall improvement I hope, and should be the basis for my chaos mods for more years to come.

]]

--[[ NOTE TO SELF: ORDER OF FILES 2 LOAD:
    - customsave_io.lua
    - critical.lua
    - fx.lua
    - spcon.lua
    - rewrite.lua
    - rewrite-hud-debug.lua
    - pong.lua (can be anywhere)
]]

--[[ CHANGELOG:
    - NOTICE: This mod was made for SRB2 v2.2.13 (with some enhancements for v2.2.14 implemented), make sure you're on that version or above before complaining about bugs.
    - ANOTHER NOTICE: Chaos City might be a little laggy in OpenGL; I apologize for this, it seems I'm hitting some kind of sector limit that causes SRB2 to misbehave. You can quickly switch to Software Rendering by pressing F10!

    - You can view this changelog in the new in-game settings menu!

    - Massive rewrite has been implemented!
        * This is the rewrite I did for Tails' Khaos Kontraption, this should 500% improve stability.
        * Also included are some things I didn't do for TKK, like the faster roulette table (thus meaning your rolls will go faster)
        * This will also include expandability (i.e. mod support), so you can create your own effects that can be loaded on-top of this!

    - New single-player content!
        * A hub level has been added called "Chaos City"; interact with Amy to buy new things such as effects and such, and interact with Tails to adjust your options! Recruit more characters to expand your ways of interacting with this mod!
        * When you finish a level, you deposit all your rings into a bank, that you can then use to spend on items in Amy's shop! There's also an important (and cheap) upgrade you can get to keep your rings when damaged!
            - Because of all of this, you are now incentivized to go for as many rings as possible in a level, if you can avoid the chaos.
        * This mod *should* be compatible with other level packs, as it only replaces the tutorial level. If there's any incompatibilities, let me know, no guarantees I can fix it but I can try.

    - New effects added:
        From TKK:
        * Hey! Listen!
        * Timebomb
        * Laggy Model
        * Kids These Days
        * Hop!
        * Super Explosive 64

        Brand New:
        * STOP! (unlockable)
        * Downside Up
        * Money! (unlockable)
        * Omochao (unlockable)
        * Flamethrower
        * Broken Screen (unlockable)
        * but here's the spinner (unlockable)
        * The Treadmill Effect
        * Eye for an Eye
        * Pong
        * Rainbow Sneakers
        * :3

    - Universal Effects added, from Tails' Khaos Kontraption.
        * This is an unlockable you can buy! Comes with an assortment of effects that don't just affect the player, it affects the environment around you, and has a seperate global timer.
        * Effects are:
            ^ Timestop
            ^ BOMBZ
            ^ You Get The Bees
            ^ Extra Life
            ^ The Most Attractive Player (takes the place of "Black Hole", since it gives you it)
            ^ Out Of This World
            ^ Wacky Textures
            ^ Negative Rings (this replaces the effect of the same name)

    - New menu system introduced, mainly for single-player content, but allows for general options to be changed.
        * All the general settings have been put in here, allowing for a non-confusing way to change up the mod to your liking.
        * Blacklist system has been overhauled and put into here, being called "Toggle Effects", which can affect your ring multiplier!
            ^ This will also show descriptions for the effects, giving you a brief glance at what they do.

    - TKK specific changes, brought over to here:
        * Icons have all been revamped, and are slightly bigger in size. They should be more legible, hopefully!
        * A custom text renderer has been introduced (courtesy of l_fontlib.lua), just for a different font to be used, one custom-made for TKK and this mod.
        * The HUD text system has been overhauled, looking a lot prettier with colored borders.
        * The timer bar now accounts for your screen resolution, that is to say, it takes up the whole screen now.
        * You can disable the HUD altogether; this also disables many effect's starting sounds, meaning you won't have much idea of what effects are being applied to you!

    - Tweaked certain effects:
        * "Bouncy!" has the wall bouncing logic changed.
        * "Rewind" now resets your camera, making it slightly less disorienting. I think.
        * "Earthquake" utilizes the new 2.2.14 functions, now making it so not everyone gets the same effect, due to there being a radius and epicenter. This also means more than 1 person can have it applied at the same time.
        * "Change Character" now actually chooses random characters, instead of heavily biasing towards picking one.
        * "Nuke" has a much larger blast radius now, and comes with 0 risk of any shield shenanigans happening.
        * "Give Good Stuff" now adds to the timers instead of just replacing them.
        * "the pizza is aggressive" has had it's code ripped out and replaced; the short of it is, this will work much better now.
            ^ Also, the sounds have been upgraded to be a bit more audible.
        * "Tiny" makes you a little bit less... little. Meant to make things a little more bearable. (sorry)
        * "Change Sprite" now copies the roll and pitch of the player.
            ^ It also has a bunch of new sprites to choose from!
        * "Random Music" now looks through all the maps to get all the music tracks. So, if you're on a custom level pack, music from those levels can be selected!
            ^ It's also been changed to no longer have the music get changed back to the normal music by things that would ordinarily restore the music (i.e. 1-up jingle, invincibility, etc.)
        * "Black Hole" is now classified as a Universal Effect.
            ^ It also no longer attracts any solid objects if you're not in Amplified mode.
        * "Negative Rings" is also now a Universal Effect.
            ^ Instead of how it worked before, random rings are now turned red, and those deplete your current ring count.
        * In general a lot of other effects have had their properties changed too, there's too many changes to list here at this point :P

    - General changed things:
        * Effects are no longer preserved between levels.
            ^ Justification is that this just isn't really needed, and caused a lot more headaches programming-wise than it was worth.
        * On top of having a settings menu in Chaos City, you have access to it via "SRB2CE". Use that instead of typing out confusing commands!
        * The save system has been re-worked entirely; settings should now no longer cause "NetXCMD" issues if you were having them before.
        * The rainbow timer bar now runs slower, and has a smoother transition when looping.

    - There are enhancements exclusive for 2.2.14 (as stated above, this mod will work in 2.2.13):
        * "Ghost" and "Invisible" use the alpha tag; ergo, player smoothly transitions to an invisible state.
        * Text borders have proper transparency
        * Earthquakes now have proper radiuses; not everyone will experience them now!

    Future plans:
        * Make "Toggle Effects" work for Universal Effects
        * Give Fang a purpose besides just standing there.
        * Add some new shop items.
        * Expand Chaos City to have more stuff to do (mainly, let you enter the clock tower).
        * Re-add missing effects and functions (such as "No Abilities", and "no incompatibilities" mode)
        * Import some of the changes made to this back to Tails' Khaos Kontraption (and also fix the issues it has with SRB2Kart)
        * And, of course, add more effects.
]]

--[[ FINISH THE UPDATE ALREADY
    - I need to finish this up, this has gone on far long enough.

    - [ ] FINALIZING CHANGES TODO BEFORE RELEASE:
        - Rename the mapfile (CHOOSE THE CHRISTMAS ONE) to MAPZ0!
        - Triple check all needed textures are in, they should be!
        - Update the forum post, duh!

]]

--[[BUG FIXES & THINGS TO DO:
    - [?] When an object is in time-stop, if the current state changes, nil it's timestop properties.
        - Check the interactions with a sign-post, that's how I noticed this problem.
        - Implementing a fix doesn't fix that specific issue, but it might've fixed other issues... keep an eye on this.

    - [ ] Remove "description" and "description2", replace it with "descriptiont".
        - The spacing between items will have to be dynamic now... so if it exceeds 2, have some variable at the top to keep track of how much extra padding is needed.

    - [ ] Make a custom version of C. Fighter's Text Box Library Thing, make it use fontlib instead.
        - Unsure how easy this'll be though... plus, there's the fact it will use *two different* parsers for effects.

    - [ ] Add a new thing: If you beat a level without taking any hits, your ring multiplier goes up by 0.2x

    - [ ] Implement a proper solution to the effects that just toggle a player's variable (like ipadkid or :3)

    - [ ] Next update will bring FX Pax 3:
        - Randomized Sounds
        -

    - [ ] If an invalid effect is set to the starter effect, shit gets fucked, fix that.
        - I should check for how it interacts with loading save files in general, if it loads the wrong type of variable, it should error out properly.

    - [ ] Add a new sprite to "Change Sprite": the spinning skull.
        - maybe make it a funny unlockable, after drowning in the pool in chaos city, or getting spooked by metal sonic

    - [ ] See what happens when you black-list all effects... if it breaks things, don't allow it.
        - [ ] Maybe just cap the maximum allowed blacklisted items to like, 10. Would be a good game balance change anyways.
            - If doing that, add this condition to save verification!

    - [ ] Fix up the names of variables in the HUD, make sure they don't use the same name between each other function.
        - Also check spcon, since it has it's own HUD stuff.

    - [ ] Implement "Emergency Effect"
        - Have it be two menus, one for they key, one for the effect. Nested sub-menu.
            - Effect submenu should be the same as the starter-effect one, just for this one.
            - Give it a submenu, where you press a key and that's what it's set to.
                - Make it a generic submenu, which can call back to what thing called for it, so that it can set itself to that.

    - [?] The text effect shows the chaos mod having "nil" effects, something went wrong.
        - Disabled for now, will re-implement in next version.

    - [ ] "Eye for an Eye" does not work with Flamethrower. Maybe the player isn't being set as the afflictor?

    - [ ] It seeems exiting the level counts for getting the time attack record. Is it possible to fix that somehow?

    - [ ] Re-add "no incompatibilities" mode, maybe name it something better.
        - all hell breaks loose mode

    - [ ] Double check that all options are now in the settings menu.

    - [?] "Change Color" might not work properly if the player is super? Doesn't work in NiGHTs, at least.
        - Change Color has "nonights" added to it.

    - [ ] Have the pong machine cost less, but require Chaos Rings to play.
        - Reward the player with rings for every 5 or 10 wins, maybe.

    - [ ] Determine story elements to justify shop system... and everything.
        - "Chaos Rings" are what's collected at the end of every level. They are the pure essence of chaos found in rings.
        - Collecting enough of the Chaos Rings will eventually restore order to the world, somehow.
            - Have the tower be the place where you deposit all your unspent Chaos Rings.
        - Maybe have one final challenge to end it all off, somehow.

    - [ ] Have the "set timer for menu music" thing detect for what music is currently playing, as opposed to what level the player is in.
        - Change Music can actually pick chaos city's music, so, yeah.

    - [?] "Change Music" tried to load "MAP99M", and it couldn't be loaded. Why is that? Why did it find that, and it wasn't a valid music thing?
        - I use "S_MusicExists" now, that should hopefully fix it.

    - [ ] Implement blacklist for universal effects.

    - [ ] See if it's somehow possible to time-stop moving FOFs, and polyobjects. Might be very hard...

    - [?] Consider extending the chaos memory default.

    - [ ] New idea for Fang: Chaos Augmentations
        - Add mutators/alterations to the player that permanently affect them, which affects the ring multiplier.
        - Ideas:
            - Maybe move "Starter Effect" here.
            - Also move "Rigged Dice" here, but also make an anti-variant of it that really increases the ring multiplier.
                - In "Shoptions", have the option to choose between normal, rigged, and Evil rigged dice.
            -

    - [?] Implement statistics system for chaos effects
        - Instead of loading all the information at once, have it only load it when calling the save function.
            - It'll load in the relevant info, by adding it to your current information, then save it, and then purge that information.
            - Alternatively, make some command that handles this all locally, as to not have any net-game interference happen at all.

    - [ ] Replace the options with "options.dat", and use SRB2CE.optionstableofdoom2 to replace it all.
        - Also remove the old options load system.
        - And remove all the commands for stuff that isn't admin/local/netvar/etc.

    - [ ] Finish up things:
        - Emergency Effect
        - Double or Nothing

    - [?] Pancake is way too slow in amplify mode.

    - [ ] Are rings multiplying in multiplayer? It seems so...
        - Like, I end the level with more rings than I should.
        - Might happen after a longer play session, pay attention to the ring count before finishing a level, and what the banking system says.

    - [ ] Consider using Jana Sol's camera mod for Chaos City, particularly for that sewer secret.

    - [ ] Do validation checks on loaded items, ESPECIALLY string checks.

    - [?] Optmize saves/loads
        - Don't blank out values that are nil, and don't load them as false either (unless they're numbers).
        - Instead of using the save data table, use the SRB2CE.items2save table, and compare it to the save data being loaded.

    - [?] 2.2.14 changes to consider:
        - "tofixed" looks interesting, and according to a dev there's an inverse function too. I already made a function to convert numbers, but this might be useful still.
        - Consider using "P_DoTailsOverlay" for the fake NPCs. Or use P_DoFollowMobj, not sure which one to use.
        - Use P_IsLocalPlayer for save stuff.
        - Use P_ResetCamera to reset the camera after rewind.
            - It works like P_ResetCamera(player, cam)... how do I access the camera variables again?
        - Use input.ignoregameinputs to disable menu inputs, and use the input library in general maybe.
        - v.drawFill will get transparency support, maybe add pre-emptive support for it?
            - HUD Text will definitely need this.
        - (NOT MERGED YET) An "AddonLoaded" hook may be added; look into how I can utilize this to ensure effects get loaded properly.

    - [?] Don't allow players to go to Chaos City when in special stages, that messes things up.

    - [ ] Idea to fix "Kids These Days": Use actual screen coordinates to match the edges up.
        - Also, make it a little less severe. Just a little.

    - [ ] Get all the textures I used from the community textures compiled together. There should be some way to sort textures by which ones I already used in UZB.

    - [ ] Icons for the shop:
        - all 44x44
            FX Pax 2
            Double or Nothing
                - Two rings, but one of them is fading out (like the negative rings icon)
            Fix The Shop's Roof

            Attributions
            HUD Position
            Keep Rings On Hit

    - [ ] Re-implement the bot text feature
        - And of course, let the player be allowed to disable it.

    - [ ] Unlockables:
        - Chaos City:
            - Getting 100% unlocks and achievements will unlock a golden statue to be shown on the city map.
            - Arcade Machine becomes unlockable after spending money to fix it (shop item).
            - Restoration Station
                - Buyable thing that plonks a machine in the city to restore your character back to normal (just restores your player character)
                - Alternatively, make it a character swapper thing? Maybe just plonk the one from OLDC, if that's allowed...
                - [ ] ...maybe have Fang serve this purpose instead? I might scrap the mission system, since most of it's ideas would work better for randomizer.
            - Omochao
                - Have him behind the clock tower when you unlock him, just have him simply say "Hehe, thanks for freeing me, sucka" and have him fly away after.
        - Chaos Effects:
            - Baseball Minigame and/or Pong
        - Universal Effects:
            - The Wheels Of Fate
                - Unlocked by beating a certain amount of Fang's missions.

    - [?] Omochao doesn't work in reverse gravity.
        - I think I need to subtract the height of omochao itself? Idk?

    - [ ] Reset the gravity after 2 times the length of the universal effects timer has passed
        - [ ] PUT THE VARIABLE IN NETVAR

    - [ ] Fang Missions
        - Collect 100 rings and the goal is unlocked.
        - Defeat X amount of enemies and the goal is unlocked.
        - Find any emblem on the stage and clear it.

    - [ ] Amplify Mode isn't working in netgames, figure out why.
        - Is this even still the case?

    - [ ] See how abusable the Ring Multiplier system is, and if I need to tone things down, or just increase prices.

    - [ ] Triple-check all missing effects, from all versions of chaos.
        - Busted Controller, v2
        - No Abilities, v2
            - Make this a lot more rare, it's very annoying for what it is.
            - Also, make this just disable spin altogether, maybe custom buttons too.

    - [ ] Don't roll for "The Most Attractive Player" if the player has the effect disabled.
        - [ ] Also, check for if all players have the effect disabled. If so, don't roll this universal effect.

    - [ ] TEST ANALOG MODE, OR WHATEVER THE OTHER CONTROL SCHEME IS CALLED
        - Need to test it on effects like "but here's the spinner", with the thok specifically.

    - [?] Fix up isPlayerRespawning, since this is relevant for multiplayer.

    - [ ] Restore old behavior for bleeding; should make the player briefly invulnerable to prevent the rings being immediately re-grabbed.

    - [ ] Make "You Get The Bees" not happen when there's just a single player.

    - [?] Deny the player entry into the menus if their settings haven't fully loaded yet.
        - I don't want them to have issues saving their settings when it hasn't all loaded yet.

    - [?] Weird issue where sometimes players just don't really get hit in Match Mode. Is this the same fuckin issue as before? Check mobjdamage.

    - [?] When selecting "Retry", sometimes the player scale doesn't restore properly. Even when the player isn't currently giant/tiny when restarting. Why?
        - New system in place for removing effects for just a tic after a level starts, hopefully that fixed it?

    - [ ] Re-evaulate all "nospecialstage" effects; they were meant for Ring Racers, not SRB2.

    - [ ] Fix HUD elements that rely on gametime, if the game's paused, shit gets fucked.
        - Maybe detect if the game gets paused, and just stop all behavior that moves elements around.

    NIGHTs bugs:
        - [ ] Evaluate if "Giant" and "Tiny" should have the nonights tag.
        - [ ] Make "I WAS FROZEN TODAY!" work in nights
        - [ ] Make "No Turning" work
        - [ ] "Drifting" is completely borked.

    - [?] Change Sprite and Downside Up don't work together.

    - [?] Super Explosive 64 needs to check for walls.
        - [ ] Also, make bubbles occasionally explosive.
        - [ ] And monitors.

    - [?] When dying with "Gravity Coil", you keep the extra jump height.
        - I presume the same will apply to all things that affect player stats.

    - [ ] Remove monitors from being attracted by black hole.
        - Unless in Amplified Mode, ofc.

    - [ ] Marked for Death makes you not get hit by enemies at all.
        - Replace it's function; when you get hit, you lose much more rings than normal. Do this when the ring-keep system is implemented.

    - [?] Replace all instances of player.flashing

    - [ ] Colors are definitely not 100% accurate in l_fontlib. Fix that.

    - [?] If you die with a text box, you just get stuck.
        - Maybe fixed now that I've redone the respawn checks?
]]


--[[ Effects to do:
missing from chaotic edition: (marked with [p] means it's in the previous version of this script)
    - Busted Controller

]]


--[[ FX Ideas

    - [ ] Marked for Death
        - Re-implement this, have this make it so every time you take damage, you lose way more rings than usual.

    - [ ] Hell Of Mirrors
        - This is like "Wacky Textures", but all the textures on the walls get removed. Floor and ceiling textures stay though.
        - Ideally this could be something I can reverse somehow...
            - I don't think maps can make new linedefs after the start, so everything should remain the same, so a static table should be fine.
            - Make it like local restoretextures = {{[uppertexture],[midtexture],[lowertexture]},etc.}

    - New universal effect idea: Colormap hell
        - Change all the sector's colormaps to be random ones
        - Need to figure out how it's done for binary, should be easy for UDMF though, I think...

    - [ ] Fish Mode
        - You drown in air, and you can breathe in water (or space, i guess, it'll just reset that timer)
        - Run a detection for any water in the map; if none is found, this effect won't run.
            - Or maybe not, considering the next thing.
        - Rings restore your "air" timer, by a quarter of a second each (so, TICRATE/4)

    - [ ] Pop-Ups
        - Old IE style pop-ups show up, you have to hold jump to close each one.
        - The pop-ups vibrate more violently as you hold jump, maybe they shatter?
        - Multiple ones can show up at the same time, use a table to handle this

    - [ ] They Get The Bees
        - All enemies get certain chaos effects applied to them. I'll need to make a whole other system for this somehow...
        - It'll definitely be a limited subsection of effects, things like "Bleeding" will work, or "Flamethrower", etc.
        - Have certain effects have a "theymobj" toggle; if this is true, then it'll allow the effect to be ran on enemies.
            - Will need to make any effect that does this account for mobj's instead of players. Maybe instead of supplying a player variable, they get their own mobj value?

    - [ ] Hide the HUD
        - Find all the variables that refer to the HUD enable convar, and just have it also check for "player.khaos.HideHUD" or something.
            - Make it a function?
        - Also, add a new thing to "addEffect" that allows the override tag to be applied to effect names, like this.

    - [ ] Resizer
        - Scales random objects around the player.
        - Unsure how to go about this w/o lag, but I'd like there to be a timer that decrements if the object has been scaled.
        - Also, maybe a universal effect?

    - [ ] Noclip
        - Very brief and very rare effect where the player gets to noclip.
        - Icon is just the gmod icon, why not.
        - Use t-posing character sprites if they exist, that'd be really funny.
            - I know amy has a sprite for that... what is it called?
        - If the player is found to not be in bounds shortly after the effect ends, the player dies.
            - player.pflags = $|PF_NOCLIP
            - have to turn this off on end function

    - [ ] It's Goop Time
        - Turn all Water FOFs into Goop.
            - Only run this on maps with water; do a pre-check first. If the pre-check is a success, store it all in a table so we don't have to do it twice.
        for sector in sectors.iterate do
            for rover in sector.ffloors() do -- mmm love this
                if rover and rover.valid then
                    if rover.flags & FF_SWIMMABLE then
                        rover.flags = $1 & ~FF_GOOWATER
                    end
                end
            end
        end

    - [ ] Amy Invasion (universal effect)
        - Make a mobjthinker for MT_ROSY, have it only work if it has the "khaostime" tag applied.
        - Search the map for players in a much bigger radius.
            - Alternatively, do the same thing as "BOMBZ", just find all objects, and have a random chance to spawn an amy around that object.
        - Also increase the speed, somehow.
        - Also do noclip, so she can go anywhere.

    - [ ] Taking what you love
        - Remove many useful objects in the nearby area, like existing air bubbles (not the spawners... unless in amplify :3c), rings, monitors, etc.

    - [ ] Insta-Deton
        - Enemies turn into detons when destroyed

    - [ ] Object Scramble
        - Objects are turned into different ones, like enemies, hazards, etc.

    - [ ] Anti-Tiktok
        - It's "Kids These Days", but the blacked out sides are reversed. Center of the screen is blacked out now.
        - Amplify only effect?

    - Baseball
        - A giant baseball rolls across the screen, taking up the screen height. It happens at random intervals.

    - [ ] BEES
        - Directly ripped from Reelism 2, because I'm shameless. Might even rip the DKC sprites just because.
        - Not to be confused with "you get the bees"
        - Covers your character in bees! Meant to be a visual distraction for the player, but can also hinder other player's visions too since these will be a bunch of bees objects.

    - [ ] Carbonated Poop Water (universal effect)
        - play the funny sound from the tiktok
        - Check the map if water FOFs exist, if it does, make all the water brown.
            - ...though I should check if they don't exist, because if they don't, then there's no point in running this.
        - Also, player drains rings while in water now.
        for sector in sectors.iterate do
            for rover in sector.ffloors() do
                if rover and rover.valid then
                    if rover.flags & FF_SWIMMABLE then
                        --rover.toppic, rover.bottompic
                        --??
                    end
                end
            end
        end

    - The Helicopter is now avaialble.
        - like that one simpleflips video
        - after a bit, it shows "The Helicopter is no longer available"

    - Make "Bouncy!" be "Bouncy" again
        - Use (player.mo.eflags & MFE_JUSTBOUNCEDWALL) for the check.
        - For SRB2, use this: https://mb.srb2.org/addons/wall-check.5085/
            - (this'll also be needed for super explosive 64)

    - Rat Swarm
        - You are swarmed in rats that will protect you.
            - Kinda like super flickies, but Not
        - Use S_FLICKY_12_AIM from SRB2.

    - So Many Colors...
        - Effect that displays multi-color bands on the player's screen.
        - Need transparency to do this; additionally, try to see if I can use additive color.

    - Water Walker
        - does what is says on the tin
        - spawn a solid mobj that floats right on the surface of water, tracking the x and y of the player, thus stopping the player from going underwater

    - Wacky Item Roulette
        - Next item roulette the player gets, will have stupid busted items.
        - Have a table with all the possible sets.
        - ...also figure out how the roulette is set up internally.

    - Instantly Use Shield Ability
        - Makes you instantly use your shield ability, whether you want to or not.

    - Bullet Time
        - This just slows down the player compared to everyone else. All actions are slower.
        - Would be a bitch to do though, I'd need to slow down *all* the animations somehow.
            - Is it possible to update the mobj tics, that reference a state's tics? That'd be enough to do it I think...

    - Press Spin to Crunch
        - Press the spin dash button and it squishes the player, causing damage but doesn't put the player in a pain state.
        - Make a squish animation
        - Use the ring spill script I made awhile ago.
        - There should be a function/action to make the player play the pain sound, use that for custom character compatibility.

    - Randomized Sounds
        - Re-do this from the ground up
        - Or just look at this again? I used it for MMT, so...
            - https://mb.srb2.org/addons/personal-sounds.2924/
        - Sound functions are local to the player only, avoid trying to base too much code off that.
            - Have it cycle through a list of sounds, if it detects a sound, it just reads the current cycle and replaces that sound.
            - Maybe have the table randomized on start, but other than that, that should work. Other players probably won't hear the same sounds, but oh well

    - The Wheels Of Fate
        - Universal effect that applies to everyone.
        - This spins up 3 slots that will give the player certain minor attributes.
            - This can be giving rings, taking away rings, buffed speed, or certain items.
        - Use the same roulette noises as the one from the Egg Dealer in shadow the hedgehog
            - Noise lasts 89.6 tics, play the loop after that, which lasts 23.5 tics. Round down.

    - Scramble
        - Everyone trades items.
        - And khaos effects.

    - Randomize Followers

    - Make Followers make sounds at random
]]

--[[ FEATURES TO ADD:
    - [ ] Add the missing effects:
        - Busted Controller

    - [ ] Make the spawned turret instantly fire.
        - Set the duration (tics) to a very low number, like, 5.

    - [ ] Nuke needs more visual flair.

    - [ ] Re-do "the pizza is aggressive" entirely. Just simplify it.

    - [ ] Use the new music functions to properly restore the music when using the menus.
        - Things like S_GetMusicPosition() and S_SetMusicPosition(int position)

    - [ ] Add new blacklist commands for Universal Effects.

    - [ ] Show Universal Effects in "SRB2CE_List_Effects", and have the descriptions show up in "SRB2CE_What_Is" too.

    - [ ] Add an option to set a y offset to the HUD. Good for CRTs.

    - [ ] Revamp the nonconflict checks for effects, and add a similar system for Universal Effects
        - Use a table to compare things, and have there be a few different types of comparisons done.
        - This is to allow people to add their own compatibility checks into their mods.
        - Won't be done for v1.0, maybe for a v1.1 update?

    - [ ] Admin option to disable user blacklists.

    - [ ] Consider using "CV_StealthSet" instead of running an entire console command.
        - Is it netplay safe though? Biggest question.

    - [ ] Re-add command to set effects they can immediately apply to themselves.
        - [ ] But also, have an admin command that disables this. Or rather, enables it, since it's disabled by default.
            - When a player tries to use it, have it print to the console that they can't use it.

    - [ ] Add a textbox that dynamically shows your current character's name.
        - And maybe also your profile name, that'd be funny too.

    - [ ] Don't apply Bleeding when you have the Lightning Shield.
        - Make a parameter for this, "no magnet shield"

    - Option to randomly enable Amplify Mode for a round.

    - Have mods that interface with this require that they specify what mod they are
        - Can do stuff with this later, just make this a requirement.
        - Also, make all internal effects use "base" as the set mod.
            - Stop mods from using "base" as it's mod.

    - Make the timer bars glow a little bit when they are about to reach the end.
        - It'd be like, the last 80% of the bar.
        - Unable to do this for the rainbow bar, but whatever.

    - Have "Wacky Textures" have a chance to instead swap around all the textures on a course.
        - Would make things more visually interesting.
        - Re-use the code done for position swap

    - Restore the player-table roulette system to discard incompatible effects
        - Faster rolls can be done like this.
        - Only store numbers to make sure things stay quick and simple.
        - Fall back on the original table if the player table doesn't currently exist.
            - Alternatively, store the table inside the roulette effect itself.

    - Add text boxes that references 5secondfilms

    - Look through the ghostmobj function to see everything that needs to be copied over for effects that hide the player and replace it with a different object.
        - I don't know what "standingslope" is but it doesn't seem to like being set directly.

    - Change up the SFX for a lot of effects.
        - Make "Moonjump" play a sound when the key combo is pressed.

    - Have "nuke" still affect you if someone else got it at the same time as you.
        - Makes rolling for it in "you get the bees" more chaotic.
        - If you get hit by something that has you as the target, then ignore the damage. Else, continue on with mobjdamage (return false?)

    - Make dodging a hit with "Lucky!" play the Lucky sound from Paper Mario, and use the text from there too.
        - Might be a bit much actually, considering the sound...

    - "the pizza is aggressive" should use mapobjectscale instead of FRACUNIT.

    - [?] Have the text box show the button prompt for accelerate.
        - [ ] Uncomment this and fix the position, maybe change up the text box UI to accomodate for the button.

    - Maybe come up with a different name for "Universal Effects", one that's more fitting when there's just 1 player.
        - Would make it more convienent when I backport all this back into SRB2 Chaotic Edition

    - Have effects that would apply multiple times instead just add to the timer.
        - Include exceptions of course, or do the inverse and make it a "add2timer" variable.
        - In the non-conflict check, make it pass the already exists thing if it plans to add to the timer.
            - And then, on SRB2CE.ApplyEffect, check for the conditions, and instead find that effect to apply the additional timer to. Make the HUD say "Extended timer for $EFFECT"

    - [?] Now that everything is stored in a global table variable... allow mods to interface with this. As in, add effects.
        - [?] Have a seperate "addEffect" function, that works to facilitate mods better.
            - Disallow this from adding effects with the same name.
            - Check for valid sounds.
        - Also have a seperate "addUniversalEffect" thing too. And make Universal effects a global table.
        - Get a function that adds incompatibilities to effects.

    - Fix all effects that use pw_powers for flashing, replace it with "player.flashing" instead.

    - Add icon next to the universal effects text, show it actively rolling for effects too, that'd be fun.

]]

local ueTimer = 0
local ueTimerBuffer = 0
local ueTimerMax = 69 * TICRATE -- DON'T CHANGE THIS!

local initiateUE = false
local hasanyonefinished = false
local mapobjectscale = mapobjectscale or FRACUNIT -- SRB2 might get this in the future, so we're doing this for in case

local PreviousUniversalEffect = "" -- done to prevent the same effect from being applied again
SRB2CE.PreviousUniversalEffect = PreviousUniversalEffect

SRB2CE.botcount = 0
SRB2CE.playercount = 0

local function isPlayerRespawning(player) -- TODO: fix me for SRB2
    --[[if VERSION >=2 and player.respawn.state then
        return true
    end

    if VERSION < 2 and player.kartstuff[k_respawn] then
        return true
    end]]

    if player.playerstate != PST_LIVE then
        return true
    end

    return false
end

local function indexOfName(array, value) --thanks to amperbee for telling me the obvious that i can just modify this function lol
    for i, v in ipairs(array) do
        if v.name == value then
            return i
        end
    end
    return nil
end

addHook("PlayerSpawn", function(player)
    if not player.khaos then
        player.khaos = {}
        player.khaosmemory = {}
        player.khaoseffects = {}
        player.khaostext = {}
        player.khaoshud = {}
        player.khaosHUDNumber = 1 -- seperate value, as to not get reset by anything hopefully
    end

    player.khaos.IsLucky = false
    player.khaos.DodgeHit = false
    player.khaos.starteffectdone = false
    player.khaos.Check4Hit = 0

    if not player.khaosoptions then -- keep seperate to not have this reset on map change
        player.khaosoptions = {}
        player.khaos.OptionsLoadAmount = 0
        --TODO: move all this to the spcon loading settings thing instead.
        player.khaos.OptionDelay = (netgame and 3*TICRATE) or (not netgame and 11)
        for key, value in pairs(SRB2CE.optionstableofdoom) do
            if not value.localonly then
                local TheGame = value.name --you lost it
                local TheMame = CV_FindVar(TheGame).defaultvalue
                if value.typer == "number" then
                    TheMame = tonumber($)
                end
                player.khaosoptions[TheGame] = TheMame
            end
        end
    else
        player.khaos.OptionDelay = -1
        player.khaos.OptionsTextConfirm = true
        player.khaos.OptionsLoaded = true
    end


    if player.khaos.OptionsLoaded and not player.khaossp.visitedcity and not netgame and not mapheaderinfo[gamemap].srb2cecity then
        SRB2CE.addHUDText(player, "\201\199" + "Chaos City" + "\200 \x80" + "has shops for upgrades, and lets you change settings!", true, 8)
        SRB2CE.addHUDText(player, "\201" + "Hey!" + "\200 \x80" + "Check " + "\199" + "the warp" + "\x80 at spawn to go to Chaos City!", true, 8)
    end

    player.khaos.timer = 0
    player.khaos.WaitTime = 0
    player.khaos.failsafetimer = 0
    player.khaos.addHUDTextRando = 0
    player.khaos.HUDMoveUp = 0
    player.khaos.TextPopUpTimer = 0
    player.khaos.TextPopUpFade = 0
    if not SRB2CE_Enabled then
        player.khaos.TextPopUpTimer = 6969
        player.khaos.TextPopUpFade = 6969
    end
    player.khaos.UniPopUpTimer = 0
    player.khaos.UniEffectSubTextBuffer = ""
    player.khaos.UniEffectTextBuffer = ""
    player.khaos.UniEffecTextBufferNumber = 1
    player.khaos.UniTextPopUpFade = 0
    player.khaos.TextBoxAdvancePage = 0
    player.khaos.TextBoxFader = 0
    player.khaos.Tumbler = 0

    if player.mo then -- match mode doesn't like it when i access mo, smh
        player.normalspeed = skins[player.mo.skin].normalspeed
        player.runspeed = skins[player.mo.skin].runspeed
        player.actionspd = skins[player.mo.skin].actionspd
        player.thrustfactor = skins[player.mo.skin].thrustfactor
        player.mindash = skins[player.mo.skin].mindash
        player.maxdash = skins[player.mo.skin].maxdash
        player.camerascale = skins[player.mo.skin].camerascale
        player.khaos.RewindX = player.mo.x
        player.khaos.RewindY = player.mo.y
        player.khaos.RewindZ = player.mo.z
        player.khaos.RewindAngle = player.mo.angle
        --player.mo.flags2 = $ & ~MF2_DONTDRAW
        if not player.khaos.OGFriction then
            player.khaos.OGFriction = player.mo.friction
        end
    end

    player.khaos.InvertControlsNow = nil
    player.khaos.BorkedSteering = nil

end)

--[[addHook("ThinkFrame", function()
    if (leveltime % TICRATE == 0) then
        --print("hi")
        --print(KhaosEffectsTable[1])
        --print(KhaosEffectsTable[1].name)
    end
end)]]


--[[
    addHook("PlayerCmd", function(player, ticcmd)
        if not player.khaos then return end
        if player.khaos.BorkedSteering then -- i think this fucking crashes SRB2?
            ticcmd.turning = $ + player.khaos.BorkedSteering
        end
        if player.khaos.InvertControlsNow then
            ticcmd.turning = -$
        end
    end)
]]


addHook("PreThinkFrame", function()
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            if player.khaos.FrozenNow then
                if ((tonumber(player.khaos.postfowardmove) != tonumber(player.cmd.forwardmove)) or
                (tonumber(player.khaos.postsidemove) != tonumber(player.cmd.sidemove)))
                and player.khaos.FrozenWait == 0 then
                    player.khaos.FrozenShake = true -- isn't that just a slushie?
                end
                player.khaos.postfowardmove = player.cmd.forwardmove
                player.khaos.postsidemove = player.cmd.sidemove
                player.cmd.buttons = $ & ~BT_JUMP
                player.cmd.buttons = $ & ~BT_SPIN
            end
            if player.khaos.InvertControlsNow then
                player.cmd.forwardmove = -($)
                player.cmd.sidemove = -($)
            end
            if player.khaos.BustedController then
                if player.khaos.holdbutton then
                    player.cmd.buttons = $ | player.khaos.holdbutton
                    player.khaos.holdtime = $ - 1
                    if player.khaos.holdtime <= 0 then
                        player.khaos.holdbutton = nil
                    end
                else
                    if P_RandomChance(FRACUNIT/5) then
                        if P_RandomChance(FRACUNIT/2) then -- or not P_IsObjectOnGround(player.mo)
                            player.khaos.holdbutton = BT_JUMP
                            SRB2CE.DebugPrint(player, "press jump")
                        else
                            player.khaos.holdbutton = BT_SPIN
                            SRB2CE.DebugPrint(player, "press spin")
                        end
                        player.khaos.holdtime = P_RandomRange(2, TICRATE - TICRATE/4)
                        SRB2CE.DebugPrint(player, "duration: " .. player.khaos.holdtime)
                    end
                end

                if P_RandomChance(FRACUNIT/3) then
                    if P_RandomChance(FRACUNIT/2) then
                        player.cmd.forwardmove = -50
                    else
                        player.cmd.forwardmove = 50
                    end
                end

                if P_RandomChance(FRACUNIT/15) then
                    player.mo.angle = (player.cmd.angleturn << FRACBITS) + (P_RandomRange(-60, 60)*ANG1)
                end
            end
        end
    end
end)

addHook("PostThinkFrame", function()
    for player in players.iterate do
        if SRB2CE.ValidityCheck(player) then
            if player.khaos.GhostNow then  -- thanks to Faye on the SRB2 discord for discovering how to get MF2_SHADOW working
                player.mo.flags2 = $ | MF2_SHADOW
            end
            if player.khaos.GhostBegone then
                player.mo.flags2 = $ & ~MF2_SHADOW
                player.khaos.GhostBegone = false
            end

            if player.khaos.InvisibleNow then
                player.mo.flags2 = $ | MF2_DONTDRAW
            end

            if player.khaos.InvisibleBegone then
                player.mo.flags2 = $ & ~MF2_DONTDRAW
                player.khaos.InvisibleBegone = false
            end

            if player.khaos.FrozenNow and not player.khaos.ChangeSpriteNow and player.khaos.FrozenState and not isPlayerRespawning(player) then
                player.mo.state = player.khaos.FrozenState
                player.mo.frame = player.khaos.FrozenFrame
                player.nocontrol = 1
                if VERSION < 2 then
                    player.powers[pw_nocontrol] = 1
                end
                if player.followmobj
                and player.khaos.FollowFrozenState and player.khaos.FollowFrozenFrame then -- redundancy beacuse script errord during testing for some unknown reason
                    player.followmobj.state = player.khaos.FollowFrozenState
                    player.followmobj.frame = player.khaos.FollowFrozenFrame
                end
            end

            if player.khaos.FrozenShake and player.khaos.FrozenWait > 2 then
                player.mo.spritexoffset = P_RandomRange(-5, 5) * FRACUNIT
                player.mo.spriteyoffset = P_RandomRange(-5, 5) * FRACUNIT
            end
        end

        if player.khaos.GhostNow then  -- thanks to Faye on the SRB2 discord for discovering how to get MF2_SHADOW working
            player.mo.flags2 = $ | MF2_SHADOW
        end
        if player.khaos.GhostBegone then
            player.mo.flags2 = $ & ~MF2_SHADOW
            player.khaos.GhostBegone = false
        end

        if player.khaos.InvisibleNow then
            player.mo.flags2 = $ | MF2_DONTDRAW
        end

        if player.khaos.InvisibleBegone then
            player.mo.flags2 = $ & ~MF2_DONTDRAW
            player.khaos.InvisibleBegone = false
        end
    end
end)

addHook("PlayerThink", function(player)
    if tonumber(player.khaos.FlutterTimer) then
        player.khaos.FlutterTimer = $ + 1
        if player.khaos.FlutterTimer >= 2 then
            if player.khaos.ItsFreakingFluttershy and player.khaos.ItsFreakingFluttershy.valid then
                P_KillMobj(player.khaos.ItsFreakingFluttershy)
            end
            player.khaos.ItsFreakingFluttershy = nil
            player.khaos.FlutterTimer = nil
        end
    end

    for index, value in ipairs(player.khaostext) do
        -- starting position, then add 10 for each value
        value.timer = $ + 1

        if (value.expire - value.timer) > 1*TICRATE then
            value.flags = 0
        else
            if (((value.timer-2*TICRATE)*2) % 5 == 0) then
                value.trans = $ + 1
            end
            value.flags = value.trans<<V_ALPHASHIFT
        end

        if value.timer > value.expire then
            table.remove(player.khaostext, index)
            player.khaos.HUDMoveUp = $ + 9
        end
    end

    if player.khaos.HUDMoveUp > 0 then
        player.khaos.HUDMoveUp = $ - 2
    end

    if player.khaos.HUDMoveUp < 0 or #player.khaostext == 0 then -- the latter shouldn't happen, but who knows if the screen gets flooded with text
        player.khaos.HUDMoveUp = 0
    end

    if player.khaos.addHUDTextRando > 0 then
        player.khaos.addHUDTextRando = $ - 1
    end

    if player.khaos.OptionDelay > -9 then
        player.khaos.OptionDelay = $ - 1
    end

    if player.khaos.OptionDelay <= 0 and not player.khaos.OptionsLoaded and player.khaossp.loadedsettings and player.khaossp.blacklistloaded then
    --TODO: move this to spcon, and remove all the console variables that aren't local
        if not player.bot then
            Khaos_LoadSettings(player)
            SRB2CE.DebugPrint(player, "LOADING KHAOS SETTINGS")
            if netgame then player.khaos.OptionDelay = $ + 5 end
            player.khaos.OptionsPreviousLoad = player.khaos.OptionsLoadAmount
            --print("ATTEMPT LOAD")
        else -- player is bot, or we're in srb2kart, don't try to load anything
            player.khaos.OptionsLoaded = true
            player.khaos.OptionsLoadAmount = 1
        end
    end

    --[[if player.khaos.OptionDelay <= -4 and not player.khaos.BLOptionsLoaded then
        if not player.bot or VERSION < 2 then
            LoadBlacklistSettings(player)
        else -- player is bot, or we're in srb2kart, don't try to load anything
            player.khaos.BLOptionsLoaded = true
        end
    end]]

    if player.khaos.OptionsLoaded and not player.khaos.OptionsTextConfirm and player.khaos.OptionDelay <= -8 then -- need to wait just a little bit for the slow-ass command to actually change the load failed value
        if player.khaos.LoadFailed or player.khaos.LoadFailedSP then
            SRB2CE.addHUDText(player, "\x85" + "Error loading options!\x80 Check the\x82 console\x80 &\x82 luafiles\x80 folder.", true, 4)
            S_StartSound(player.mo, sfx_adderr, player)
        else
            if player.khaos.OptionsLoadAmount > 0 then
                if not player.khaossp.visitedcity and not netgame and not mapheaderinfo[gamemap].srb2cecity then
                    SRB2CE.addHUDText(player, "\201\199" + "Chaos City" + "\200 \x80" + "has shops for upgrades, and lets you change settings!", true, 8)
                    SRB2CE.addHUDText(player, "\201" + "Welcome!" + "\200 \x80" + "Check " + "\199" + "the warp" + "\x80 at spawn to go to Chaos City!", true, 8)
                else
                    SRB2CE.addHUDText(player, "\199" + "Loaded SRB2:CE save data!", true)
                end
                if netgame then S_StartSound(player.mo, sfx_gshc4, player) end --sfx_tmxsuc
            else --newbie, try to tell them to look at the SRB2CE_Help command. sucks to be a controller user i guess
                --SRB2CE.addHUDText(player, "And be sure to go to '" + "\199" + "Chaos City" + "\x80' some time!", true, 8)
                if not netgame then
                    SRB2CE.addHUDText(player, "\201\199" + "Chaos City" + "\200 \x80" + "has shops, and let's you change settings!", true, 8)
                    SRB2CE.addHUDText(player, "\201" + "Welcome!" + "\200 \x80" + "Check " + "\199" + "the warp" + "\x80 at spawn to go to Chaos City!", true, 8)
                else
                    SRB2CE.addHUDText(player, "Welcome to" + "\201 SRB2: Chaotic Edition\200!", true, 8)
                end
                --TODO: if they fix addon settings to work everywhere else, replace this text!
                S_StartSound(player.mo, sfx_addfil, player) --sfx_tmxsuc
                S_StartSound(player.mo, sfx_addfil, player)
                --S_StartSound(player.mo, sfx_gshc4, player)
                --S_StartSound(player.mo, sfx_gshc4, player)
            end
        end
        if SUBVERSION < 13 and VERSION == 202 then
            SRB2CE.addHUDText(player, "\x85" + "WARNING:\x80 You are not on the latest version of SRB2! Please update!", true, 8)
        end
        player.khaos.OptionsTextConfirm = true
    end

    if player.khaossp.starteffect and player.khaossp.starteffectchosen and not player.khaos.starteffectdone and player.khaos.timer > (TICRATE/2) then
        --TODO: either do save validation or check the effect here to make sure players aren't applying effects they shouldn't! the value to check is "startexception"
        SRB2CE.ApplyEffect(player, player.khaossp.starteffectchosen)
        player.khaos.starteffectdone = true
    end


    if not SRB2CE.ValidityCheck(player) return end -- if turning playerthink into a thinkframe hook, be sure to change this around too
    player.starpostscale = mapobjectscale -- this might break maps, but i can't think of literally any other way to fix this
    local Freezer = false

    if leveltime < 2*TICRATE or player.exiting or (player.pflags & PF_FINISHED) or stoppedclock or player.playerstate == PST_DEAD then -- freeze all effects if the level is finished, or if the player Fucking Dies
        Freezer = true
        player.khaossp.warpcooldown = (1*TICRATE/2)
    end

    player.khaos.Freezer = Freezer

    if player.khaos.WaitTime == 0 then
        player.khaos.WaitTime = P_RandomRange(TICRATE*3, TICRATE*10) -- TODO: replace with "minwait" and "maxwait" settings
        if player.khaos.IsLucky then
            player.khaos.WaitTime = $ + P_RandomRange(0, TICRATE) -- adds an extra second if lucky
        end
    end

    if not Freezer then
        player.khaos.timer = $ + 1
    end

    if (player.khaos.timer > player.khaos.WaitTime) then --activate roulette
        SRB2CE.ApplyEffect(player, "Roulette")
        player.khaos.timer = 0
        player.khaos.WaitTime = 0
    end

    --[[if not player.bot and player.khaos.Check4Hit > 0 then
        if P_PlayerInPain(player) or player.playerstate == PST_DEAD then
            player.khaossp.timeshitbykhaos = $ + 1
            player.khaos.Check4Hit = $ + 5
            SRB2CE.DebugPrint(player, "PLAYER GOT HIT! HAH!")
        end
        --print(player.khaos.Check4Hit)
        player.khaos.Check4Hit = max(0, $ - 1) -- not really nessecary, i just like to be safe
    end]]

    if P_PlayerInPain(player) or player.playerstate == PST_DEAD then
        player.khaossp.timeshitbykhaos = ($ and $ + 1) or 1 -- there's a chance this could error if you delete your save file, dunno why it's random tho
    end

    if CV_FindVar("SRB2CEDebug").value then
        if (player.cmd.buttons & BT_CUSTOM1) then
            if not player.khaos.holdabutton then
                for mo in mobjs.iterate() do
                    if mo.tics >= 0 then
                        mo.tics = 5*TICRATE
                    end
                end
            end
            player.mo.rollangle = $ + ANGLE_180
            --player.pflags = $|PF_NOCLIP
            player.khaos.holdabutton = true
        else
            player.khaos.holdabutton = false
        end

        if (player.cmd.buttons & BT_CUSTOM3) then
            local Gay = "PRESS X NOW"
            --COM_BufAddText(player, "TESTER")
            if not player.khaos.holddebugbutton and SRB2CE.CheckNonConflict(player, indexOfName(KhaosEffectsTable, Gay), 3) then
                SRB2CE.ApplyEffect(player, Gay)
            end
            player.khaos.holddebugbutton = true
        else
            player.khaos.holddebugbutton = false
        end


        --[[if (player.cmd.buttons & BT_CUSTOM3) then
            local Gay = "Change Color"
            if not player.khaos.hold3rdbutton then
                for p2 in players.iterate do
                    if SRB2CE.ValidityCheck(player) then
                        SRB2CE.ApplyEffect(p2, Gay)
                    end
                    --SRB2CE.DebugPrint(player, "playerstate: " .. player.playerstate)
                end
            end
            player.khaos.hold3rdbutton = true
        else
            player.khaos.hold3rdbutton = false
        end]]
    end

    --[[if player.cmd.buttons & BT_DRIFT and player.cmd.buttons & BT_BRAKE and player.cmd.buttons & BT_ACCELERATE and player.cmd.buttons & BT_ATTACK
        print(players[0].khaosoptions.SRB2CE_Memory)
    end]]

    if isPlayerRespawning(player) then
        player.khaos.failsafetimer = $ + 1
    else
        player.khaos.failsafetimer = 0
    end

    -- [[ -- Effect Logic -- ]] --
    for index, value in ipairs(player.khaoseffects) do
        local effecter = KhaosEffectsTable[value.effect]

        if effecter.func and not Freezer and not value.removeme then
            effecter.func(player, index, value)

            if effecter.drainrespawntimer or (not isPlayerRespawning(player) or player.khaos.failsafetimer > 5*TICRATE) then -- don't decrease timer during respawn, unless the player's been respawning for > 5 seconds
                value.timer = $ + 1
            end
            if player.khaos.failsafetimer > 9*TICRATE then -- player seems to be taking awfully long to respawn, double the timer speeds to get rid of effects quick
                value.timer = $ + 1
            end
        end

        if isPlayerRespawning(player) and effecter.restartrespawn then
            value.started = false
        end

        if not value.removeme and ((value.duration > 0 and (value.timer > value.duration*TICRATE)) or value.fxdone or player.playerstate == PST_DEAD or (player.exiting or stoppedclock or (player.pflags & PF_FINISHED)))
        or player.khaos.mapchange then
            SRB2CE.DebugPrint(player, "MARKING 4 REMOVAL: " + effecter.name)
            if effecter.endfunc then
                effecter.endfunc(player, index, value)
            end
            if player.khaosoptions.SRB2CE_HUD or effecter.forceend and (not player.exiting) then -- last one is to prevent all the effect end sounds from playing at the same time
                if effecter.endsfx then
                    if value.localsfx then
                        S_StartSound(player.mo, effecter.endsfx, player)
                    else
                        S_StartSound(player.mo, effecter.endsfx)
                    end
                end
            end
            --table.remove(player.khaoseffects, index)
            player.khaoseffects[index].removeme = 1 -- timer that counts up to 35, then it gets removed completely
            --table.remove(player.khaoshud, index)
        end

        if value.removeme then
            value.removeme = $ + 1
            if value.removeme > 40 or effecter.removeimmediately then
                SRB2CE.DebugPrint(player, "NOW REMOVING: " + effecter.name)
                table.remove(player.khaoseffects, index)
            end
        end
    end

    if player.khaos.mapchange then
        player.khaoseffects = {} -- forcibly reset it; all effects should be removed by now.
        player.khaos.mapchange = false
    end

    -- Rewind --
    if (leveltime % (3*TICRATE)) == 0 then --every 2 seconds is a lot more generous
        --print("set new position")
        player.khaos.RewindX = player.mo.x
        player.khaos.RewindY = player.mo.y
        player.khaos.RewindZ = player.mo.z
        player.khaos.RewindAngle = player.mo.angle
    end

    -- Restore angle'd view --
    if player.khaos.RestoreViewAngle then
        if player.khaos.SpinOtherWay then
            if player.viewrollangle < 0
                player.viewrollangle = $ + (ANG10/4)
                player.mo.rollangle = $ - (ANG10/4)
            else
                player.viewrollangle = 0
                player.mo.rollangle = 0
                player.khaos.RestoreViewAngle = false
                player.khaos.SpinOtherWay = false
            end
        else
            if player.viewrollangle > 0 then
                player.viewrollangle = $ - (ANG10/4)
                player.mo.rollangle = $ + (ANG10/4)
            else
                player.viewrollangle = 0
                player.mo.rollangle = 0
                player.khaos.RestoreViewAngle = false
            end
        end
    end

    -- Restore Sprite Stuff --
    if player.khaos.RestoreSpriteYScaleUp then
        if player.mo.spriteyscale < FRACUNIT then
            player.mo.spriteyscale = $ + FRACUNIT/10
        elseif player.mo.spriteyscale >= FRACUNIT then
            player.mo.spriteyscale = FRACUNIT
            player.khaos.RestoreSpriteYScaleUp = false
        end
    end

    if player.khaos.RestoreSpriteXScaleDown then
        if player.mo.spritexscale > FRACUNIT then
            player.mo.spritexscale = $ - FRACUNIT/10
        elseif player.mo.spritexscale <= FRACUNIT then
            player.mo.spritexscale = FRACUNIT
            player.khaos.RestoreSpriteXScaleDown = false
        end
    end

    if player.khaos.RestoreSpriteYScaleDown then
        if player.mo.spriteyscale > FRACUNIT then
            player.mo.spriteyscale = $ - FRACUNIT/10
        elseif player.mo.spriteyscale <= FRACUNIT then
            player.mo.spriteyscale = FRACUNIT
            player.khaos.RestoreSpriteYScaleDown = false
        end
    end

    -- [[ -- HUD Logic -- ]] --

    player.khaos.TextPopUpTimer = $ + 1
    if (player.khaos.TextPopUpTimer > 5*TICRATE and (((player.khaos.TextPopUpTimer-2*TICRATE)*2) % 5 == 0)) then
        player.khaos.TextPopUpFade = $ + 1
    end

    if player.khaos.TextPopUpTimer > 10*TICRATE
        player.khaos.respawnmeinchief = true
    end

    -- [[ -- Universal Effect Player HUD Logic -- ]] --

    player.khaos.ueTimer = ueTimer -- for the hud
    player.khaos.ueTimerMax = ueTimerMax
    if initiateUE then
        player.khaos.ueTimer = 0 -- just to make it look smooth when an effect is being rolled
    end

    if player.khaos.UniEffectText
        local MaxLength = #player.khaos.UniEffectText
        local SubMaxLength = 0
        if player.khaos.UniEffectSubText then
            SubMaxLength = #player.khaos.UniEffectSubText
        end

        if player.khaos.UniPopUpTimer % 1 == 0 then -- every one tic. used to be two but i'm leaving this here so it's easier to revert this if i want to
            player.khaos.UniEffecTextBufferNumber = $ + 1
        end

        player.khaos.UniEffectTextBuffer = string.sub(player.khaos.UniEffectText, 1, min(player.khaos.UniEffecTextBufferNumber, MaxLength))
        if player.khaos.UniEffectSubText then
            player.khaos.UniEffectSubTextBuffer = string.sub(player.khaos.UniEffectSubText, 1, min(player.khaos.UniEffecTextBufferNumber, SubMaxLength))
        end

        player.khaos.UniPopUpTimer = $ + 1
        if (player.khaos.UniPopUpTimer > 5*TICRATE and (((player.khaos.UniPopUpTimer-2*TICRATE)*2) % 5 == 0)) then
            player.khaos.UniTextPopUpFade = $ + 1
        end

        if player.khaos.UniPopUpTimer > 8*TICRATE
            player.khaos.UniPopUpTimer = 0
            player.khaos.UniTextPopUpFade = 0
            player.khaos.UniEffecTextBufferNumber = 1
            player.khaos.UniEffectText = nil
            player.khaos.UniEffectSubText = nil
            player.khaos.UniEffectSubTextBuffer = ""
            player.khaos.UniEffectTextBuffer = ""
        end
    end
end)

addHook("MapChange", function() --leftover from TKK, but keeping because i think this'll make netgames play a bit nicer like this
    ueTimer = 0
    ueTimerBuffer = 0
    ueTimerMax = tonumber(CV_FindVar("SRB2CE_Universal_Timer").value) * TICRATE
    SRB2CE_Enabled = tonumber(CV_FindVar("SRB2CE_Enabled").value)
    SRB2CE.GravityTimer = nil
    initiateUE = false
    SRB2CE.RunFunc = nil
    PreviousUniversalEffect = ""
    for player in players.iterate do
        --[[player.khaos = nil
        player.khaosmemory = nil
        player.khaoseffects = nil
        player.khaostext = nil]]
        player.khaos.mapchange = true
        player.khaos.UniPopUpTimer = 0
        player.khaos.UniTextPopUpFade = 0
        player.khaos.UniEffecTextBufferNumber = 1
        player.khaos.UniEffectText = nil
        player.khaos.UniEffectSubText = nil
        player.khaos.UniEffectSubTextBuffer = ""
        player.khaos.UniEffectTextBuffer = ""
    end
end)


-- gametype == GT_BATTLE

local prevSRB2CE_Amplify = false
local prevSRB2CE_Enabled = 1

local function UEChecker(index)
    local StatusEffect = UniversalEffectsTable[index]

    if StatusEffect.requiredplayers and (SRB2CE.playercount - SRB2CE.botcount) < StatusEffect.requiredplayers then
        return false
    end

    if StatusEffect.gamemodeneedslives and not G_GametypeUsesLives() then
        return false
    end

    if StatusEffect.nobattle and gametype == GT_BATTLE then
        return false
    end

    if StatusEffect.nonights and (mapheaderinfo[gamemap].typeoflevel & TOL_NIGHTS) then
        return false
    end

    if StatusEffect.nospecialstage and G_IsSpecialStage(gamemap) then
        return false
    end

    return true
end

addHook("ThinkFrame", function()
    if (leveltime % TICRATE == 0) then -- don't spam it, it might cause unnessecary lag to check every tic
        ueTimerMax = tonumber(CV_FindVar("SRB2CE_Universal_Timer").value) * TICRATE
        if prevSRB2CE_Amplify != SRB2CE.AmplifyChecker() then
            prevSRB2CE_Amplify = SRB2CE.AmplifyChecker()
            local Texter = "\x85" + "Amplify Mode" + "\x80" + " is now "
            local Sounder = sfx_tmxsuc
            if prevSRB2CE_Amplify == 0 then
                Texter = $ + "\x82" + "deactivated."
            else
                Texter = "\202" + $ + "\x82" + "activated!!!"
                Sounder = sfx_gshd9
            end
            for player in players.iterate do
                if SRB2CE.ValidityCheck(player) then
                    SRB2CE.addHUDText(player, Texter, true)
                    S_StartSound(player.mo, Sounder, player)
                end
            end
        end
        if prevSRB2CE_Enabled != tonumber(CV_FindVar("SRB2CE_Enabled").value) then
            prevSRB2CE_Enabled = tonumber(CV_FindVar("SRB2CE_Enabled").value)
            local Texter = "\x82" + "SRB2: Chaotic Edition" + "\x80" + " will be "
            if prevSRB2CE_Enabled == 0 then
                Texter = $ + "\x85" + "deactivated "
            else
                Texter = $ + "\x83" + "activated "
            end
            Texter = $ + "\x80" + "next round."
            chatprint("* " + Texter)
            for player in players.iterate do
                CONS_Printf(player, Texter)
                S_StartSound(player.mo, sfx_tmxsuc, player)
                if SRB2CE.ValidityCheck(player) then
                    SRB2CE.addHUDText(player, Texter, true)
                end
            end
        end
        SRB2CE.botcount = 0
        SRB2CE.playercount = 0
        hasanyonefinished = false
        for player in players.iterate do
            SRB2CE.playercount = $ + 1
            if player.bot then
                SRB2CE.botcount = $ + 1
            end
            if player.exiting or (player.pflags & PF_FINISHED) then
                --print("SOMEONE FINISHED")
                hasanyonefinished = true
            end
        end
    end

    if not SRB2CE.playercount then return end -- no players exist! assume this is the title screen.

    if SRB2CE.GravityTimer != nil then
        SRB2CE.GravityTimer = $ - 1
        if SRB2CE.GravityTimer <= 0 then
            gravity = mapheaderinfo[gamemap].gravity or FRACUNIT/2
            for player in players.iterate do
                SRB2CE.addHUDText(player, "Gravity has been returned to normal!", true)
            end
            SRB2CE.GravityTimer = nil
        end
    end

    -- Universal Effects --
    if not tonumber(CV_FindVar("SRB2CE_Universal_Fx_Enabled").value)
    or hasanyonefinished or stoppedclock
    or (players[0] and not (players[0].khaossp and players[0].khaossp.universalfx)) -- gets first player's khaossp stuff; this would presumbably be the host of the server. they need to unlock it!
    or mapheaderinfo[gamemap].srb2cecity
    then
        SRB2CE.ueTimerDisabled = true
    return end
    -- don't run if the option is disabled!
    ueTimer = $ + 1
    SRB2CE.ueTimerDisabled = false

    if ueTimer > ueTimerMax and not initiateUE and not SRB2CE.RunFunc then -- don't run if SRB2CE.RunFunc is still going! hopefully the script didn't break
        S_StartSoundAtVolume(nil, sfx_gshd2, 255)
        for player in players.iterate do
            if SRB2CE.ValidityCheck(player)
                if not player.exiting and not (player.pflags & PF_FINISHED) then
                    player.khaos.UniEffectText = "Rolling Universal Effect..."
                    -- reset everything just to make sure that the text actually shows up if the timer is set to be way too short somehow
                    player.khaos.UniPopUpTimer = 0
                    player.khaos.UniTextPopUpFade = 0
                    player.khaos.UniEffecTextBufferNumber = 1
                    player.khaos.UniEffectSubText = nil
                    player.khaos.UniEffectSubTextBuffer = ""
                end
            end
        end
        initiateUE = true
        ueTimer = 0
    end

    if initiateUE then
        local StatusRand = P_RandomRange(1, #UniversalEffectsTable)
        local StatusEffect = UniversalEffectsTable[StatusRand]
        ueTimer = 0 -- keep timer at 0, for in case it's taking extrodionarly long to roll an effect due to bad RNG

        if StatusEffect then -- sfx_s3kd2s
            if ueTimerBuffer > TICRATE then -- only apply after a second passes, should give enough time to plonk in the text
                if P_RandomChance(StatusEffect.chance) and StatusEffect.name != PreviousUniversalEffect and UEChecker(StatusRand) then
                    initiateUE = false
                    SRB2CE.RunFunc = StatusRand

                    for player in players.iterate do
                        if SRB2CE.ValidityCheck(player) then
                            if player.khaosoptions.SRB2CE_Universal_Announcer > 0 then
                                local TheSnable = StatusEffect.sounds
                                if player.khaosoptions.SRB2CE_Universal_Announcer == 2 and StatusEffect.microsounds then
                                    TheSnable = StatusEffect.microsounds
                                end
                                if SRB2CE.AmplifyChecker(player) and StatusEffect.amplifysounds then
                                    TheSnable = StatusEffect.amplifysounds
                                    if player.khaosoptions.SRB2CE_Universal_Announcer == 2 and StatusEffect.microamplifysounds then
                                        TheSnable = StatusEffect.microamplifysounds
                                    end
                                end
                                local TheSound = TheSnable[1]
                                if #TheSnable > 1 then
                                    TheSound = TheSnable[P_RandomRange(1, #TheSnable)]
                                end

                                S_StartSoundAtVolume(nil, TheSound, 255, player) -- do it 2 times, because SRB2 has reasonable audio levels unlike ring racers :)
                                S_StartSoundAtVolume(nil, TheSound, 255, player)
                            end

                            player.khaos.UniEffectText = StatusEffect.name

                            if SRB2CE.AmplifyChecker(player) and StatusEffect.amplifyname then
                                player.khaos.UniEffectText = StatusEffect.amplifyname
                            end
                            if player.khaosoptions.SRB2CE_HUD and player.khaosoptions.SRB2CE_Print_Effect then
                                CONS_Printf(player, "*\x87 Universal Effect\x80: " .. player.khaos.UniEffectText)
                            end

                            if StatusEffect.hudtextprefix then
                                player.khaos.UniEffectText = StatusEffect.hudtextprefix + $
                            end

                            if StatusEffect.subtext then
                                local TheGlable = StatusEffect.subtext
                                if SRB2CE.AmplifyChecker(player) and StatusEffect.amplifysubtext then
                                    TheGlable = StatusEffect.amplifysubtext
                                end
                                player.khaos.UniEffectSubText = TheGlable[1]
                                if #TheGlable > 1 then
                                    local TheChosenOne = P_RandomRange(1, #TheGlable)
                                    player.khaos.UniEffectSubText = TheGlable[TheChosenOne]
                                end
                            else -- force clear
                                player.khaos.UniEffectSubText = ""
                            end
                            player.khaos.UniPopUpTimer = 0
                            player.khaos.UniTextPopUpFade = 0
                            player.khaos.UniEffecTextBufferNumber = 1
                        end

                        if player.spectator and VERSION >= 2 then
                            local TheSnable = StatusEffect.sounds
                            if tonumber(CV_FindVar("SRB2CE_Universal_Announcer").value) then
                                TheSnable = StatusEffect.microsounds
                            end
                            local TheSound = TheSnable[1]
                            S_StartSoundAtVolume(nil, TheSound, 255, player)
                            S_StartSoundAtVolume(nil, TheSound, 255, player)
                        end
                    end
                    PreviousUniversalEffect = StatusEffect.name
                    ueTimerBuffer = 0
                end
            else
                ueTimerBuffer = $ + 1
                ueTimer = 0 -- reset timer, just to be sure.
                return -- not sure what this is doing here actually. i think it's to stop runfunc from going ahead?
            end
        end
    end

    if SRB2CE.RunFunc then
        UniversalEffectsTable[SRB2CE.RunFunc].func()
    end

    SRB2CE.ueTimer = ueTimer
    SRB2CE.ueTimerMax = ueTimerMax
    SRB2CE.PreviousUniversalEffect = PreviousUniversalEffect
end)

addHook("MobjThinker", function(mobj)
    if not mobj.khaostime then return false end

    if P_IsObjectOnGround(mobj) and mobj.state == S_TNTBARREL_STND1 then
        P_KillMobj(mobj)
    end

end, MT_TNTBARREL)

--[[addHook("MobjThinker", function(mobj)
    --if not mobj.khaostime then return false end


    if not mobj.started then
        mobj.state = S_ROSY_WALK
        mobj.pdist = 9000*FRACUNIT
        P_LookForPlayers(mobj, 0, true)
        mobj.started = true
    end

    --this is basically just a port of the SRB2 source code, forced to make it *always* make amy run (unless hugging)

    --TODO: that

    if not mobj.target then
        P_LookForPlayers(mobj, 0, true)
        return false -- resume normal behavior just for now
    end




    if mobj.state == S_ROSY_WALK then
        local x = mobj.x
        local y = mobj.y
        local z = mobj.z
        local angletoplayer = R_PointToAngle2(x, y, mobj.target.x, mobj.target.y);
        --local allowed = P_TryMove(mobj, mobj.target.x, mobj.target.y, false);
        local allowed = true --lol

        --P_UnsetThingPosition(mobj);
        mobj.x = x;
        mobj.y = y;
        mobj.z = z;
        --P_SetThingPosition(mobj);

        if (allowed) then
            fixed_t mom, max;
            P_Thrust(mobj, angletoplayer, (3*FRACUNIT) >> 1);
            mom = FixedHypot(mobj.momx, mobj.momy);
            max = pdist;
            if ((--mobj.extravalue1) <= 0)
            {
                if (++mobj.frame > mobj.state.frame + mobj.state.var1)
                    mobj.frame = mobj.state.frame;
                if (mom > 12*mobj.scale)
                    mobj.extravalue1 = 2;
                else if (mom > 6*mobj.scale)
                    mobj.extravalue1 = 3;
                else
                    mobj.extravalue1 = 4;
            }
            if (max < (mobj.radius + mobj.target.radius))
            {
                mobj.momx = mobj.target.player.cmomx;
                mobj.momy = mobj.target.player.cmomy;
                if ((mobj.cvmem > TICRATE && !player.exiting) || !targonground)
                    P_SetMobjState(mobj, (stat = S_ROSY_STND));
                else
                {
                    mobj.target.momx = mobj.momx;
                    mobj.target.momy = mobj.momy;
                    P_SetMobjState(mobj, (stat = S_ROSY_HUG));
                    S_StartSound(mobj, sfx_cdpcm6);
                    mobj.angle = angletoplayer;
                }
            }
            else
            {
                max /= 3;
                if (max > 30*mobj.scale)
                    max = 30*mobj.scale;
                if (mom > max && max > mobj.scale)
                {
                    max = FixedDiv(max, mom);
                    mobj.momx = FixedMul(mobj.momx, max);
                    mobj.momy = FixedMul(mobj.momy, max);
                }
                if (abs(mobj.momx) > mobj.scale || abs(mobj.momy) > mobj.scale)
                    mobj.angle = R_PointToAngle2(0, 0, mobj.momx, mobj.momy);
            }

    end


    return true -- forcefully override default behavior
end, MT_ROSY)]]

addHook("MobjThinker", function(mobj)
    mobj.color = SKINCOLOR_ICY
    local target = mobj.target

    if not mobj.revolvingangle then
        mobj.revolvingangle = 1
        mobj.zpos = 0
        mobj.xystart = 0
        S_StartSound(mobj, sfx_chhiom)
    end
    if target and mobj.state == S_CHAOSOMOCHAO then
        if not S_OriginPlaying(mobj) then
            S_StartSound(mobj, sfx_chanom)
        end

        if mobj.xystart < 40 then
            mobj.xystart = $ + 1
        end
        mobj.scale = target.scale
        mobj.angle = mobj.revolvingangle

        local xmove = target.x + P_ReturnThrustX(mobj, mobj.angle, FixedMul(-mobj.xystart*FRACUNIT, mobj.scale)) -- stole this from SRB2's source code lol
        local ymove = target.y + P_ReturnThrustY(mobj, mobj.angle, FixedMul(-mobj.xystart*FRACUNIT, mobj.scale))
        local zmove = target.z + 25*FRACUNIT
        if (target.flags2 & MF2_OBJECTFLIP) or (target.eflags & MFE_VERTICALFLIP) then
            zmove = (target.z + (target.height/2)) - 25*FRACUNIT --+ target.height
        end
        P_MoveOrigin(mobj, xmove, ymove, zmove)

        mobj.flags2 = target.flags2 & MF2_OBJECTFLIP
        mobj.eflags = target.eflags & MFE_VERTICALFLIP

        mobj.revolvingangle = $ + ANG2 + ANG1
    else
        S_StopSoundByID(mobj, sfx_chanom)
    end

    if mobj.killtimer then
        mobj.killtimer = $ - 1
        if mobj.killtimer == 0 then
            S_StartSound(mobj, sfx_pop)
            P_KillMobj(mobj)
        end
    end
end, MT_CHAOSOMO)

addHook("MobjThinker", function(aggressive)
    if not aggressive.valid or not aggressive.target or not aggressive.target.valid then
        return
    end

    if aggressive.state != S_CHAOTICPIZZAFACE and aggressive.state != S_CHAOSSKULL then
        aggressive.rollangle = $ + ANG15
    end

    A_HomingChase(aggressive, aggressive.speed, 0)

    if (R_PointToDist2(aggressive.target.x, aggressive.target.y, aggressive.x, aggressive.y))/FRACUNIT < 35 and
    (abs((aggressive.target.z - aggressive.z)/FRACUNIT) < 25) and not (aggressive.target.player and aggressive.target.player.playerstate == PST_DEAD)
    then
        if SRB2CE.AmplifyChecker() then
            P_RemoveShield(aggressive.target.player)
            aggressive.target.player.powers[pw_invulnerability] = 0
            P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
        else
            P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_SPIKE)
        end

        S_StartSound(aggressive.target, sfx_chapki)
        if aggressive.state == S_CHAOTICXPLOSPIZZA then
            local Explode = P_SpawnMobj(aggressive.target.x, aggressive.target.y, aggressive.target.z, MT_BOSSEXPLODE)
            S_StartSound(explode, sfx_bkpoof)
            Explode.destscale = 3*FRACUNIT
            Explode.scalespeed = FRACUNIT/7
        end
        aggressive.target.player.khaos.PizzaHitTarget = true
        P_RemoveMobj(aggressive)
    end

    if aggressive.valid and aggressive.target.player.khaos.PizzaRemoval then
        aggressive.target.player.khaos.PizzaRemoval = false
        P_RemoveMobj(aggressive)
    end
end, MT_JIMBO)

--[[addHook("MobjThinker", function(aggressive) -- this code is directly lifted from my MMT script, i have no shame
    if aggressive.valid and aggressive.target and aggressive.target.valid then
        if not aggressive.initializedtime then
            aggressive.shadowscale = 3*mapobjectscale
            if aggressive.sizeoption == "large" then
                aggressive.speed = 140*mapobjectscale
                aggressive.scale = mapobjectscale + mapobjectscale*(2/4)
            elseif aggressive.sizeoption == "medium" then
                aggressive.speed = 180*mapobjectscale
                aggressive.scale = mapobjectscale
            elseif (aggressive.sizeoption == "small" or aggressive.sizeoption == "personal") then
                aggressive.speed = 250*mapobjectscale
                aggressive.scale = mapobjectscale/2
            elseif (aggressive.sizeoption == "xl" or aggressive.sizeoption == "extra large" or aggressive.sizeoption == "extralarge") then
                aggressive.speed = 140*mapobjectscale
                aggressive.scale = mapobjectscale*2
            else
                aggressive.speed = 280*mapobjectscale
                aggressive.scale = mapobjectscale/2
            end
            aggressive.ogspeed = aggressive.speed
            aggressive.initializedtime = true
        end
        if not aggressive.rollanglesprite then
            aggressive.rollanglesprite = 0
        end
        aggressive.rollanglesprite = $ + ANG15

        if aggressive.state != S_CHAOTICPIZZAFACE then
            aggressive.rollangle = aggressive.rollanglesprite
        end

        if isPlayerRespawning(aggressive.target.player) or aggressive.target.player.playerstate != PST_LIVE then --slow down if the player is respawning or dead
            aggressive.speed = aggressive.ogspeed / 10
        else
            aggressive.speed = aggressive.ogspeed
        end

        if (R_PointToDist2(aggressive.target.x, aggressive.target.y, aggressive.x, aggressive.y))/mapobjectscale < 35 and
        (abs((aggressive.target.z - aggressive.z)/mapobjectscale) < 25) and (aggressive.target.player and aggressive.target.player.playerstate == PST_LIVE)
        and not aggressive.Remove
        and not isPlayerRespawning(aggressive.target.player) then
            P_DamageMobj(aggressive.target, aggressive, aggressive, DMG_NORMAL)

            S_StartSound(aggressive.target, sfx_chapki)
            if aggressive.state == S_CHAOTICXPLOSPIZZA then
                local Explode = P_SpawnMobj(aggressive.target.x, aggressive.target.y, aggressive.target.z, MT_BOSSEXPLODE)
                S_StartSound(Explode, sfx_bkpoof) -- only just now realizing this didn't fucking play the sound because i typo'd the variable name. 3 years this typo existed! 3 years!!!
                Explode.destscale = 5*mapobjectscale
                Explode.scalespeed = FRACUNIT/7
            end
            aggressive.target.player.khaos.PizzaHitTarget = true
            aggressive.Remove = true
        end

        if aggressive.target and not aggressive.target.player.khaos.PizzaRemoval then
            A_HomingChase(aggressive, aggressive.speed, 0)
        else
            P_RemoveMobj(aggressive)
        end
    else
        P_RemoveMobj(aggressive)
    end
end, MT_JIMBO)]]

addHook("MobjThinker", function(cs)
    if cs.despawn then P_RemoveMobj(cs) return end
    if not cs.gayer then P_RemoveMobj(cs) return end
    if not cs.gayer.valid then P_RemoveMobj(cs) return end
    if cs.gayer.RemoveChangeSprite then P_RemoveMobj(cs) return end
    local player = cs.gayer

    -- replicate a lot of player stuff here
    P_MoveOrigin(cs, player.mo.x, player.mo.y, player.mo.z)
    cs.shadowscale = player.mo.shadowscale
    if cs.summonedasstate != S_CHAOSMAC then
        cs.scale = player.mo.scale
    else
        cs.scale = player.mo.scale/2
        cs.shadowscale = player.mo.shadowscale*4
    end
    if cs.summonedasstate != S_CHAOSWIN9X
        cs.color = player.mo.color
    end
    cs.spritexoffset = player.mo.spritexoffset
    cs.spriteyoffset = player.mo.spriteyoffset
    cs.spritexscale = player.mo.spritexscale
    cs.spriteyscale = player.mo.spriteyscale
    cs.angle = player.drawangle or player.mo.angle
    --cs.flags = player.mo.flags
    cs.flags2 = player.mo.flags2 & MF2_OBJECTFLIP
    cs.eflags = player.mo.eflags & MFE_VERTICALFLIP
    cs.flags2 = player.mo.flags2 & MF2_SHADOW
    cs.flags2 = player.mo.flags2 & MF2_DONTDRAW
    cs.roll = player.mo.roll
    cs.pitch = player.mo.pitch
    cs.alpha = player.mo.alpha -- hopefully this fixes the issue w/ ghost

    if cs.summonedasstate == S_CHAOTICPIZZA then
        cs.rollangle = $ + ANG15
    else
        cs.rollangle = player.mo.rollangle
    end

    if cs.summonedasstate == S_SOSS_RUN1 then -- summoned ass tate
        if player.speed/FRACUNIT == 0 then
            cs.tics = -1
        else
            if cs.tics == -1 then
                cs.tics = 1
            end
            if player.speed-2*FRACUNIT >= 10*FRACUNIT and cs.tics == 3 then
                cs.tics = 2
            end
        end
    elseif (cs.summonedasstate == S_YHORIZ1 or cs.summonedasstate == S_YELLOWSPRING) then
        if abs(player.mo.momz) > 0 and not cs.sprung then
            if cs.summonedasstate == S_YHORIZ1 then
                cs.state = S_YHORIZ2
            else
                cs.state = S_YELLOWSPRING2
            end
            cs.sprung = true
        end
    end

    if P_IsObjectOnGround(player.mo) then
        cs.sprung = false
    end

end, MT_CHANGESPRITE)

addHook("MobjThinker", function(mo)
    if mo.despawn then P_RemoveMobj(cs) return end
    if not mo.gayer then P_RemoveMobj(cs) return end
    if not mo.gayer.valid then P_RemoveMobj(cs) return end

    local player = mo.gayer
    if not mo.initialized then
        --P_MoveOrigin(mo, player.mo.x, player.mo.y, player.mo.z)
        mo.skin = player.mo.skin
        mo.state = player.mo.state
        mo.sprite2 = player.mo.sprite2
        --mo.sprite = player.mo.sprite
        mo.color = player.mo.color
        mo.angle = player.mo.angle
        mo.scale = player.mo.scale
        mo.spritexoffset = player.mo.spritexoffset
        mo.spriteyoffset = player.mo.spriteyoffset
        mo.colorized = player.mo.colorized
        --mo.flags = player.mo.flags
        --mo.flags2 = player.mo.flags2
        mo.eflags = player.mo.eflags
        mo.rollangle = player.mo.rollangle
        mo.frame = player.mo.frame
        mo.initialized = true
    end

end, MT_LAGPLAYER)

addHook("MobjThinker", function(mo)
    if mo.despawn then P_RemoveMobj(mo) return end
    if not mo.gayer then P_RemoveMobj(mo) return end
    if not mo.gayer.valid then P_RemoveMobj(mo) return end

    local player = mo.gayer

    mo.angle = player.mo.angle
    mo.scale = player.mo.scale/2
    mo.spritexoffset = player.mo.spritexoffset
    mo.spriteyoffset = player.mo.spriteyoffset
    mo.eflags = player.mo.eflags
    mo.rollangle = player.mo.rollangle

    local zoffset = 105*mo.scale
    if (mo.flags2 & MF2_OBJECTFLIP) or (mo.eflags & MFE_VERTICALFLIP) then
        zoffset = $ * -1
    end

    P_MoveOrigin(mo, player.mo.x, player.mo.y, player.mo.z+zoffset)

end, MT_MARKED4DEATH)

addHook("MobjThinker", function(mo)
    if mo.despawn then P_RemoveMobj(mo) return end
    if not mo.gayer then P_RemoveMobj(mo) return end
    if not mo.gayer.valid then P_RemoveMobj(mo) return end

    local player = mo.gayer

    if not mo.timer then
        mo.scale = player.mo.scale
        mo.scaleresizer = player.mo.scale
        mo.timer = 0
        mo.funnyoffset = 0
    end

    mo.timer = $ + 1

    local themultiplier = 1

    if (mo.flags2 & MF2_OBJECTFLIP) or (mo.eflags & MFE_VERTICALFLIP) then
        themultiplier = -1
    end

    if mo.timer < (TICRATE)-4 then
        mo.scale = $ - mo.scaleresizer/37
    else
        mo.funnyoffset = $ + (16*mo.scale *themultiplier)
    end


    P_MoveOrigin(mo, player.mo.x, player.mo.y, player.mo.z+mo.funnyoffset)

    if mo.timer > 4*TICRATE then
        P_RemoveMobj(mo)
    end

end, MT_KHAOSMINICRIT)

addHook("MobjThinker", function(pieces) -- this is re-used code from the Rock Shield script. which in turn i have also stolen this from my ice springs code. lots of re-used code :3c
    pieces.shadowscale = FRACUNIT
    local function RestoreState()
        if pieces.number == 1
            pieces.state = S_ICESHRAPCH
        elseif pieces.number == 2
            pieces.state = S_ICESHRAP2CH
        elseif pieces.number == 3
            pieces.state = S_ICESHRAP3CH
        else
            pieces.state = S_ICESHRAP4CH
        end
    end

    -- [[ remove shrapnel ]] --
    if pieces.timer == nil
        RestoreState()
        pieces.timer = P_RandomRange(0, TICRATE)
        pieces.transstate = 0
        P_SetObjectMomZ(pieces, P_RandomRange(6, 12)*FRACUNIT, false)
        P_InstaThrust(pieces, pieces.angle, -FixedMul((P_RandomRange(5, 35)*FRACUNIT)/12 + FRACUNIT + P_RandomFixed(), pieces.scale))
    end

    if not P_IsObjectOnGround(pieces)
        pieces.rollangle = $ + FixedAngle(10<<FRACBITS)
    else
        pieces.rollangle = $ + FixedAngle(4<<FRACBITS)
    end

    pieces.timer = $ + 1

    if (pieces.timer % 10) == 0
        pieces.transstate = $ + 1
    end

    if pieces.transstate == 1
        RestoreState()
        pieces.frame = $1 | TR_TRANS10
    elseif pieces.transstate == 2
        RestoreState()
        pieces.frame = $1 | TR_TRANS20
    elseif pieces.transstate == 3
        RestoreState()
        pieces.frame = $1 | TR_TRANS30
    elseif pieces.transstate == 4
        RestoreState()
        pieces.frame = $1 | TR_TRANS40
    elseif pieces.transstate == 5
        RestoreState()
        pieces.frame = $1 | TR_TRANS50
    elseif pieces.transstate == 6
        RestoreState()
        pieces.frame = $1 | TR_TRANS60
    elseif pieces.transstate == 7
        RestoreState()
        pieces.frame = $1 | TR_TRANS70
    elseif pieces.transstate == 8
        RestoreState()
        pieces.frame = $1 | TR_TRANS80
    elseif pieces.transstate == 9
        RestoreState()
        pieces.frame = $1 | TR_TRANS90
    elseif pieces.transstate >= 10
        P_RemoveMobj(pieces)
        return
    end
end, MT_ICESHRAPNELCHAOTIC)



addHook("MobjDamage", function(mo, inflictor, source, damage, damagetype)
    local player = mo.player
    if not player then return false end
    if not player.khaos then return false end -- just in case
    if (not player.powers[pw_flashing] or player.khaos.DodgeHit) and ((player.khaos.IsLucky and P_RandomChance(FRACUNIT/2 + FRACUNIT/8)) or player.khaos.DodgeHit) then -- not P_PlayerInPain(player) and
        --player.spinouttype = KSPIN_IFRAMES -- required to have player.flashing not get immediately reset to 0
        player.powers[pw_flashing] = 76
        if not player.khaos.DodgeHit and player.khaos.IsLucky then
            S_StartSound(player.mo, sfx_s3k7a)
            SRB2CE.addHUDText(player, "Dodged a hit, lucky!", true)
        end
        return true -- by returning, we override the rest of mobjdamage's functions, ergo no damage
    end
    --[[if player.khaos.ImDyingSquirtle then
        player.khaos.Tumbler = $ + 2
        if player.khaos.Tumbler < 15 -- stops the player from being whacked too much... hopefully.
            local thing = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z+32*player.mo.scale, MT_KHAOSMINICRIT)
            thing.gayer = player
            S_StartSound(player.mo, sfx_chmini)
            K_TumblePlayer(player)
        end
    end]]

    if player.khaos.Eye4Eye then
        local enemy = inflictor
        if enemy and enemy.valid then
            if (enemy.flags & MF_MISSILE) and inflictor.target and inflictor.target.valid then
                enemy = inflictor.target
            end
            P_DamageMobj(enemy, nil, nil, DMG_SPIKE)
        end
    end

    if SRB2CE_Enabled and not SRB2CE.IsMatchMode() and player.khaossp.ringkeep and tonumber(CV_FindVar("SRB2CE_Ringkeep").value) then
        if not player.powers[pw_shield] and player.rings > 1
            P_PlayRinglossSound(mo, player)
            player.khaos.RingPainTime = true
            P_DoPlayerPain(player, source, inflictor)
            return true
        else
            return false
        end
    else
        return false
    end
end, MT_PLAYER)

addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
    if not inflictor then return false end
    if not inflictor.player then return false end
    local player = inflictor.player
    if not player.khaos then return false end

    if player.khaos.Eye4Eye then
        P_DamageMobj(inflictor, nil, nil, DMG_SPIKE)
        player.khaossp.timeshitbykhaos = $ + 1
    end

end)

local function SanityCheck(var1, var2) -- this is here to prevent players from loading invalid settings
    local var1place = false
    for index, value in ipairs(SRB2CE.optionstableofdoom) do
        if var1 == value.name then
            var1place = index
        end
    end

    if var1place then
        local snable = SRB2CE.optionstableofdoom[var1place]
        if snable.localonly then return false end -- don't load local variables. TODO: allow this to be loaded by admins?
        if (snable.typer == "number" and tonumber(var2) != nil) and (tonumber(var2) >= snable.values[1] and tonumber(var2) <= snable.values[2]) then
            return true
        end
        if snable.typer == "string" then
            local var2exists = false
            for key, value in pairs(snable.values) do
                if value == var2 then
                    var2exists = true
                end
            end
            return var2exists
        end
    end

    return false
end

local function numbercunktion(player, var1, var2)
    if var1 == nil or var2 == nil or not SanityCheck(var1, var2) then
        CONS_Printf(player, "\x85" + "* INVALID SETTINGS: \x80" .. tostring(var1) .. ", " .. tostring(var2))
        if player and player.khaos then player.khaos.LoadFailed = true end
        return
    end

    if player and player.khaosoptions then player.khaosoptions[var1] = tonumber(var2) end
end

local function stringcunktion(player, var1, var2)
    if var1 == nil or var2 == nil or not SanityCheck(var1, var2) then
        CONS_Printf(player, "\x85" + "* INVALID SETTINGS: \x80" .. tostring(var1) .. ", " .. tostring(var2))
        if player and player.khaos then player.khaos.LoadFailed = true end
        return
    end

    if player and player.khaosoptions then player.khaosoptions[var1] = tostring(var2) end
end


COM_AddCommand("zkhaosNUM", numbercunktion)
COM_AddCommand("zkhaosSTR", stringcunktion)

COM_AddCommand("SRB2CE_Load_Settings", Khaos_LoadSettings)
COM_AddCommand("SRB2CE_Save_Settings", Khaos_SaveSettings)


COM_AddCommand("SRB2CE_Attributions", function(player)
    for key, value in pairs(SRB2CE.attributions) do
        CONS_Printf(player, "* \x82"+value[1]+"\x80: "+value[2])
    end
end)

local function TheNumberFeeder(var)
    if consoleplayer then -- stops this thing from bitching in the logs
        COM_BufInsertText(consoleplayer, "zkhaosNUM " + var.name + " " + var.value)
        --CONS_Printf(consoleplayer, "zkhaosNUM " + var.name + " " + var.value)
    end
end

local function TheStringFeeder(var)
    if consoleplayer then
        COM_BufInsertText(consoleplayer, "zkhaosSTR " + var.name + " " + var.value)
        --CONS_Printf(consoleplayer, "zkhaosSTR " + var.name + " " + var.value)
    end
end

--[[COM_AddCommand("SRB2CE_Blacklist_Change", function(player, var1, silent)
    local effecter = (var1 and indexOfName2(KhaosEffectsTable, var1))
    if var1 then
        effecter = KhaosEffectsTable[effecter]
    else
        local silly = tonumber(CV_FindVar("SRB2CE_Blacklist_Set").value)
        for key, value in pairs(CV_TKKBlackList) do
            if value == silly then
                silly = key
            end
        end
        effecter = KhaosEffectsTable[indexOfName2(KhaosEffectsTable, silly)]
    end

    if not effecter then
        CONS_Printf(player, "* \x85"+"Invalid effect name: \x80\"\x82" + var1 + "\x80\" Be sure to use \"quotation marks\"! Use \"SRB2CE_List_Effects\" to show all effect names.") -- this assumes var1 exists. if it doesn't, then i fucked up big time
        if player and player.khaos then player.khaos.LoadFailed = true end
        return
    end

    --print(effecter.name)
    --print(indexOfName2(KhaosEffectsTable, TheName))

    local TheName = effecter.name

    if effecter.noblacklist then
        CONS_Printf(player, "* \x85"+"ERROR:\x80" + " You can't set this effect to be blacklisted.")
        if player and player.khaos then player.khaos.LoadFailed = true end
        return
    end

    if not SRB2CE.ValidityCheck(player) then
        CONS_Printf(player, "* \x85"+"ERROR:\x80" + " You must be in-game for this to be changed.")
        return
    end

    local alreadyexists = false

    for idx, value in ipairs(player.khaosblacklist) do
        if value == TheName then
            table.remove(player.khaosblacklist, idx)
            alreadyexists = true
        end
    end

    if alreadyexists then
        if not silent then CONS_Printf(player, "* \x84"+"Removed: \x80 " + effecter.name) end
    else
        table.insert(player.khaosblacklist, TheName)
        if not silent then CONS_Printf(player, "* \x84"+"Added: \x80 " + effecter.name) end
    end

    if not silent then
        local name = "client/srb2chaoticedition/" .. "SRB2CEBlackList" .. ".dat"
        local savefile = io.openlocal(name, "w")
        if savefile != nil then
            for key, value in pairs(player.khaosblacklist) do
                savefile:write(value)
                savefile:write("\n")
            end
        end

        --CONS_Printf(player, "* \x83"+"\x80" + "Saved blacklist options!")
    end
end)

COM_AddCommand("SRB2CE_Blacklist_Show", function(player)
    if not SRB2CE.ValidityCheck(player) then
        CONS_Printf(player, "* \x85"+"ERROR:\x80" + " You must be in-game to use this.")
        return
    end

    for idx, value in ipairs(player.khaosblacklist) do
        CONS_Printf(player, "* \x83" + value)
    end
end)]]

COM_AddCommand("SRB2CE_List_Effects", function(player) --if it exceeds 70 characters, line break
    local TableMan = {}
    local StringLength = 0
    local TableInd = 1
    for index, value in ipairs(KhaosEffectsTable) do
        local comma = "\x80, "
        if index == 1 then
        comma = ""
        end
        local Phrase = "\x82" + value.name + comma
        local ThisLength = string.len(Phrase)
        if StringLength + ThisLength > 70 then
            StringLength = 0
            TableInd = $ + 1
        end
        TableMan[TableInd] = (TableMan[TableInd] and Phrase + $) or (Phrase)
        StringLength = $ + ThisLength
    end

    local TableMan2 = {}
    StringLength = 0
    TableInd = 1
    for index, value in ipairs(UniversalEffectsTable) do
        local comma = "\x80, "
        if index == 1 then
        comma = ""
        end
        local Phrase = "\x82" + value.name + comma
        local ThisLength = string.len(Phrase)
        if StringLength + ThisLength > 70 then
            StringLength = 0
            TableInd = $ + 1
        end
        TableMan2[TableInd] = (TableMan2[TableInd] and Phrase + $) or (Phrase)
        StringLength = $ + ThisLength
    end



    CONS_Printf(player, "* \x83" + "List of Khaos Effects:")
    for i = #TableMan, 1, -1 do
        CONS_Printf(player, "\x81" + TableMan[i])
    end


    CONS_Printf(player, "* \x83" + "List of Universal Effects:")
    for i = #TableMan2, 1, -1 do
        CONS_Printf(player, "\x81" + TableMan2[i])
    end
end)

COM_AddCommand("SRB2CE", function(player)
    --summon the menu!
    SRB2CE.SetPlayerMenu(player, "optionsmenu", "optionsmenu")
end)

local function indexOfName2(array, value)
    for i, v in ipairs(array) do
        if string.lower(v.name) == string.lower(value) then -- makes this case insensitive
            return i
        end
    end
    return nil
end

COM_AddCommand("SRB2CE_What_Is", function(player, var1)
    local effecter = var1 and indexOfName2(KhaosEffectsTable, var1)
    if not effecter then
        CONS_Printf(player, "* \x85"+"Invalid effect name!\x80" + "Be sure to use \"quotation marks\"! Use \"SRB2CE_List_Effects\" to show all effect names.")
        return
    else
        effecter = KhaosEffectsTable[effecter]
    end
    local description = ""

    if effecter.description then
        description = effecter.description
        if effecter.description2 then
            description = $ + effecter.description2
        end
    else
        description = "Sorry, this effect doesn't have a description yet!"
    end

    local Color = "\x86"
    if effecter.etype == "good" then
        Color = "\x83"
    elseif effecter.etype == "bad" then
        Color = "\x85"
    end

    CONS_Printf(player, "*\x81 Name" + "\x80: " + Color + effecter.name)
    if effecter.amplifyname then
        CONS_Printf(player, "*\x81 Amplify Name" + "\x80: " + Color + effecter.amplifyname)
    end
    CONS_Printf(player, "*\x81 Description" + "\x80: " + description)
    if effecter.savereq then
        local gex = khaos_intnametable[effecter.savereq] or "???" -- latter will mostly be used for unlockables that don't have names... maybe
        CONS_Printf(player, "*\x81 Requires Unlock" + "\x80: " + gex)
    end

    local Duration = (effecter.duration < 0 and "Infinite")
    or (effecter.duration == 1 and tostring(effecter.duration) + " second")
    or (effecter.duration > 1 and tostring(effecter.duration) + " seconds")
    or "Instant"

    if effecter.duration > 0 and effecter.range then
        local mintime = max(effecter.duration - (effecter.range/3), effecter.duration/2 + 1) -- done to prevent effect from going under half it's time
        local maxtime = effecter.duration + (effecter.range + 2)
        Duration = mintime + "-" + maxtime + " seconds"
    end
    CONS_Printf(player, "*\x81 Duration" + "\x80: " + Duration)


    if effecter.incompatible then
        local incompat = ""
        for index, value in ipairs(effecter.incompatible) do
            incompat = $ + value
            if index != #effecter.incompatible then
                incompat = $ + ", "
            end
        end
        CONS_Printf(player, "*\x81 Incompatible Effects" + "\x80: " + incompat)
    end

    local Attributes = {}

    if effecter.addition then
        local secondman = " second"
        local lengthener = " more"
        if abs(effecter.addition) > 1 then
            secondman = " seconds"
        end
        if effecter.addition < 0 then
            lengthener = " less"
        end
        table.insert(Attributes, " * Lasts " + abs(effecter.addition) + secondman + lengthener + " in Amplify Mode.")
    end

    for key, value in pairs(SRB2CE.AttributeCompareTable) do
        if effecter[key] then
            table.insert(Attributes, value)
        end
    end

    if #Attributes > 0 then
        CONS_Printf(player, "*\x81 Attributes" + "\x80:")
        for index, value in ipairs(Attributes) do
            CONS_Printf(player, value)
        end
    end
end)

--NOTICE: When adding/modifying options, be sure to put it in "SRB2CE.optionstableofdoom", otherwise it won't work.

CV_RegisterVar({name = "SRB2CE_Enabled", defaultvalue = 1, PossibleValue = CV_OnOff, flags = CV_NETVAR, description = "If disabled, the chaos mod will be turned off on the next map."})
CV_RegisterVar({name = "SRB2CE_Amplify", defaultvalue = 0, PossibleValue = CV_OnOff, flags = CV_NETVAR, description = "Makes most effects more intense, often worse."})
CV_RegisterVar({name = "SRB2CE_Universal_Timer", defaultvalue = 30, PossibleValue = {MIN = 10, MAX = 85}, flags = CV_NETVAR, description = "Sets the timer for when Universal Effects are rolled for."})
CV_RegisterVar({name = "SRB2CE_Universal_Fx_Enabled", defaultvalue = 1, PossibleValue = CV_OnOff, flags = CV_NETVAR, description = "If disabled, Universal Effects will be disabled."})
CV_RegisterVar({name = "SRB2CEDebug", defaultvalue = 0, PossibleValue = CV_OnOff, flags = CV_NETVAR, description = "Enables debugging features like printing."})

CV_RegisterVar({name = "SRB2CE_Memory", defaultvalue = 14, PossibleValue = {MIN = 0, MAX = 28}, func = TheNumberFeeder, flags = CV_CALL, description = "Sets how many effects are remembered to not roll for."})
CV_RegisterVar({name = "SRB2CE_Universal_Announcer", defaultvalue = 1, PossibleValue = {off = 0, original = 1, microsam = 2}, func = TheNumberFeeder, flags = CV_CALL, description = "Changes the voice of the announcer for Universal Effects."})
CV_RegisterVar({name = "SRB2CE_Print_Effect", defaultvalue = 1, PossibleValue = CV_OnOff, func = TheNumberFeeder, flags = CV_CALL, description = "Prints to the console when you get an effect applied to you."})
CV_RegisterVar({name = "SRB2CE_Roulette_SFX", defaultvalue = 1, PossibleValue = CV_OnOff, func = TheNumberFeeder, flags = CV_CALL, description = "Plays little sounds when the roulette is active."})
CV_RegisterVar({name = "SRB2CE_HUD", defaultvalue = 1, PossibleValue = CV_OnOff, flags = CV_CALL, func = TheNumberFeeder, description = "If changed, the HUD will be disabled, and most sounds won't play."})

CV_RegisterVar({name = "SRB2CE_HUD_Timer_Bar", defaultvalue = 1, PossibleValue = CV_OnOff, flags = CV_CALL, func = TheNumberFeeder, description = "If changed, the visual timer between effects will be disabled."})
CV_RegisterVar({name = "SRB2CE_HUD_Position", defaultvalue = 0, PossibleValue = {bottom = 0 , top = 1}, flags = CV_CALL, func = TheNumberFeeder, description = "Changes the position of the effects and timerbar."})
CV_RegisterVar({name = "SRB2CE_HUD_Text_Border", defaultvalue = 1, PossibleValue = CV_OnOff, flags = CV_CALL, func = TheNumberFeeder, description = "Enables a border around the text on the HUD."})
CV_RegisterVar({name = "SRB2CE_HUD_Text_Border_Type", defaultvalue = 0, PossibleValue = {default = 0, tkk = 1, round = 2, tri = 3}, flags = CV_CALL, func = TheNumberFeeder, description = "What kind of border do you want?"})
--CV_RegisterVar({name = "SRB2CE_HUD_Override", defaultvalue = 0, PossibleValue = CV_OnOff, flags = CV_CALL, func = TheNumberFeeder, description = "When spectating/viewing replays, your HUD settings will take priority."})
CV_RegisterVar({name = "SRB2CE_HUD_Timer_Color", defaultvalue = 2, PossibleValue = {classic = 0, player = 1, rainbow = 2}, flags = CV_CALL, func = TheNumberFeeder, description = "Changes the colors of the timer bar."})


CV_RegisterVar({name = "SRB2CE_Ringkeep", defaultvalue = 1, PossibleValue = CV_OnOff, func = TheNumberFeeder, flags = CV_CALL})
CV_RegisterVar({name = "SRB2CE_Gamblecore", defaultvalue = 1, PossibleValue = CV_OnOff, func = TheNumberFeeder, flags = CV_CALL})
CV_RegisterVar({name = "SRB2CE_Menu_Music", defaultvalue = 1, PossibleValue = CV_OnOff, func = TheNumberFeeder, flags = CV_CALL})


addHook("NetVars", function(net) -- DO NOT TRY TO USE NETVARS TO SYNC FUNCTIONS. THIS WILL FAIL.
    PreviousUniversalEffect = net(PreviousUniversalEffect)
    SRB2CE.RunFunc = net(SRB2CE.RunFunc)
    ueTimer = net(ueTimer)
    SRB2CE.GravityTimer = net(SRB2CE.GravityTimer)
    ueTimerBuffer = net(ueTimerBuffer)
    ueTimerMax = net(ueTimerMax)
    initiateUE = net(initiateUE)
    SRB2CE.botcount = net(SRB2CE.botcount)
    SRB2CE.playercount = net(SRB2CE.playercount)
    hasanyonefinished = net(hasanyonefinished)
    SRB2CE_Enabled = net(SRB2CE_Enabled)

    --KhaosEffectsTable = net(KhaosEffectsTable) -- < NOTICE: DO NOT DO THIS UNDER ANY FUCKING CIRCUMSTANCES.
end)

--[[table.insert(SRB2CE.TextBoxTable, -- at the bottom just so we get an accurate count of all effects
{{"Did you know? You are",
"playing " + "\x87" + "Tails' Khaos",
"\x87"+ "Kontraption " + "\x81" + SRB2CE.version + "\x80" + ",",
"with \x85" + KhaosTotal + "\x80 effects!"},
{"This was rewritten from",
"the ground up in a few",
"weeks! Now you know!",""}}
)]]

--[[COM_AddCommand("SRB2CE_Help", function(player, var1)
    CONS_Printf(player, "\x83 Specify 1 or 2 for pages!")

    if not var1 or tonumber(var1) == 1 then
        CONS_Printf(player, "List of commands: ")
        CONS_Printf(player, "* " + "SRB2CE_Attributions")
        CONS_Printf(player, "* " + "SRB2CE_Blacklist_Change")
        CONS_Printf(player, "* " + "SRB2CE_Blacklist_Show")
        CONS_Printf(player, "* " + "SRB2CE_List_Effects")
        CONS_Printf(player, "* " + "SRB2CE_What_Is")
        CONS_Printf(player, "* " + "SRB2CE_Save_Settings")
        CONS_Printf(player, "* " + "SRB2CE_Load_Settings")
    end

    if tonumber(var1) == 2 then
        CONS_Printf(player, "List of convars: ")
        for key, value in pairs(SRB2CE.optionstableofdoom) do
            CONS_Printf(player, "* " + value.name)
        end
    end

    CONS_Printf(player, "SRB2: Chaotic Edition v" .. SRB2CE.version ..", effects count: " .. SRB2CE.KhaosTotal .. ", & universal effects: " .. #SRB2CE.UniversalEffectsTable)
end)]]



Khaos_LoadSettings() -- force load the cvars

print("SRB2: Chaotic Edition v" .. SRB2CE.version .." loaded! Total number of effects: " .. SRB2CE.KhaosTotal .. ", and universal effects: " .. #SRB2CE.UniversalEffectsTable)


-- DEBUG TOOL USED TO SEE IF THERE'S ANY INCOMPATIBLE EFFECTS THAT DON'T ACTUALLY EXIST
--[[local FunnyFumber = 0

repeat
    FunnyFumber = $ + 1
    local Gexfect = KhaosEffectsTable[FunnyFumber]
    if Gexfect.incompatible then
        for key, value in pairs(Gexfect.incompatible) do
            local ThePlacer = indexOfName(KhaosEffectsTable, value)
            if not ThePlacer then
                print("INVALID INCOMPATIBLE EFFECT: " + value, ", FOR: " + Gexfect.name)
            end
        end
    end

until FunnyFumber == #KhaosEffectsTable]]
